@@ -10,7 +10,7 @@
 
 ;;; Code:
 
-(defvar agda2-version "2.4.2.5"
+(defvar agda2-version "2.5.1"
   "The version of the Agda mode.
 Note that the same version of the Agda executable must be used.")
 
@@ -78,14 +78,6 @@ All the elements of XS are included, in their original order."
   "Major mode for interactively developing Agda programs."
   :group 'languages)
 
-(defcustom agda2-include-dirs
-  '(".")
-  "The directories Agda uses to search for files.
-The directory names should either be absolute or be relative to
-the root of the current project."
-  :type '(repeat directory)
-  :group 'agda2)
-
 (defcustom agda2-program-name "agda"
   "The name of the Agda executable."
   :type 'string
@@ -205,7 +197,7 @@ to this variable to take effect."
                         (modify-syntax-entry keys "w" tbl)))
                     (standard-syntax-table))
     ;; Then override the remaining special cases.
-    (dolist (cs '((?{ . "(}1n") (?} . "){4n") (?- . "w 123b") (?\n . "> b")
+    (dolist (cs '((?- . "w 12b") (?\n . "> b")
                   (?. . ".") (?\; . ".") (?_ . ".") (?! . ".")))
       (modify-syntax-entry (car cs) (cdr cs) tbl))
     tbl)
@@ -246,6 +238,7 @@ constituents.")
     (agda2-why-in-scope-maybe-toplevel       "\C-c\C-w"           (local global) "Explain why a particular name is in scope")
     (agda2-goal-and-context                  ,(kbd "C-c C-,")     (local)        "Goal type and context")
     (agda2-goal-and-context-and-inferred     ,(kbd "C-c C-.")     (local)        "Goal type, context and inferred type")
+    (agda2-search-about-toplevel             ,(kbd "C-c C-z")     (local global) "Search About")
     (agda2-module-contents-maybe-toplevel    ,(kbd "C-c C-o")     (local global) "Module contents")
     (agda2-compute-normalised-maybe-toplevel "\C-c\C-n"           (local global) "Evaluate term to normal form")
     (describe-char                           nil                  (global)       "Information about the character at point")
@@ -373,6 +366,10 @@ The command which arrived last is stored first in the list.")
   "The Agda buffer.
 Note that this variable is not buffer-local.")
 
+(defvar agda2-in-agda2-file-buffer nil
+  "Was `agda2-file-buffer' active when `agda2-output-filter' started?
+Note that this variable is not buffer-local.")
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;; agda2-mode
 
@@ -397,17 +394,15 @@ The following paragraph does not apply to Emacs 23 or newer.
 Special commands:
 \\{agda2-mode-map}"
 
- ;; Check that the right version of Agda is used.
- (let* ((coding-system-for-read 'utf-8)
-        (output (with-output-to-string
-                  (call-process agda2-program-name
-                                nil standard-output nil "--version")))
-        (version (and (string-match "^Agda version \\([0-9.]+\\)$" output)
-                      (match-string 1 output))))
-   (unless (equal version agda2-version)
-     (error "The Agda mode's version (%s) does not match that of %s (%s)."
-            agda2-version
-            agda2-program-name (or version "unknown"))))
+ (if (boundp 'agda2-include-dirs)
+     (display-warning 'agda2 "Note that the variable agda2-include-dirs is
+no longer used. You may want to update your configuration. You
+have at least two choices:
+* Use the library management system.
+* Set the include path using agda2-program-args.
+
+One way to avoid seeing this warning is to make sure that
+agda2-include-dirs is not bound." :warning))
 
  (setq local-abbrev-table agda2-mode-abbrev-table
        indent-tabs-mode   nil
@@ -445,38 +440,60 @@ Special commands:
  (add-hook 'change-major-mode-hook 'agda2-quit nil 'local))
 
 (defun agda2-restart ()
-  "Kill and restart the *agda2* buffer and load `agda2-toplevel-module'."
+  "Tries to start or restart the Agda process."
   (interactive)
-  (save-excursion (let ((agda2-bufname "*agda2*"))
-                    (condition-case nil
-                      (agda2-term)
-                      (error nil))
 
-                    ;; Start the Agda2 process in a new buffer.
-                    (message "Starting agda process `%s'." agda2-program-name)
-                    (let ((process-connection-type nil)) ; pipes are faster than PTYs
-                      (apply 'start-process "Agda2" agda2-bufname
-                             agda2-program-name
-                             (cons "--interaction" agda2-program-args)))
+  ;; Kill any running instance of the Agda process.
+  (condition-case nil
+      (agda2-term)
+    (error nil))
 
-                    ;; Select agda buffer temporarily.
-                    (with-current-buffer agda2-bufname
-                      (setq process-adaptive-read-buffering t)
+  ;; Check that the right version of Agda is used.
+  (let* ((coding-system-for-read 'utf-8)
+         (output (with-output-to-string
+                   (call-process agda2-program-name
+                                 nil standard-output nil "--version")))
+         (version (and (string-match "^Agda version \\([0-9.]+\\)$" output)
+                       (match-string 1 output))))
+    (unless (equal version agda2-version)
+      (error "The Agda mode's version (%s) does not match that of %s (%s)."
+             agda2-version
+             agda2-program-name (or version "unknown"))))
 
-                      ;; Clear message area.
-                      (message "")
+  (let ((all-program-args (cons "--interaction" agda2-program-args)))
 
-                      (setq agda2-process        (get-buffer-process agda2-bufname)
-                            agda2-process-buffer (process-buffer agda2-process)
-                            agda2-in-progress    nil
-                            mode-name            "Agda executable"
-                            agda2-last-responses nil)
-                      (set-buffer-file-coding-system 'utf-8)
-                      (set-buffer-process-coding-system 'utf-8 'utf-8)
-                      (set-process-query-on-exit-flag agda2-process nil))))
-  (setq agda2-file-buffer (current-buffer))
-  (set-process-filter agda2-process 'agda2-output-filter)
-  (agda2-remove-annotations))
+    ;; Check that the arguments are not malformed.
+    (let* ((coding-system-for-read 'utf-8)
+           (status)
+           (output
+            (with-output-to-string
+              (setq status
+                    (apply 'call-process agda2-program-name
+                           nil standard-output nil all-program-args)))))
+      (unless (equal status 0)
+        (error "Failed to start the Agda process:\n%s" output)))
+
+    ;; Start the Agda process.
+    (let ((agda2-bufname "*agda2*"))
+
+      (let ((process-connection-type nil)) ; Pipes are faster than PTYs.
+        (setq agda2-process
+              (apply 'start-process "Agda2" agda2-bufname
+                     agda2-program-name all-program-args)))
+
+      (set-process-coding-system agda2-process 'utf-8 'utf-8)
+      (set-process-query-on-exit-flag agda2-process nil)
+      (set-process-filter agda2-process 'agda2-output-filter)
+      (setq agda2-in-progress nil
+            agda2-file-buffer (current-buffer))
+
+      (with-current-buffer agda2-bufname
+        (setq agda2-process-buffer (current-buffer)
+              mode-name            "Agda executable"
+              agda2-last-responses nil)
+        (set-buffer-file-coding-system 'utf-8))
+
+      (agda2-remove-annotations))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;; Communicating with Agda
@@ -578,6 +595,9 @@ reloaded from `agda2-highlighting-file', unless
   ;; extract the fontified text and kill the temp buffer; so when Agda
   ;; finally answers, the temp buffer is long gone.
   (when (buffer-live-p agda2-file-buffer)
+  (setq agda2-in-agda2-file-buffer
+        (and agda2-file-buffer
+             (equal (current-buffer) agda2-file-buffer)))
   (let (;; The input lines in the current chunk.
         (lines (split-string chunk "\n"))
 
@@ -609,19 +629,23 @@ reloaded from `agda2-highlighting-file', unless
                               (if (and (listp (car result))
                                        (= (cdr result) (length line)))
                                   (car result)))
-                          (error nil))))
+                          (error nil)))
+                   (is-highlighting-command
+                    (and cmd
+                         (symbolp (car cmd))
+                         (let ((case-fold-search nil))
+                           (string-match "^agda2-highlight-"
+                                         (symbol-name (car cmd)))))))
 
+              ;; Do not echo highlighting commands.
+              (unless is-highlighting-command
+                (with-current-buffer agda2-process-buffer
+                  (save-excursion
+                    (goto-char (point-max))
+                    (insert line)
+                    (insert "\n"))))
               (when cmd
-                (unless (and (symbolp (car cmd))
-                             (let ((case-fold-search nil))
-                               (string-match "^agda2-highlight-"
-                                             (symbol-name (car cmd)))))
-                    ;; Do not echo highlighting commands.
-                  (with-current-buffer agda2-process-buffer
-                    (save-excursion
-                      (goto-char (point-max))
-                      (insert line)
-                      (insert "\n")))
+                (unless is-highlighting-command
                   (incf agda2-responses))
                 (if (equal 'last (car-safe (car cmd)))
                     (push (cons (cdr (car cmd)) (cdr cmd))
@@ -724,7 +748,7 @@ An error is raised if no responses are received."
             (t (setq input-from-goal t)))
       (apply 'agda2-go t t t cmd
              (format "%d" g)
-             (if input-from-goal (agda2-goal-Range o) "noRange")
+             (if input-from-goal (agda2-goal-Range o) (agda2-mkRange nil))
              (agda2-string-quote txt) args))))
 
 ;; Note that the following function is a security risk, since it
@@ -746,7 +770,7 @@ An error is raised if no responses are received."
   (interactive)
   (agda2-go t t t "Cmd_load"
             (agda2-string-quote (buffer-file-name))
-            (agda2-list-quote agda2-include-dirs)
+            (agda2-list-quote agda2-program-args)
             ))
 
 (defun agda2-measure-load-time
@@ -779,7 +803,7 @@ The variable `agda2-backend' determines which backend is used."
   (agda2-go t t t "Cmd_compile"
             agda2-backend
             (agda2-string-quote (buffer-file-name))
-            (agda2-list-quote agda2-include-dirs)
+            (agda2-list-quote agda2-program-args)
             ))
 
 (defun agda2-give()
@@ -904,49 +928,73 @@ buffer, and point placed after this text.
 If APPEND is nil, then any previous text is removed before TEXT
 is inserted, and point is placed before this text."
   (interactive)
-  (with-current-buffer (agda2-info-buffer)
-    ;; In some cases the jump-to-position-mentioned-in-text
-    ;; functionality (see compilation-error-regexp-alist above) didn't
-    ;; work: Emacs jumped to the wrong position. However, it seems to
-    ;; work if compilation-forget-errors is used. This problem may be
-    ;; related to Emacs bug #9679
-    ;; (http://debbugs.gnu.org/cgi/bugreport.cgi?bug=9679). The idea
-    ;; to use compilation-forget-errors comes from a comment due to
-    ;; Oleksandr Manzyuk
-    ;; (https://github.com/haskell/haskell-mode/issues/67).
-    (compilation-forget-errors)
-    (unless append (erase-buffer))
-    (save-excursion
-      (goto-char (point-max))
-      (insert text))
-    (put-text-property 0 (length name) 'face '(:weight bold) name)
-    (setq mode-line-buffer-identification name)
-    (save-selected-window
-      (let (;; If there is only one window, then the info window
-            ;; should be created above or below the code window, not
-            ;; to the left or right.
-            (split-width-threshold nil)
-            (buf (current-buffer))
-           )
-        ;; Andreas, 2014-02-23, issue 1061
-        ;; If the buffer is already displayed in some window,
-        ;; do not display it again.  Also, do not raise its frame.
-        ;; This allows undisturbed working on something else while
-        ;; Agda is type-checking.
-        ;; The solution is to query for the window displaying the buffer
-        ;; via get-buffer-window.  Only if it return nil, we pop-to-buffer.
-        ;; Credits go to Iqbal Ansari who anwered my question on
-        ;; http://stackoverflow.com/questions/21955162/emacs-how-to-display-a-buffer-without-switching-window-and-without-raising-fram
-        ;; 2014-03-01 DISABLED FIX
-        ;; (unless (get-buffer-window buf t)
-          (pop-to-buffer buf nil 'norecord)
-          (fit-window-to-buffer nil
-            (truncate
-              (* (frame-height) agda2-information-window-max-height)))
-        ;; )
-        (if append
-            (goto-char (point-max))
-          (goto-char (point-min)))))))
+  (let ((buf (agda2-info-buffer)))
+    (with-current-buffer buf
+      ;; In some cases the jump-to-position-mentioned-in-text
+      ;; functionality (see compilation-error-regexp-alist above)
+      ;; didn't work: Emacs jumped to the wrong position. However, it
+      ;; seems to work if compilation-forget-errors is used. This
+      ;; problem may be related to Emacs bug #9679
+      ;; (http://debbugs.gnu.org/cgi/bugreport.cgi?bug=9679). The idea
+      ;; to use compilation-forget-errors comes from a comment due to
+      ;; Oleksandr Manzyuk
+      ;; (https://github.com/haskell/haskell-mode/issues/67).
+      (compilation-forget-errors)
+      (unless append (erase-buffer))
+      (save-excursion
+        (goto-char (point-max))
+        (insert text))
+      (put-text-property 0 (length name) 'face '(:weight bold) name)
+      (setq mode-line-buffer-identification name)
+      (force-mode-line-update))
+    ;; If the current window displays the information buffer, then the
+    ;; window configuration is left untouched.
+    (unless (equal (window-buffer) buf)
+      (let ((agda-window
+              (and agda2-file-buffer
+                   (car-safe
+                     ;; All windows, including minibuffers, on any
+                     ;; frame on the current terminal, displaying the
+                     ;; present Agda file buffer.
+                     (get-buffer-window-list agda2-file-buffer t 0)))))
+        (save-selected-window
+          ;; Select a window displaying the Agda file buffer (if such
+          ;; a window exists). With certain configurations of
+          ;; display-buffer this should increase the likelihood that
+          ;; the info buffer will be displayed on the same frame.
+          (when agda-window
+            (select-window agda-window 'no-record))
+          (let* (;; If there is only one window, then the info window
+                 ;; should be created above or below the code window,
+                 ;; not to the left or right.
+                 (split-width-threshold nil)
+                 (window
+                   (display-buffer
+                     buf
+                     ;; Under Emacs 23 the effect of the following
+                     ;; argument is only that the current window
+                     ;; should not be used.
+                     '(nil
+                       .
+                       (;; Do not use the same window.
+                        (inhibit-same-window . t)
+                        ;; Do not raise or select another frame.
+                        (inhibit-switch-frame . t))))))
+            (if window
+                (fit-window-to-buffer window
+                  (truncate
+                    (* (frame-height)
+                       agda2-information-window-max-height))))))))
+    ;; Move point in every window displaying the information buffer.
+    ;; Exception: If we are appending, don't move point in selected
+    ;; windows.
+    (dolist (window (get-buffer-window-list buf 'no-minibuffer t))
+      (unless (and append
+                   (equal window (selected-window)))
+        (with-selected-window window
+          (if append
+              (goto-char (point-max))
+            (goto-char (point-min))))))))
 
 (defun agda2-info-action-and-copy (name text &optional append)
   "Same as agda2-info-action but also puts TEXT in the kill ring."
@@ -1147,6 +1195,13 @@ Along with their types."
   "Module name"
 )
 
+(agda2-maybe-normalised-toplevel
+  agda2-search-about-toplevel
+  "Search About an identifier"
+  "Cmd_search_about_toplevel"
+  "Name"
+)
+
 (defun agda2-module-contents-maybe-toplevel ()
   "Shows all the top-level names in the given module.
 Along with their types.
@@ -1249,7 +1304,7 @@ ways."
        ;; Don't run modification hooks: we don't want this function to
        ;; trigger agda2-abort-highlighting.
        (inhibit-modification-hooks t))
-      ((delims() (re-search-forward "[?]\\|[{][-!]\\|[-!][}]\\|--\\|\\\\begin{code}\\|\\\\end{code}" nil t))
+      ((delims() (re-search-forward "[?]\\|[{][-!]\\|[-!][}]\\|--\\|^%.*\\\\begin{code}\\|\\\\begin{code}\\|\\\\end{code}" nil t))
        ;; is-proper checks whether string s (e.g. "?" or "--") is proper
        ;; i.e., is not part of an identifier.
        ;; comment-starter is true if s starts a comment (e.g. "--")
@@ -1359,7 +1414,7 @@ text properties."
         (if (and (not (equal new-txt 'paren)) (not (equal new-txt 'no-paren)))
             (apply 'agda2-go t t nil "Cmd_highlight"
               (format "%d" old-g)
-              (agda2-mkRange p (- q 2))
+              (agda2-mkRange `(,p ,(- q 2)))
               (agda2-string-quote new-txt) nil))
     )))
 
@@ -1377,24 +1432,30 @@ text properties."
 
 (defun agda2-goal-Range (o)
   "The Haskell Range of goal overlay O."
-  (agda2-mkRange (+ (overlay-start o) 2)
-                 (- (overlay-end   o) 2)))
+  (agda2-mkRange `(,(+ (overlay-start o) 2)
+                   ,(- (overlay-end   o) 2))))
 
-(defun agda2-mkRange (p q)
-  "The Haskell Range corresponding to two points."
-  (format "(Range [Interval %s %s])"
-          (agda2-mkPos p)
-          (agda2-mkPos q)))
+(defun agda2-mkRange (points)
+  "A string representing a range corresponding to POINTS.
+POINTS must be a list of integers, and its length must be 0 or 2."
+  (if points
+      (format "(intervalsToRange (Just (mkAbsolute %s)) %s)"
+              (agda2-string-quote (file-truename (buffer-file-name)))
+              (format "[Interval %s %s]"
+                      (agda2-mkPos (car points))
+                      (agda2-mkPos (cadr points))))
+    "noRange"))
 
 (defun agda2-mkPos (&optional p)
-  "The Haskell Position corresponding to P or `point'."
+  "The Haskell PositionWithoutFile corresponding to P or `point'."
   (save-excursion
-    (if p (goto-char p))
-    (format "(Pn (Just (mkAbsolute %s)) %d %d %d)"
-            (agda2-string-quote (file-truename (buffer-file-name)))
-            (point)
-            (count-lines (point-min) (point))
-            (1+ (current-column)))))
+    (save-restriction
+      (widen)
+      (if p (goto-char p))
+      (format "(Pn () %d %d %d)"
+              (point)
+              (count-lines (point-min) (point))
+              (1+ (current-column))))))
 
 (defun agda2-char-quote (c)
   "Convert character C to the notation used in Haskell strings.
@@ -1558,11 +1619,6 @@ From the beginning of the current line to the end of the buffer."
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; Go to definition site
 
-(defun agda2-goto (filepos &optional other-window)
-  "Like `annotation-goto', unless `agda2-highlight-in-progress' is nil."
-  (if agda2-highlight-in-progress
-      (annotation-goto filepos other-window)))
-
 (defun agda2-goto-definition-keyboard (&optional other-window)
   "Go to the definition site of the name under point (if any).
 If this function is invoked with a prefix argument then another window is used
@@ -1586,6 +1642,35 @@ invoked."
   (interactive)
   (annotation-go-back))
 
+(defun agda2-maybe-goto (filepos)
+  "Might move point to the given error.
+FILEPOS should have the form (FILE . POSITION).
+
+If `agda2-highlight-in-progress' is nil, then nothing happens.
+Otherwise, if the current buffer is the one that is connected to
+the Agda process, then point is moved to POSITION in
+FILE (assuming that the FILE is readable). Otherwise point is
+moved to the given position in the buffer visiting the file, if
+any, and in every window displaying the buffer, but the window
+configuration and the selected window are not changed."
+  (when (and agda2-highlight-in-progress
+             (consp filepos)
+             (stringp (car filepos))
+             (integerp (cdr filepos)))
+    (if agda2-in-agda2-file-buffer
+        (annotation-goto-and-push (current-buffer) (point) filepos)
+      (save-excursion
+        (let ((buffer (find-buffer-visiting (car filepos))))
+          (when buffer
+            (let ((windows (get-buffer-window-list buffer
+                                                   'no-minibuffer t)))
+              (if windows
+                  (dolist (window windows)
+                    (with-selected-window window
+                      (annotation-goto-position (cdr filepos))))
+                (with-current-buffer buffer
+                  (annotation-goto-position (cdr filepos)))))))))))
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; Implicit arguments
 
