@@ -1,31 +1,36 @@
-{-# LANGUAGE CPP #-}
-{-# LANGUAGE DeriveFunctor #-}
-{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE CPP                   #-}
+{-# LANGUAGE DeriveFunctor         #-}
+{-# LANGUAGE FlexibleInstances     #-}
+{-# LANGUAGE FlexibleContexts      #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
-{-# LANGUAGE TupleSections #-}
-{-# LANGUAGE UndecidableInstances #-}
+{-# LANGUAGE ScopedTypeVariables   #-}
+{-# LANGUAGE TupleSections         #-}
+{-# LANGUAGE UndecidableInstances  #-}
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
 module Agda.Interaction.BasicOps where
 
+import Prelude hiding (null)
+
 import Control.Arrow ((***), first, second)
-import Control.Applicative
+import Control.Applicative hiding (empty)
 import Control.Monad.Reader
 import Control.Monad.State
 import Control.Monad.Identity
 
 import qualified Data.Map as Map
-import Data.List
+import Data.List hiding (null)
 import Data.Maybe
-import Data.Traversable hiding (mapM, forM)
+import Data.Traversable hiding (mapM, forM, for)
+import Data.Monoid
 
 import qualified Agda.Syntax.Concrete as C -- ToDo: Remove with instance of ToConcrete
 import Agda.Syntax.Position
-import Agda.Syntax.Abstract as A hiding (Open, Apply)
+import Agda.Syntax.Abstract as A hiding (Open, Apply, Assign)
 import Agda.Syntax.Abstract.Views as A
 import Agda.Syntax.Common
-import Agda.Syntax.Info (ExprInfo(..),MetaInfo(..),emptyMetaInfo)
+import Agda.Syntax.Info (ExprInfo(..),MetaInfo(..),emptyMetaInfo,exprNoRange)
 import qualified Agda.Syntax.Info as Info
 import Agda.Syntax.Internal as I
 import Agda.Syntax.Translation.InternalToAbstract
@@ -49,6 +54,7 @@ import Agda.TypeChecking.Records
 import Agda.TypeChecking.Irrelevance (wakeIrrelevantVars)
 import Agda.TypeChecking.Pretty (prettyTCM)
 import Agda.TypeChecking.Free
+import Agda.TypeChecking.CheckInternal
 import qualified Agda.TypeChecking.Pretty as TP
 
 import Agda.Utils.Except ( Error(strMsg), MonadError(catchError, throwError) )
@@ -57,6 +63,7 @@ import Agda.Utils.Lens
 import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
+import Agda.Utils.Null
 import Agda.Utils.Pretty
 import Agda.Utils.Permutation
 import Agda.Utils.Size
@@ -67,7 +74,11 @@ import Agda.Utils.Impossible
 -- | Parses an expression.
 
 parseExpr :: Range -> String -> TCM C.Expr
-parseExpr rng s = liftIO $ parsePosString exprParser pos s
+parseExpr rng s = do
+  C.ExprWhere e wh <- liftIO $ parsePosString exprWhereParser pos s
+  unless (null wh) $ typeError $ GenericError $
+    "where clauses are not supported in holes"
+  return e
   where pos = fromMaybe (startPos Nothing) $ rStart rng
 
 parseExprIn :: InteractionId -> Range -> String -> TCM Expr
@@ -89,13 +100,13 @@ giveExpr mi e = do
       metaTypeCheck mv (mvJudgement mv)
   where
     metaTypeCheck mv IsSort{}      = __IMPOSSIBLE__
-    metaTypeCheck mv (HasType _ t) = do
+    metaTypeCheck mv (HasType _ t) = disableDestructiveUpdate $ do
       reportSDoc "interaction.give" 20 $
         TP.text "give: meta type =" TP.<+> prettyTCM t
       -- Here, we must be in the same context where the meta was created.
       -- Thus, we can safely apply its type to the context variables.
       ctx <- getContextArgs
-      let t' = t `piApply` permute (takeP (length ctx) $ mvPermutation mv) ctx
+      t' <- t `piApplyM` permute (takeP (length ctx) $ mvPermutation mv) ctx
       traceCall (CheckExprCall e t') $ do
         reportSDoc "interaction.give" 20 $
           TP.text "give: instantiated meta type =" TP.<+> prettyTCM t'
@@ -136,7 +147,7 @@ give ii mr e = liftTCM $ do
   mi  <- lookupInteractionId ii
   whenJust mr $ updateMetaVarRange mi
   reportSDoc "interaction.give" 10 $ TP.text "giving expression" TP.<+> prettyTCM e
-  reportSDoc "interaction.give" 50 $ TP.text $ show $ deepUnScope e
+  reportSDoc "interaction.give" 50 $ TP.text $ show $ deepUnscope e
   -- Try to give mi := e
   _ <- catchError (giveExpr mi e) $ \ err -> case err of
     -- Turn PatternErr into proper error:
@@ -165,7 +176,7 @@ refine ii mr e = do
   reportSDoc "interaction.refine" 10 $
     TP.text "refining with expression" TP.<+> prettyTCM e
   reportSDoc "interaction.refine" 50 $
-    TP.text $ show $ deepUnScope e
+    TP.text $ show $ deepUnscope e
   -- We try to append up to 10 meta variables
   tryRefine 10 range scope e
   where
@@ -174,7 +185,7 @@ refine ii mr e = do
       where
         try :: Int -> Expr -> TCM Expr
         try 0 e = throwError $ strMsg "Cannot refine"
-        try n e = give ii (Just r) e `catchError` (\_ -> try (n-1) =<< appMeta e)
+        try n e = give ii (Just r) e `catchError` (\_ -> try (n - 1) =<< appMeta e)
 
         -- Apply A.Expr to a new meta
         appMeta :: Expr -> TCM Expr
@@ -189,7 +200,25 @@ refine ii mr e = do
                 , metaNameSuggestion = ""
                 }
               metaVar = QuestionMark info ii
-          return $ App (ExprRange r) e $ defaultNamedArg metaVar
+
+              count x e = getSum $ foldExpr isX e
+                where isX (A.Var y) | x == y = Sum 1
+                      isX _                  = mempty
+
+              lamView (A.Lam _ (DomainFree _ x) e) = Just (x, e)
+              lamView (A.Lam i (DomainFull (TypedBindings r (Arg ai (TBind br (x : xs) a)))) e)
+                | null xs   = Just (dget x, e)
+                | otherwise = Just (dget x, A.Lam i (DomainFull $ TypedBindings r $ Arg ai $ TBind br xs a) e)
+              lamView _ = Nothing
+
+              -- reduce beta-redexes where the argument is used at most once
+              smartApp i e arg =
+                case lamView $ unScope e of
+                  Just (x, e) | count x e < 2 -> mapExpr subX e
+                    where subX (A.Var y) | x == y = namedArg arg
+                          subX e = e
+                  _ -> App i e arg
+          return $ smartApp (ExprRange r) e $ defaultNamedArg metaVar
           --ToDo: The position of metaVar is not correct
           --ToDo: The fixity of metavars is not correct -- fixed? MT
 
@@ -294,22 +323,25 @@ instance Reify Constraint (OutputConstraint Expr Expr) where
             CheckLambda (Arg ai (xs, mt)) body target -> do
               domType <- maybe (return underscore) reify mt
               target  <- reify target
-              let bs = TypedBindings noRange $ Arg (mapArgInfoColors (const []) ai) $
+              let bs = TypedBindings noRange $ Arg ai $
                        TBind noRange xs domType
                   e  = A.Lam Info.exprNoRange (DomainFull bs) body
               return $ TypedAssign m' e target
-            CheckArgs _ _ _ args t0 t1 _ -> do
+            CheckArgs _ _ args t0 t1 _ -> do
               t0 <- reify t0
               t1 <- reify t1
               return $ PostponedCheckArgs m' (map (namedThing . unArg) args) t0 t1
+            UnquoteTactic tac _ goal -> do
+              tac <- A.App exprNoRange (A.Unquote exprNoRange) . defaultNamedArg <$> reify tac
+              OfType tac <$> reify goal
           Open{}  -> __IMPOSSIBLE__
           OpenIFS{}  -> __IMPOSSIBLE__
           InstS{} -> __IMPOSSIBLE__
           InstV{} -> __IMPOSSIBLE__
-    reify (FindInScope m mcands) = FindInScopeOF
+    reify (FindInScope m _b mcands) = FindInScopeOF
       <$> (reify $ MetaV m [])
       <*> (reify =<< getMetaType m)
-      <*> (forM (fromMaybe [] mcands) $ \ (tm, ty) -> do
+      <*> (forM (fromMaybe [] mcands) $ \ (Candidate tm ty eti) -> do
             (,) <$> reify tm <*> reify ty)
     reify (IsEmpty r a) = IsEmptyType <$> reify a
     reify (CheckSizeLtSat a) = SizeLtSat  <$> reify a
@@ -464,7 +496,8 @@ typeOfMetaMI norm mi =
           ]
         ]
       reportSDoc "interactive.meta.scope" 20 $ TP.text $ show $ getMetaScope mv
-      OfType x <$> reify (t `piApply` permute (takeP (size vs) $ mvPermutation mv) vs)
+      -- Andreas, 2016-01-19, issue #1783: need piApplyM instead of just piApply
+      OfType x <$> do reify =<< t `piApplyM` permute (takeP (size vs) $ mvPermutation mv) vs
     rewriteJudg mv (IsSort i t) = do
       ms <- getMetaNameSuggestion i
       return $ JustSort $ NamedMeta ms i
@@ -499,7 +532,7 @@ metaHelperType :: Rewrite -> InteractionId -> Range -> String -> TCM (OutputCons
 metaHelperType norm ii rng s = case words s of
   []    -> fail "C-c C-h expects an argument of the form f e1 e2 .. en"
   f : _ -> do
-    A.Application h args <- A.appView . getBody . deepUnScope <$> parseExprIn ii rng ("let " ++ f ++ " = _ in " ++ s)
+    A.Application h args <- A.appView . getBody . deepUnscope <$> parseExprIn ii rng ("let " ++ f ++ " = _ in " ++ s)
     withInteractionId ii $ do
       cxtArgs  <- getContextArgs
       -- cleanupType relies on with arguments being named 'w',
@@ -510,9 +543,9 @@ metaHelperType norm ii rng s = case words s of
       -- Remember the arity of a
       TelV atel _ <- telView a
       let arity = size atel
-          (delta1, delta2, _, a', as', vs') = splitTelForWith tel a as vs
+          (delta1, delta2, _, a', as', vs') = splitTelForWith tel a (map OtherType as) vs
       a <- local (\e -> e { envPrintDomainFreePi = True }) $ do
-        reify =<< cleanupType arity args =<< normalForm norm =<< withFunctionType delta1 vs' as' delta2 a'
+        reify =<< cleanupType arity args =<< normalForm norm =<< fst <$> withFunctionType delta1 vs' as' delta2 a'
       return (OfType' h a)
   where
     cleanupType arity args t = do
@@ -569,7 +602,6 @@ metaHelperType norm ii rng s = case words s of
       I.Level{}    -> pure v
       I.MetaV{}    -> pure v
       I.Shared{}   -> pure v
-      I.ExtLam{}   -> __IMPOSSIBLE__
     onNamesElims f = traverse $ traverse $ onNamesTm f
     onNamesArgs f  = traverse $ traverse $ onNamesTm f
     onNamesAbs f   = onNamesAbs' f (stringToArgName <.> f . argNameToString)
@@ -597,22 +629,20 @@ metaHelperType norm ii rng s = case words s of
 contextOfMeta :: InteractionId -> Rewrite -> TCM [OutputConstraint' Expr Name]
 contextOfMeta ii norm = do
   info <- getMetaInfo <$> (lookupMeta =<< lookupInteractionId ii)
-  let localVars = map ctxEntry . envContext . clEnv $ info
-      letVars = map (\(n, OpenThing _ (tm, (Dom c ty))) -> Dom c (n, ty))
-                    $ Map.toDescList . envLetBindings . clEnv $ info
-  withMetaInfo info $ gfilter visible <$> reifyContext (length letVars)
-                                                       (letVars ++ localVars)
+  withMetaInfo info $ do
+    cxt <- getContext
+    let n         = length cxt
+        localVars = zipWith raise [1..] cxt
+        mkLet (x, lb) = do
+          (tm, Dom c ty) <- getOpen lb
+          return $ Dom c (x, ty)
+    letVars <- mapM mkLet . Map.toDescList =<< asks envLetBindings
+    gfilter visible . reverse <$> mapM out (letVars ++ localVars)
   where gfilter p = catMaybes . map p
         visible (OfType x y) | not (isNoName x) = Just (OfType' x y)
                              | otherwise        = Nothing
         visible _            = __IMPOSSIBLE__
-        reifyContext skip xs =
-          reverse <$> zipWithM out
-                               -- don't escape context for letvars
-                               (replicate skip 0 ++ [1..])
-                               xs
-
-        out i (Dom _ (x, t)) = escapeContext i $ do
+        out (Dom _ (x, t)) = do
           t' <- reify =<< normalForm norm t
           return $ OfType x t'
 
@@ -656,7 +686,7 @@ introTactic pmLambda ii = do
   mv <- lookupMeta mi
   withMetaInfo (getMetaInfo mv) $ case mvJudgement mv of
     HasType _ t -> do
-        t <- reduce =<< piApply t <$> getContextArgs
+        t <- reduce =<< piApplyM t =<< getContextArgs
         -- Andreas, 2013-03-05 Issue 810: skip hidden domains in introduction
         -- of constructor.
         TelV tel' t <- telViewUpTo' (-1) notVisible t
@@ -700,7 +730,7 @@ introTactic pmLambda ii = do
             okHiding    = if allHidden then const True else okHiding0
         vars <- -- setShowImplicitArguments (imp || allHidden) $
                 (if allHidden then withShowAllArguments else id) $
-                  mapM showTCM [ setHiding h $ defaultArg $ var i :: I.Arg Term
+                  mapM showTCM [ setHiding h $ defaultArg $ var i :: Arg Term
                                | (h, i) <- zip hs $ downFrom n
                                , okHiding h
                                ]
@@ -715,8 +745,8 @@ introTactic pmLambda ii = do
         makeName (x, t)   = (x, t)
 
     introData t = do
-      let tel  = telFromList [domFromArg $ defaultArg ("_", t)]
-          pat  = [defaultArg $ unnamed $ I.VarP "c"]
+      let tel  = telFromList [defaultDom ("_", t)]
+          pat  = [defaultArg $ unnamed $ I.VarP (0,"c")]
       r <- splitLast CoInductive tel pat
       case r of
         Left err -> return []
@@ -728,7 +758,8 @@ introTactic pmLambda ii = do
       fs <- ifM showImplicitArguments
             (return $ map unArg hfs)
             (return [ unArg a | a <- hfs, getHiding a == NotHidden ])
-      let e = C.Rec noRange $ map (, C.QuestionMark noRange Nothing) fs
+      let e = C.Rec noRange $ for fs $ \ f ->
+            Left $ C.FieldAssignment f $ C.QuestionMark noRange Nothing
       return [ prettyShow e ]
 
 -- | Runs the given computation as if in an anonymous goal at the end
@@ -746,7 +777,7 @@ atTopLevel m = inConcreteMode $ do
       -- and put them into the context.
       let names :: [A.Name]
           names = reverse $ map snd $ notShadowedLocals $ scopeLocals scope
-          types :: [I.Dom I.Type]
+          types :: [Dom I.Type]
           types = map (snd <$>) $ telToList tel
           gamma :: ListTel' A.Name
           gamma = zipWith' (\ x dom -> (x,) <$> dom) names types
@@ -765,32 +796,79 @@ parseName r s = do
     _                      -> typeError $
       GenericError $ "Not an identifier: " ++ show m ++ "."
 
+-- | Check whether an expression is a (qualified) identifier.
+isQName :: C.Expr -> Maybe C.QName
+isQName m = do
+  case m of
+    C.Ident m              -> return m
+    C.RawApp _ [C.Ident m] -> return m
+    _ -> Nothing
+
+-- | Returns the contents of the given module or record.
+
+moduleContents
+  :: Rewrite
+     -- ^ How should the types be presented?
+  -> Range
+     -- ^ The range of the next argument.
+  -> String
+     -- ^ The module name.
+  -> TCM ([C.Name], [(C.Name, Type)])
+     -- ^ Module names, names paired up with corresponding types.
+
+moduleContents norm rng s = traceCall ModuleContents $ do
+  e <- parseExpr rng s
+  case isQName e of
+    -- If the expression is not a single identifier, it is not a module name
+    -- and treated as a record expression.
+    Nothing -> getRecordContents norm e
+    -- Otherwise, if it is not in scope as a module name, it is treated
+    -- as a record name.
+    Just x  -> do
+      ms :: [AbstractModule] <- scopeLookup x <$> getScope
+      if null ms then getRecordContents norm e else getModuleContents norm x
+
+-- | Returns the contents of the given record identifier.
+
+getRecordContents
+  :: Rewrite  -- ^ Amount of normalization in types.
+  -> C.Expr   -- ^ Expression presumably of record type.
+  -> TCM ([C.Name], [(C.Name, Type)])
+              -- ^ Module names, names paired up with corresponding types.
+getRecordContents norm ce = do
+  e <- toAbstract ce
+  (_, t) <- inferExpr e
+  let notRecordType = typeError $ ShouldBeRecordType t
+  (q, vs, defn) <- fromMaybeM notRecordType $ isRecordType t
+  case defn of
+    Record{ recFields = fs, recTel = tel } -> do
+      let xs   = map (nameConcrete . qnameName . unArg) fs
+          doms = telToList $ apply tel vs
+      ts <- mapM (normalForm norm) $ map (snd . unDom) doms
+      return ([], zip xs ts)
+    _ -> __IMPOSSIBLE__
+
 -- | Returns the contents of the given module.
 
-moduleContents :: Rewrite
-                  -- ^ How should the types be presented
-               -> Range
-                  -- ^ The range of the next argument.
-               -> String
-                  -- ^ The module name.
-               -> TCM ([C.Name], [(C.Name, Type)])
-                  -- ^ Module names, names paired up with
-                  -- corresponding types.
-moduleContents norm rng s = do
-  m <- parseName rng s
+getModuleContents
+  :: Rewrite  -- ^ Amount of normalization in types.
+  -> C.QName  -- ^ Module name.
+  -> TCM ([C.Name], [(C.Name, Type)])
+              -- ^ Module names, names paired up with corresponding types.
+getModuleContents norm m = do
   modScope <- getNamedScope . amodName =<< resolveModule m
   let modules :: ThingsInScope AbstractModule
       modules = exportedNamesInScope modScope
       names :: ThingsInScope AbstractName
       names = exportedNamesInScope modScope
-  types <- mapM (\(x, n) -> do
-                   d <- getConstInfo $ anameName n
-                   t <- normalForm norm =<< (defType <$> instantiateDef d)
-                   return (x, t))
-                (concatMap (\(x, ns) -> map ((,) x) ns) $
-                           Map.toList names)
+      xns = [ (x,n) | (x, ns) <- Map.toList names, n <- ns ]
+  types <- forM xns $ \(x, n) -> do
+    d <- getConstInfo $ anameName n
+    t <- normalForm norm =<< (defType <$> instantiateDef d)
+    return (x, t)
   return (Map.keys modules, types)
 
+
 whyInScope :: String -> TCM (Maybe LocalVar, [AbstractName], [AbstractModule])
 whyInScope s = do
   x     <- parseName noRange s
