@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP #-}
+{-# LANGUAGE CPP           #-}
 {-# LANGUAGE TupleSections #-}
 
 #if __GLASGOW_HASKELL__ >= 710
@@ -51,6 +51,8 @@ import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Serialise
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Primitive
+import Agda.TypeChecking.Rewriting (killCtxId)
+import Agda.TypeChecking.DeadCode
 import qualified Agda.TypeChecking.Monad.Benchmark as Bench
 
 import Agda.TheTypeChecker
@@ -72,6 +74,7 @@ import Agda.Utils.IO.Binary
 import Agda.Utils.Pretty
 import Agda.Utils.Time
 import Agda.Utils.Hash
+import qualified Agda.Utils.HashMap as HMap
 import qualified Agda.Utils.Trie as Trie
 
 #include "undefined.h"
@@ -105,23 +108,27 @@ mergeInterface i = do
             Just b1 = Map.lookup b bs
             Just b2 = Map.lookup b bi
     mapM_ check (map fst $ Map.toList $ Map.intersection bs bi)
-    addImportedThings sig bi (iHaskellImports i) (iPatternSyns i)
+    addImportedThings sig bi (iHaskellImports i) (iHaskellImportsUHC i) (iPatternSyns i) (iDisplayForms i)
     reportSLn "import.iface.merge" 20 $
       "  Rebinding primitives " ++ show prim
-    prim <- Map.fromList <$> mapM rebind prim
-    stImportedBuiltins %= (`Map.union` prim)
+    mapM_ rebind prim
     where
         rebind (x, q) = do
             PrimImpl _ pf <- lookupPrimitiveFunction x
-            return (x, Prim $ pf { primFunName = q })
+            stImportedBuiltins %= Map.insert x (Prim pf{ primFunName = q })
 
 addImportedThings ::
-  Signature -> BuiltinThings PrimFun -> Set String -> A.PatternSynDefns -> TCM ()
-addImportedThings isig ibuiltin hsImports patsyns = do
-  stImports %= \imp -> unionSignatures [imp, isig]
+  Signature -> BuiltinThings PrimFun ->
+  Set String -> -- MAlonzo imoprts
+  Set String -> -- UHC backend imports
+  A.PatternSynDefns -> DisplayForms -> TCM ()
+addImportedThings isig ibuiltin hsImports hsImportsUHC patsyns display = do
+  stImports %= \imp -> unionSignatures [imp, over sigRewriteRules killCtxId isig]
   stImportedBuiltins %= \imp -> Map.union imp ibuiltin
   stHaskellImports %= \imp -> Set.union imp hsImports
+  stHaskellImportsUHC %= \imp -> Set.union imp hsImportsUHC
   stPatternSynImports %= \imp -> Map.union imp patsyns
+  stImportedDisplayForms %= \imp -> HMap.unionWith (++) imp display
   addSignatureInstances isig
 
 -- | Scope checks the given module. A proper version of the module
@@ -373,6 +380,7 @@ getInterface' x isMain = do
                 return (False, (i, NoWarnings))
 
       typeCheckThe file = do
+          unless includeStateChanges cleanCachedLog
           let withMsgs = bracket_
                 (chaseMsg "Checking" $ Just $ filePath file)
                 (const $ chaseMsg "Finished" Nothing)
@@ -384,9 +392,10 @@ getInterface' x isMain = do
 
             -- Merge the signature with the signature for imported
             -- things.
-            sig <- getSignature
+            sig     <- getSignature
             patsyns <- getPatternSyns
-            addImportedThings sig Map.empty Set.empty patsyns
+            display <- use stImportsDisplayForms
+            addImportedThings sig Map.empty Set.empty Set.empty patsyns display
             setSignature emptySignature
             setPatternSyns Map.empty
 
@@ -402,6 +411,7 @@ getInterface' x isMain = do
             opts     <- stPersistentOptions . stPersistentState <$> get
             isig     <- getImportedSignature
             ibuiltin <- use stImportedBuiltins
+            display  <- use stImportsDisplayForms
             ipatsyns <- getPatternSynImports
             ho       <- getInteractionOutputCallback
             -- Every interface is treated in isolation. Note: Changes
@@ -425,7 +435,7 @@ getInterface' x isMain = do
                      setInteractionOutputCallback ho
                      stModuleToSource .= mf
                      setVisitedModules vs
-                     addImportedThings isig ibuiltin Set.empty ipatsyns
+                     addImportedThings isig ibuiltin Set.empty Set.empty ipatsyns display
 
                      r  <- withMsgs $ createInterface file x
                      mf <- use stModuleToSource
@@ -506,9 +516,10 @@ writeInterface file i = do
     -- i <- return $
     --   i { iInsideScope  = emptyScopeInfo
     --     }
-    i <- return $
-      i { iInsideScope  = removePrivates $ iInsideScope i
-        }
+    -- Andreas, 2016-02-02 this causes issue #1804, so don't do it:
+    -- i <- return $
+    --   i { iInsideScope  = removePrivates $ iInsideScope i
+    --     }
     encodeFile file i
     reportSLn "import.iface.write" 5 $ "Wrote interface file."
     reportSLn "import.iface.write" 50 $ "  hash = " ++ show (iFullHash i) ++ ""
@@ -548,6 +559,7 @@ createInterface file mname =
         "  visited: " ++ intercalate ", " (map prettyShow visited)
 
     previousHsImports <- getHaskellImports
+    previousHsImportsUHC <- getHaskellImportsUHC
 
     -- Parsing.
     (pragmas, top) <- Bench.billTo [Bench.Parsing] $
@@ -560,13 +572,15 @@ createInterface file mname =
         options = catMaybes $ map getOptions pragmas
     mapM_ setOptionsFromPragma options
 
+
     -- Scope checking.
     reportSLn "import.iface.create" 7 $ "Starting scope checking."
     topLevel <- Bench.billTo [Bench.Scoping] $
       concreteToAbstract_ (TopLevel file top)
     reportSLn "import.iface.create" 7 $ "Finished scope checking."
 
-    let ds = topLevelDecls topLevel
+    let ds    = topLevelDecls topLevel
+        scope = topLevelScope topLevel
 
     -- Highlighting from scope checker.
     reportSLn "import.iface.create" 7 $ "Starting highlighting from scope."
@@ -577,9 +591,23 @@ createInterface file mname =
         mapM_ (\ d -> generateAndPrintSyntaxInfo d Partial) ds
     reportSLn "import.iface.create" 7 $ "Finished highlighting from scope."
 
+
     -- Type checking.
+
+    -- invalidate cache if pragmas change, TODO move
+    cachingStarts
+    opts <- use stPragmaOptions
+    me <- readFromCachedLog
+    case me of
+      Just (Pragmas opts', _) | opts == opts'
+        -> return ()
+      _ -> do
+        reportSLn "cache" 10 $ "pragma changed: " ++ show (isJust me)
+        cleanCachedLog
+    writeToCurrentLog $ Pragmas opts
+
     reportSLn "import.iface.create" 7 $ "Starting type checking."
-    Bench.billTo [Bench.Typing] $ checkDecls ds
+    Bench.billTo [Bench.Typing] $ mapM_ checkDeclCached ds `finally_` cacheCurrentLog
     reportSLn "import.iface.create" 7 $ "Finished type checking."
 
     -- Ulf, 2013-11-09: Since we're rethrowing the error, leave it up to the
@@ -609,17 +637,25 @@ createInterface file mname =
 
       whenM (optGenerateVimFile <$> commandLineOptions) $
         -- Generate Vim file.
-        withScope_ (insideScope topLevel) $ generateVimFile $ filePath file
+        withScope_ scope $ generateVimFile $ filePath file
     reportSLn "import.iface.create" 7 $ "Finished highlighting from type info."
 
-    setScope $ outsideScope topLevel
-    reportSLn "scope.top" 50 $ "SCOPE " ++ show (insideScope topLevel)
+    setScope scope
+    reportSLn "scope.top" 50 $ "SCOPE " ++ show scope
 
     -- Serialization.
     reportSLn "import.iface.create" 7 $ "Starting serialization."
     syntaxInfo <- use stSyntaxInfo
     i <- Bench.billTo [Bench.Serialization] $ do
-      buildInterface file topLevel syntaxInfo previousHsImports options
+      buildInterface file topLevel syntaxInfo previousHsImports previousHsImportsUHC options
+
+    reportSLn "tc.top" 101 $ concat $
+      "Signature:\n" :
+      [ show x ++ "\n  type: " ++ show (defType def)
+               ++ "\n  def:  " ++ show cc ++ "\n"
+      | (x, def) <- HMap.toList $ iSignature i ^. sigDefinitions,
+        Function{ funCompiled = cc } <- [theDef def]
+      ]
     reportSLn "import.iface.create" 7 $ "Finished serialization."
 
     -- TODO: It would be nice if unsolved things were highlighted
@@ -627,6 +663,7 @@ createInterface file mname =
 
     openMetas           <- getOpenMetas
     unless (null openMetas) $ do
+      reportSLn "import.metas" 10 "We have unsolved metas."
       reportSLn "import.metas" 10 . unlines =<< showOpenMetas
     unsolvedMetas       <- List.nub <$> mapM getMetaRange openMetas
     unsolvedConstraints <- getAllConstraints
@@ -673,11 +710,13 @@ buildInterface
   -> HighlightingInfo
      -- ^ Syntax highlighting info for the module.
   -> Set String
-     -- ^ Haskell modules imported in imported modules (transitively).
+     -- ^ MAlonzo: Haskell modules imported in imported modules (transitively).
+  -> Set String
+     -- ^ UHC backend: Haskell modules imported in imported modules (transitively).
   -> [OptionsPragma]
      -- ^ Options set in @OPTIONS@ pragmas.
   -> TCM Interface
-buildInterface file topLevel syntaxInfo previousHsImports pragmas = do
+buildInterface file topLevel syntaxInfo previousHsImports previousHsImportsUHC pragmas = do
     reportSLn "import.iface" 5 "Building interface..."
     let m = topLevelModuleName topLevel
     scope'  <- getScope
@@ -687,11 +726,17 @@ buildInterface file topLevel syntaxInfo previousHsImports pragmas = do
     -- Andreas, Makoto, 2014-10-18 AIM XX: repeating the experiment
     -- with discarding also the nameBindingSite in QName:
     -- Saves 10% on serialization time (and file size)!
-    sig     <- killRange <$> getSignature
     builtin <- use stLocalBuiltins
     ms      <- getImports
     mhs     <- mapM (\ m -> (m,) <$> moduleHash m) $ Set.toList ms
     hsImps  <- getHaskellImports
+    uhcHsImps <- getHaskellImportsUHC
+    hsCode  <- use stHaskellCode
+    -- Ulf, 2016-04-12:
+    -- Non-closed display forms are not applicable outside the module anyway,
+    -- and should be dead-code eliminated (#1928).
+    display <- HMap.filter (not . null) . HMap.map (filter isClosed) <$> use stImportsDisplayForms
+    (display, sig) <- second killRange <$> (eliminateDeadCode display =<< getSignature)
     -- Andreas, 2015-02-09 kill ranges in pattern synonyms before
     -- serialization to avoid error locations pointing to external files
     -- when expanding a pattern synoym.
@@ -704,10 +749,13 @@ buildInterface file topLevel syntaxInfo previousHsImports pragmas = do
       , iImportedModules = mhs
       , iModuleName      = m
       , iScope           = empty -- publicModules scope
-      , iInsideScope     = insideScope topLevel
+      , iInsideScope     = topLevelScope topLevel
       , iSignature       = sig
+      , iDisplayForms    = display
       , iBuiltin         = builtin'
       , iHaskellImports  = hsImps `Set.difference` previousHsImports
+      , iHaskellImportsUHC = uhcHsImps `Set.difference` previousHsImportsUHC
+      , iHaskellCode     = hsCode
       , iHighlighting    = syntaxInfo
       , iPragmaOptions   = pragmas
       , iPatternSyns     = patsyns
