@@ -4,11 +4,10 @@ module Agda.Interaction.Options
     ( CommandLineOptions(..)
     , PragmaOptions(..)
     , OptionsPragma
-    , Flag
+    , Flag, OptM, runOptM
     , Verbosity
-    , IncludeDirs
     , checkOpts
-    , parseStandardOptions
+    , parseStandardOptions, parseStandardOptions'
     , parsePragmaOptions
     , parsePluginOptions
     , defaultOptions
@@ -28,13 +27,16 @@ module Agda.Interaction.Options
     , getOptSimple
     ) where
 
-import Control.Monad            ( when )
+import Control.Applicative
+import Control.Monad            ( (>=>), when )
+import Control.Monad.Trans
 
 -- base-4.7 defines the Functor instances for OptDescr and ArgDescr
 #if !(MIN_VERSION_base(4,7,0))
 import Data.Orphans             ()
 #endif
 
+import Data.Either
 import Data.Maybe
 import Data.List                ( isSuffixOf , intercalate )
 import System.Console.GetOpt    ( getOpt', usageInfo, ArgOrder(ReturnInOrder)
@@ -46,6 +48,14 @@ import Text.EditDistance
 
 import Agda.Termination.CutOff  ( CutOff(..) )
 
+import Agda.Interaction.Library
+
+import Agda.Utils.Except
+  ( ExceptT
+  , MonadError(catchError, throwError)
+  , runExceptT
+  )
+
 import Agda.Utils.TestHelpers   ( runTests )
 import Agda.Utils.QuickCheck    ( quickCheck' )
 import Agda.Utils.FileName      ( absolute, AbsolutePath, filePath )
@@ -55,8 +65,7 @@ import Agda.Utils.String        ( indent )
 import Agda.Utils.Trie          ( Trie )
 import qualified Agda.Utils.Trie as Trie
 
-import Agda.Utils.Except ( MonadError(catchError, throwError) )
-
+import Agda.Version
 -- Paths_Agda.hs is in $(BUILD_DIR)/build/autogen/.
 import Paths_Agda ( getDataFileName )
 
@@ -68,25 +77,34 @@ isLiterate file = ".lagda" `isSuffixOf` file
 
 type Verbosity = Trie String Int
 
-type IncludeDirs = Either [FilePath] [AbsolutePath]
-     -- ^ 'Left' is used temporarily, before the paths have
-     -- been made absolute. An empty 'Left' list is
-     -- interpreted as @["."]@ (see
-     -- 'Agda.TypeChecking.Monad.Options.makeIncludeDirsAbsolute').
-
 data CommandLineOptions = Options
   { optProgramName      :: String
   , optInputFile        :: Maybe FilePath
-  , optIncludeDirs      :: IncludeDirs
+  , optIncludePaths     :: [FilePath]
+  , optAbsoluteIncludePaths :: [AbsolutePath]
+  , optLibraries        :: [LibName]
+  , optOverrideLibrariesFile :: Maybe FilePath
+  -- ^ Use this (if Just) instead of .agda/libraries
+  , optDefaultLibs      :: Bool
+  -- ^ Use ~/.agda/defaults or look for .agda-lib file.
   , optShowVersion      :: Bool
   , optShowHelp         :: Bool
   , optInteractive      :: Bool
   , optRunTests         :: Bool
   , optGHCiInteraction  :: Bool
-  , optCompile          :: Bool
   , optCompileNoMain    :: Bool
+  , optGhcCompile       :: Bool
+  , optGhcCallGhc       :: Bool
+  , optGhcFlags         :: [String]
   , optEpicCompile      :: Bool
   , optJSCompile        :: Bool
+  , optUHCCompile       :: Bool
+  , optUHCBin           :: Maybe FilePath
+  , optUHCTextualCore   :: Bool
+  , optUHCCallUHC       :: Bool
+  , optUHCTraceLevel    :: Int
+  , optUHCFlags         :: [String]
+  , optOptimSmashing    :: Bool
   , optCompileDir       :: Maybe FilePath
   -- ^ In the absence of a path the project root is used.
   , optGenerateVimFile  :: Bool
@@ -98,10 +116,11 @@ data CommandLineOptions = Options
   , optCSSFile          :: Maybe FilePath
   , optIgnoreInterfaces :: Bool
   , optForcing          :: Bool
-  , optGhcFlags         :: [String]
   , optPragmaOptions    :: PragmaOptions
   , optEpicFlags        :: [String]
   , optSafe             :: Bool
+  , optSharing          :: Bool
+  , optCaching          :: Bool
   }
   deriving Show
 
@@ -128,9 +147,11 @@ data PragmaOptions = PragmaOptions
   , optWithoutK                  :: Bool
   , optCopatterns                :: Bool  -- ^ Allow definitions by copattern matching?
   , optPatternMatching           :: Bool  -- ^ Is pattern matching allowed in the current file?
+  , optExactSplit                :: Bool
+  , optEta                       :: Bool
   , optRewriting                 :: Bool  -- ^ Can rewrite rules be added and used?
   }
-  deriving Show
+  deriving (Show,Eq)
 
 -- | The options from an @OPTIONS@ pragma.
 --
@@ -154,16 +175,29 @@ defaultOptions :: CommandLineOptions
 defaultOptions = Options
   { optProgramName      = "agda"
   , optInputFile        = Nothing
-  , optIncludeDirs      = Left []
+  , optIncludePaths     = []
+  , optAbsoluteIncludePaths = []
+  , optLibraries        = []
+  , optOverrideLibrariesFile = Nothing
+  , optDefaultLibs      = True
   , optShowVersion      = False
   , optShowHelp         = False
   , optInteractive      = False
   , optRunTests         = False
   , optGHCiInteraction  = False
-  , optCompile          = False
   , optCompileNoMain    = False
+  , optGhcCompile       = False
+  , optGhcCallGhc       = True
+  , optGhcFlags         = []
   , optEpicCompile      = False
   , optJSCompile        = False
+  , optUHCCompile       = False
+  , optUHCBin           = Nothing
+  , optUHCTextualCore   = False
+  , optUHCCallUHC       = True
+  , optUHCTraceLevel    = 0
+  , optUHCFlags         = []
+  , optOptimSmashing    = True
   , optCompileDir       = Nothing
   , optGenerateVimFile  = False
   , optGenerateLaTeX    = False
@@ -174,10 +208,11 @@ defaultOptions = Options
   , optCSSFile          = Nothing
   , optIgnoreInterfaces = False
   , optForcing          = True
-  , optGhcFlags         = []
   , optPragmaOptions    = defaultPragmaOptions
   , optEpicFlags        = []
   , optSafe             = False
+  , optSharing          = False
+  , optCaching          = False
   }
 
 defaultPragmaOptions :: PragmaOptions
@@ -201,6 +236,8 @@ defaultPragmaOptions = PragmaOptions
   , optWithoutK                  = False
   , optCopatterns                = True
   , optPatternMatching           = True
+  , optExactSplit                = False
+  , optEta                       = True
   , optRewriting                 = False
   }
 
@@ -219,43 +256,49 @@ defaultLaTeXDir = "latex"
 defaultHTMLDir :: String
 defaultHTMLDir = "html"
 
-prop_defaultOptions :: Bool
-prop_defaultOptions = case checkOpts defaultOptions of
-  Left  _ -> False
-  Right _ -> True
+prop_defaultOptions :: IO Bool
+prop_defaultOptions =
+  either (const False) (const True) <$> runOptM (checkOpts defaultOptions)
+
+type OptM = ExceptT String IO
+
+runOptM :: OptM a -> IO (Either String a)
+runOptM = runExceptT
 
 {- | @f :: Flag opts@  is an action on the option record that results from
      parsing an option.  @f opts@ produces either an error message or an
      updated options record
 -}
-type Flag opts = opts -> Either String opts
+type Flag opts = opts -> OptM opts
 
 -- | Checks that the given options are consistent.
 
 checkOpts :: Flag CommandLineOptions
 checkOpts opts
-  | not (atMostOne [optAllowUnsolved . p, \x -> optCompile x]) = Left
+  | not (atMostOne [optAllowUnsolved . p, \x -> optGhcCompile x]) = throwError
       "Unsolved meta variables are not allowed when compiling.\n"
-  | optCompileNoMain opts && not (optCompile opts) = Left
+  | optCompileNoMain opts && (not (optGhcCompile opts || optUHCCompile opts)) = throwError
       "--no-main only allowed in combination with --compile.\n"
   | not (atMostOne [optGHCiInteraction, isJust . optInputFile]) =
-      Left "Choose at most one: input file or --interaction.\n"
-  | not (atMostOne $ interactive ++ [\x -> optCompile x, optEpicCompile, optJSCompile]) =
-      Left "Choose at most one: compilers/--interactive/--interaction.\n"
+      throwError "Choose at most one: input file or --interaction.\n"
+  | not (atMostOne $ interactive ++ [\x -> optGhcCompile x, optEpicCompile, optJSCompile]) =
+      throwError "Choose at most one: compilers/--interactive/--interaction.\n"
   | not (atMostOne $ interactive ++ [optGenerateHTML]) =
-      Left "Choose at most one: --html/--interactive/--interaction.\n"
+      throwError "Choose at most one: --html/--interactive/--interaction.\n"
   | not (atMostOne $ interactive ++ [isJust . optDependencyGraph]) =
-      Left "Choose at most one: --dependency-graph/--interactive/--interaction.\n"
-  | not (atMostOne [ optUniversePolymorphism . p
-                   , not . optUniverseCheck . p
-                   ]) =
-      Left "Cannot have both universe polymorphism and type in type.\n"
+      throwError "Choose at most one: --dependency-graph/--interactive/--interaction.\n"
   | not (atMostOne $ interactive ++ [optGenerateLaTeX]) =
-      Left "Choose at most one: --latex/--interactive/--interaction.\n"
+      throwError "Choose at most one: --latex/--interactive/--interaction.\n"
   | (not . null . optEpicFlags $ opts)
       && not (optEpicCompile opts) =
-      Left "Cannot set Epic flags without using the Epic backend.\n"
-  | otherwise = Right opts
+      throwError "Cannot set Epic flags without using the Epic backend.\n"
+  | (isJust $ optUHCBin opts)
+      && not (optUHCCompile opts) =
+      throwError "Cannot set uhc binary without using UHC backend.\n"
+  | (optUHCTextualCore opts)
+      && not (optUHCCompile opts) =
+      throwError "Cannot set --uhc-textual-core without using UHC backend.\n"
+  | otherwise = return opts
   where
   atMostOne bs = length (filter ($ opts) bs) <= 1
 
@@ -270,7 +313,6 @@ unsafePragmaOptions opts =
   [ "--allow-unsolved-metas"                     | optAllowUnsolved opts             ] ++
   [ "--no-positivity-check"                      | optDisablePositivity opts         ] ++
   [ "--no-termination-check"                     | not (optTerminationCheck opts)    ] ++
-  [ "--no-coverage-check"                        | not (optCompletenessCheck opts)   ] ++
   [ "--type-in-type"                             | not (optUniverseCheck opts)       ] ++
   -- [ "--sized-types"                              | optSizedTypes opts                ] ++
   [ "--injective-type-constructors"              | optInjectiveTypeConstructors opts ] ++
@@ -304,6 +346,12 @@ helpFlag o = return $ o { optShowHelp = True }
 safeFlag :: Flag CommandLineOptions
 safeFlag o = return $ o { optSafe = True }
 
+sharingFlag :: Bool -> Flag CommandLineOptions
+sharingFlag b o = return $ o { optSharing = b }
+
+cachingFlag :: Bool -> Flag CommandLineOptions
+cachingFlag b o = return $ o { optCaching = b }
+
 proofIrrelevanceFlag :: Flag PragmaOptions
 proofIrrelevanceFlag o = return $ o { optProofIrrelevance = True }
 
@@ -346,13 +394,19 @@ noPositivityFlag o = return $ o { optDisablePositivity = True }
 dontTerminationCheckFlag :: Flag PragmaOptions
 dontTerminationCheckFlag o = return $ o { optTerminationCheck = False }
 
+-- The option was removed. See Issue 1918.
 dontCompletenessCheckFlag :: Flag PragmaOptions
-dontCompletenessCheckFlag o = return $ o { optCompletenessCheck = False }
+dontCompletenessCheckFlag _ =
+  throwError "the --no-coverage-check option has been removed"
 
 dontUniverseCheckFlag :: Flag PragmaOptions
 dontUniverseCheckFlag o = return $ o { optUniverseCheck        = False
-                                     , optUniversePolymorphism = False
                                      }
+etaFlag :: Flag PragmaOptions
+etaFlag o = return $ o { optEta = True }
+
+noEtaFlag :: Flag PragmaOptions
+noEtaFlag o = return $ o { optEta = False }
 
 sizedTypes :: Flag PragmaOptions
 sizedTypes o = return $ o { optSizedTypes = True }
@@ -390,6 +444,12 @@ noCopatternsFlag o = return $ o { optCopatterns = False }
 noPatternMatchingFlag :: Flag PragmaOptions
 noPatternMatchingFlag o = return $ o { optPatternMatching = False }
 
+exactSplitFlag :: Flag PragmaOptions
+exactSplitFlag o = return $ o { optExactSplit = True }
+
+noExactSplitFlag :: Flag PragmaOptions
+noExactSplitFlag o = return $ o { optExactSplit = False }
+
 rewritingFlag :: Flag PragmaOptions
 rewritingFlag o = return $ o { optRewriting = True }
 
@@ -399,12 +459,19 @@ interactiveFlag  o = return $ o { optInteractive    = True
                                                       { optAllowUnsolved = True }
                                 }
 
-compileFlag :: Flag CommandLineOptions
-compileFlag o = return $ o { optCompile = True }
-
 compileFlagNoMain :: Flag CommandLineOptions
 compileFlagNoMain o = return $ o { optCompileNoMain = True }
 
+compileGhcFlag :: Flag CommandLineOptions
+compileGhcFlag o = return $ o { optGhcCompile = True }
+
+ghcDontCallGhcFlag :: Flag CommandLineOptions
+ghcDontCallGhcFlag o = return $ o { optGhcCallGhc = False }
+
+-- NOTE: Quadratic in number of flags.
+ghcFlag :: String -> Flag CommandLineOptions
+ghcFlag f o = return $ o { optGhcFlags = optGhcFlags o ++ [f] }
+
 -- The Epic backend has been removed. See Issue 1481.
 compileEpicFlag :: Flag CommandLineOptions
 -- compileEpicFlag o = return $ o { optEpicCompile = True}
@@ -413,19 +480,37 @@ compileEpicFlag o = throwError "the Epic backend has been disabled"
 compileJSFlag :: Flag CommandLineOptions
 compileJSFlag  o = return $ o { optJSCompile = True }
 
+compileUHCFlag :: Flag CommandLineOptions
+compileUHCFlag o = return $ o { optUHCCompile = True}
+
 compileDirFlag :: FilePath -> Flag CommandLineOptions
 compileDirFlag f o = return $ o { optCompileDir = Just f }
 
--- NOTE: Quadratic in number of flags.
-ghcFlag :: String -> Flag CommandLineOptions
-ghcFlag f o = return $ o { optGhcFlags = optGhcFlags o ++ [f] }
-
 -- NOTE: Quadratic in number of flags.
 -- The Epic backend has been removed. See Issue 1481.
 epicFlagsFlag :: String -> Flag CommandLineOptions
 -- epicFlagsFlag s o = return $ o { optEpicFlags = optEpicFlags o ++ [s] }
 epicFlagsFlag s o = throwError "the Epic backend has been disabled"
 
+uhcBinFlag :: String -> Flag CommandLineOptions
+uhcBinFlag s o = return $ o { optUHCBin  = Just s }
+
+uhcTextualCoreFlag :: Flag CommandLineOptions
+uhcTextualCoreFlag o = return $ o { optUHCTextualCore = True }
+
+uhcDontCallUHCFlag :: Flag CommandLineOptions
+uhcDontCallUHCFlag o = return $ o { optUHCCallUHC = False }
+
+uhcTraceLevelFlag :: String -> Flag CommandLineOptions
+-- TODO proper parsing and error handling
+uhcTraceLevelFlag i o = return $ o { optUHCTraceLevel = read i }
+
+uhcFlagsFlag :: String -> Flag CommandLineOptions
+uhcFlagsFlag s o = return $ o { optUHCFlags = optUHCFlags o ++ [s] }
+
+optimNoSmashing :: Flag CommandLineOptions
+optimNoSmashing o = return $ o {optOptimSmashing = False }
+
 htmlFlag :: Flag CommandLineOptions
 htmlFlag o = return $ o { optGenerateHTML = True }
 
@@ -439,8 +524,16 @@ cssFlag :: FilePath -> Flag CommandLineOptions
 cssFlag f o = return $ o { optCSSFile = Just f }
 
 includeFlag :: FilePath -> Flag CommandLineOptions
-includeFlag d o = return $ o { optIncludeDirs = Left (d : ds) }
-  where ds = either id (const []) $ optIncludeDirs o
+includeFlag d o = return $ o { optIncludePaths = d : optIncludePaths o }
+
+libraryFlag :: String -> Flag CommandLineOptions
+libraryFlag s o = return $ o { optLibraries = optLibraries o ++ [s] }
+
+overrideLibrariesFileFlag :: String -> Flag CommandLineOptions
+overrideLibrariesFileFlag s o = return $ o { optOverrideLibrariesFile = Just s }
+
+noDefaultLibsFlag :: Flag CommandLineOptions
+noDefaultLibsFlag o = return $ o { optDefaultLibs = False }
 
 verboseFlag :: String -> Flag PragmaOptions
 verboseFlag s o =
@@ -461,7 +554,7 @@ terminationDepthFlag s o =
        return $ o { optTerminationDepth = CutOff $ k-1 }
     where usage = throwError "argument to termination-depth should be >= 1"
 
-integerArgument :: String -> String -> Either String Int
+integerArgument :: String -> String -> OptM Int
 integerArgument flag s =
     readM s `catchError` \_ ->
         throwError $ "option '" ++ flag ++ "' requires an integer argument"
@@ -474,10 +567,13 @@ standardOptions =
                     "start in interactive mode"
     , Option []     ["interaction"] (NoArg ghciInteractionFlag)
                     "for use with the Emacs mode"
-    , Option ['c']  ["compile"] (NoArg compileFlag)
-                    "compile program using the MAlonzo backend (experimental)"
+    , Option ['c']  ["compile", "ghc"] (NoArg compileGhcFlag)
+                    "compile program using the GHC backend (experimental)"
+    , Option []     ["ghc-dont-call-ghc"] (NoArg ghcDontCallGhcFlag) "Don't call ghc, just write the GHC Haskell files."
+    , Option []     ["ghc-flag"] (ReqArg ghcFlag "GHC-FLAG")
+                    "give the flag GHC-FLAG to GHC when compiling using the GHC backend"
     , Option []     ["no-main"] (NoArg compileFlagNoMain)
-                    "when compiling using the MAlonzo backend (experimental), do not treat the requested module as the main module of a program"
+                    "when compiling using the GHC backend or the UHC backend (experimental), do not treat the requested module as the main module of a program"
 
     -- The Epic backend has been removed. See Issue 1481.
     , Option []     ["epic"] (NoArg compileEpicFlag)
@@ -485,10 +581,16 @@ standardOptions =
                     "the Epic backend has been removed"
 
     , Option []     ["js"] (NoArg compileJSFlag) "compile program using the JS backend"
+    , Option []     ["uhc"] (NoArg compileUHCFlag) "compile program using the UHC backend"
+    , Option []     ["uhc-bin"] (ReqArg uhcBinFlag "UHC") "The uhc binary to use when compiling with the UHC backend."
+    , Option []     ["uhc-textual-core"] (NoArg uhcTextualCoreFlag) "Use textual core as intermediate representation instead of binary core."
+    , Option []     ["uhc-dont-call-uhc"] (NoArg uhcDontCallUHCFlag) "Don't call uhc, just write the UHC Core files."
+    , Option []     ["uhc-gen-trace"] (ReqArg uhcTraceLevelFlag "TRACE") "Add tracing code to generated executable."
+    , Option []     ["uhc-flag"] (ReqArg uhcFlagsFlag "UHC-FLAG")
+                    "give the flag UHC-FLAG to UHC when compiling using the UHC backend"
+    , Option []     ["no-smashing"] (NoArg optimNoSmashing) "Don't apply the smashing optimization."
     , Option []     ["compile-dir"] (ReqArg compileDirFlag "DIR")
                     ("directory for compiler output (default: the project root)")
-    , Option []     ["ghc-flag"] (ReqArg ghcFlag "GHC-FLAG")
-                    "give the flag GHC-FLAG to GHC when compiling using MAlonzo"
 
     -- The Epic backend has been removed. See Issue 1481.
     , Option []     ["epic-flag"] (ReqArg epicFlagsFlag "EPIC-FLAG")
@@ -517,10 +619,24 @@ standardOptions =
                     "ignore interface files (re-type check everything)"
     , Option ['i']  ["include-path"] (ReqArg includeFlag "DIR")
                     "look for imports in DIR"
+    , Option ['l']  ["library"] (ReqArg libraryFlag "LIB")
+                    "use library LIB"
+    , Option []     ["library-file"] (ReqArg overrideLibrariesFileFlag "FILE")
+                    "use FILE instead of the standard libraries file"
+    , Option []     ["no-default-libraries"] (NoArg noDefaultLibsFlag)
+                    "don't use default libraries"
     , Option []     ["no-forcing"] (NoArg noForcingFlag)
                     "disable the forcing optimisation"
     , Option []     ["safe"] (NoArg safeFlag)
                     "disable postulates, unsafe OPTION pragmas and primTrustMe"
+    , Option []     ["sharing"] (NoArg $ sharingFlag True)
+                    "enable sharing and call-by-need evaluation (experimental) (default: OFF)"
+    , Option []     ["no-sharing"] (NoArg $ sharingFlag False)
+                    "disable sharing and call-by-need evaluation"
+    , Option []     ["caching"] (NoArg $ cachingFlag True)
+                    "enable caching of typechecking (experimental) (default: OFF)"
+    , Option []     ["no-caching"] (NoArg $ cachingFlag False)
+                    "disable caching of typechecking"
     ] ++ map (fmap lift) pragmaOptions
   where
   lift :: Flag PragmaOptions -> Flag CommandLineOptions
@@ -547,7 +663,7 @@ pragmaOptions =
     , Option []     ["termination-depth"] (ReqArg terminationDepthFlag "N")
                     "allow termination checker to count decrease/increase upto N (default N=1)"
     , Option []     ["no-coverage-check"] (NoArg dontCompletenessCheckFlag)
-                    "do not warn about possibly incomplete pattern matches"
+                    "the option has been removed"
     , Option []     ["type-in-type"] (NoArg dontUniverseCheckFlag)
                     "ignore universe levels (this makes Agda inconsistent)"
     , Option []     ["sized-types"] (NoArg sizedTypes)
@@ -576,6 +692,12 @@ pragmaOptions =
                     "disable definitions by copattern matching"
     , Option []     ["no-pattern-matching"] (NoArg noPatternMatchingFlag)
                     "disable pattern matching completely"
+    , Option []     ["exact-split"] (NoArg exactSplitFlag)
+                    "require all clauses in a definition by pattern matching to hold as definitional equalities (except those marked as CATCHALL)"
+    , Option []     ["no-exact-split"] (NoArg noExactSplitFlag)
+                    "do not require all clauses in a definition by pattern matching to hold as definitional equalities (ignore those marked as CATCHALL)"
+    , Option []     ["no-eta-equality"] (NoArg noEtaFlag)
+                    "disable eta rules for records"
     , Option []     ["rewriting"] (NoArg rewritingFlag)
                     "enable declaration and use of REWRITE rules"
     ]
@@ -633,10 +755,13 @@ getOptSimple argv opts fileArg = \ defaults ->
       sugs as  = "any of " ++ intercalate " " as
 
 -- | Parse the standard options.
-parseStandardOptions :: [String] -> Either String CommandLineOptions
-parseStandardOptions argv =
-  checkOpts =<<
-    getOptSimple argv standardOptions inputFlag defaultOptions
+parseStandardOptions :: [String] -> OptM CommandLineOptions
+parseStandardOptions argv = parseStandardOptions' argv defaultOptions
+
+parseStandardOptions' :: [String] -> Flag CommandLineOptions
+parseStandardOptions' argv opts = do
+  opts <- getOptSimple (stripRTS argv) standardOptions inputFlag opts
+  checkOpts opts
 
 -- | Parse options from an options pragma.
 parsePragmaOptions
@@ -644,7 +769,7 @@ parsePragmaOptions
      -- ^ Pragma options.
   -> CommandLineOptions
      -- ^ Command-line options which should be updated.
-  -> Either String PragmaOptions
+  -> OptM PragmaOptions
 parsePragmaOptions argv opts = do
   ps <- getOptSimple argv pragmaOptions
           (\s _ -> throwError $ "Bad option in pragma: " ++ s)
@@ -661,29 +786,21 @@ parsePluginOptions argv opts =
 
 -- | The usage info message. The argument is the program name (probably
 --   agda).
-usage :: [OptDescr ()] -> [(String, String, [String], [OptDescr ()])] -> String -> String
-usage options pluginInfos progName =
-        usageInfo (header progName) options ++
-        "\nPlugins:\n" ++
-        indent 2 (concatMap pluginMsg pluginInfos)
-
+usage :: [OptDescr ()] -> String -> String
+usage options progName = usageInfo (header progName) options
     where
-        header progName = unlines [ "Agda"
-                                  , ""
-                                  , "Usage: " ++ progName ++ " [OPTIONS...] [FILE]"
-                                  ]
+        header progName = unlines [ "Agda version " ++ version, ""
+                                  , "Usage: " ++ progName ++ " [OPTIONS...] [FILE]" ]
 
-        pluginMsg (name, help, inherited, opts)
-            | null opts && null inherited = optHeader
-            | otherwise = usageInfo (optHeader ++
-                                     "  Plugin-specific options:" ++
-                                     inheritedOptions inherited
-                                     ) opts
-            where
-                optHeader = "\n" ++ name ++ "-plugin:\n" ++ indent 2 help
-                inheritedOptions [] = ""
-                inheritedOptions pls =
-                    "\n    Inherits options from: " ++ unwords pls
+-- Remove +RTS .. -RTS from arguments
+stripRTS :: [String] -> [String]
+stripRTS [] = []
+stripRTS ("--RTS" : argv) = argv
+stripRTS (arg : argv)
+  | is "+RTS" arg = stripRTS $ drop 1 $ dropWhile (not . is "-RTS") argv
+  | otherwise     = arg : stripRTS argv
+  where
+    is x arg = [x] == take 1 (words arg)
 
 ------------------------------------------------------------------------
 -- Some paths
@@ -702,6 +819,6 @@ defaultLibDir = do
 
 tests :: IO Bool
 tests = runTests "Agda.Interaction.Options"
-  [ quickCheck' prop_defaultOptions
+  [ prop_defaultOptions
   , defaultPragmaOptionsSafe
   ]
