@@ -1,10 +1,10 @@
-{-# LANGUAGE CPP #-}
-{-# LANGUAGE DeriveDataTypeable #-}
-{-# LANGUAGE DeriveFoldable #-}
-{-# LANGUAGE DeriveFunctor #-}
-{-# LANGUAGE DeriveGeneric #-}
-{-# LANGUAGE DeriveTraversable #-}
-{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE CPP                        #-}
+{-# LANGUAGE DeriveDataTypeable         #-}
+{-# LANGUAGE DeriveFoldable             #-}
+{-# LANGUAGE DeriveFunctor              #-}
+{-# LANGUAGE DeriveGeneric              #-}
+{-# LANGUAGE DeriveTraversable          #-}
+{-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 
 {-| Some common syntactic entities are defined in this module.
@@ -12,6 +12,7 @@
 module Agda.Syntax.Common where
 
 import Control.Applicative
+import Control.DeepSeq
 
 import Data.ByteString.Char8 (ByteString)
 import qualified Data.ByteString.Char8 as ByteString
@@ -68,6 +69,10 @@ instance CoArbitrary Induction where
   coarbitrary Inductive   = variant 0
   coarbitrary CoInductive = variant 1
 
+instance NFData Induction where
+  rnf Inductive   = ()
+  rnf CoInductive = ()
+
 ---------------------------------------------------------------------------
 -- * Hiding
 ---------------------------------------------------------------------------
@@ -88,8 +93,13 @@ instance Monoid Hiding where
 instance KillRange Hiding where
   killRange = id
 
+instance NFData Hiding where
+  rnf Hidden    = ()
+  rnf Instance  = ()
+  rnf NotHidden = ()
+
 -- | Decorating something with 'Hiding' information.
-data WithHiding a = WithHiding Hiding a
+data WithHiding a = WithHiding !Hiding a
   deriving (Typeable, Eq, Ord, Show, Functor, Foldable, Traversable)
 
 instance Decoration WithHiding where
@@ -108,6 +118,9 @@ instance SetRange a => SetRange (WithHiding a) where
 instance KillRange a => KillRange (WithHiding a) where
   killRange = fmap killRange
 
+instance NFData a => NFData (WithHiding a) where
+  rnf (WithHiding _ a) = rnf a
+
 -- | A lens to access the 'Hiding' attribute in data structures.
 --   Minimal implementation: @getHiding@ and one of @setHiding@ or @mapHiding@.
 class LensHiding a where
@@ -153,6 +166,13 @@ notVisible a = getHiding a /= NotHidden
 hide :: LensHiding a => a -> a
 hide = setHiding Hidden
 
+hideOrKeepInstance :: LensHiding a => a -> a
+hideOrKeepInstance x =
+  case getHiding x of
+    Hidden -> x
+    Instance -> x
+    NotHidden -> setHiding Hidden x
+
 makeInstance :: LensHiding a => a -> a
 makeInstance = setHiding Instance
 
@@ -176,6 +196,10 @@ instance Ord Big where
   Big <= Small = False
   _   <= _     = True
 
+instance NFData Big where
+  rnf Big   = ()
+  rnf Small = ()
+
 -- | A function argument can be relevant or irrelevant.
 --   See "Agda.TypeChecking.Irrelevance".
 data Relevance
@@ -213,6 +237,13 @@ instance Arbitrary Relevance where
 instance Ord Relevance where
   (<=) = moreRelevant
 
+instance NFData Relevance where
+  rnf Relevant   = ()
+  rnf NonStrict  = ()
+  rnf Irrelevant = ()
+  rnf (Forced a) = rnf a
+  rnf UnusedArg  = ()
+
 -- | A lens to access the 'Relevance' attribute in data structures.
 --   Minimal implementation: @getRelevance@ and one of @setRelevance@ or @mapRelevance@.
 class LensRelevance a where
@@ -331,90 +362,70 @@ nonStrictToIrr rel       = rel
 
 -- | A function argument can be hidden and/or irrelevant.
 
-data ArgInfo c = ArgInfo
+data ArgInfo = ArgInfo
   { argInfoHiding    :: Hiding
   , argInfoRelevance :: Relevance
-  , argInfoColors    :: [c]
-  } deriving (Typeable, Eq, Ord, Functor, Foldable, Traversable, Show)
+  } deriving (Typeable, Eq, Ord, Show)
 
-instance KillRange c => KillRange (ArgInfo c) where
-  killRange (ArgInfo h r cs) = killRange3 ArgInfo h r cs
+instance KillRange ArgInfo where
+  killRange (ArgInfo h r) = killRange2 ArgInfo h r
 
-{- FAILED to define a lens for ArgInfo, since it is parametrized by c
-
-   can't instantiate the following to f c = Arg c e
-   since Haskell does not have lambda abstraction
-
-class LensArgInfo f where
-  getArgInfo :: f c -> ArgInfo c
-  setArgInfo :: ArgInfo c' -> f c -> f c'
+class LensArgInfo a where
+  getArgInfo :: a -> ArgInfo
+  setArgInfo :: ArgInfo -> a -> a
   setArgInfo ai = mapArgInfo (const ai)
-  mapArgInfo :: (ArgInfo c -> ArgInfo c') -> f c -> f c'
+  mapArgInfo :: (ArgInfo -> ArgInfo) -> a -> a
   mapArgInfo f a = setArgInfo (f $ getArgInfo a) a
 
 instance LensArgInfo ArgInfo where
   getArgInfo = id
   setArgInfo = const
   mapArgInfo = id
--}
-{- FAILS because map is too restricted
-class LensArgInfo c a where
-  getArgInfo :: a -> ArgInfo c
-  setArgInfo :: ArgInfo c -> a -> a
-  setArgInfo ai = mapArgInfo (const ai)
-  mapArgInfo :: (ArgInfo c -> ArgInfo c) -> a -> a
-  mapArgInfo f a = setArgInfo (f $ getArgInfo a) a
 
-instance LensArgInfo c (ArgInfo c) where
-  getArgInfo = id
-  setArgInfo = const
-  mapArgInfo = id
--}
+instance NFData ArgInfo where
+  rnf (ArgInfo a b) = rnf a `seq` rnf b
 
-instance LensHiding (ArgInfo c) where
+instance LensHiding ArgInfo where
   getHiding = argInfoHiding
   setHiding h ai = ai { argInfoHiding = h }
   mapHiding f ai = ai { argInfoHiding = f (argInfoHiding ai) }
 
-instance LensRelevance (ArgInfo c) where
+instance LensRelevance ArgInfo where
   getRelevance = argInfoRelevance
   setRelevance h ai = ai { argInfoRelevance = h }
   mapRelevance f ai = ai { argInfoRelevance = f (argInfoRelevance ai) }
 
-mapArgInfoColors :: ([c] -> [c']) -> ArgInfo c -> ArgInfo c'
-mapArgInfoColors f info = info { argInfoColors = f $ argInfoColors info }
-
-defaultArgInfo :: ArgInfo c
+defaultArgInfo :: ArgInfo
 defaultArgInfo =  ArgInfo { argInfoHiding    = NotHidden
-                          , argInfoRelevance = Relevant
-                          , argInfoColors    = [] }
+                          , argInfoRelevance = Relevant }
+
 
 ---------------------------------------------------------------------------
 -- * Arguments
 ---------------------------------------------------------------------------
 
-data Arg c e  = Arg
-  { argInfo :: ArgInfo c
+data Arg e  = Arg
+  { argInfo :: ArgInfo
   , unArg :: e
   } deriving (Typeable, Ord, Functor, Foldable, Traversable)
 
-instance Decoration (Arg c) where
+instance Decoration Arg where
   traverseF f (Arg ai a) = Arg ai <$> f a
 
-instance HasRange a => HasRange (Arg c a) where
+instance HasRange a => HasRange (Arg a) where
     getRange = getRange . unArg
 
-instance SetRange a => SetRange (Arg c a) where
+instance SetRange a => SetRange (Arg a) where
   setRange r = fmap $ setRange r
 
-instance (KillRange c, KillRange a) => KillRange (Arg c a) where
+instance KillRange a => KillRange (Arg a) where
   killRange (Arg info a) = killRange2 Arg info a
 
-instance (Eq a, Eq c) => Eq (Arg c a) where
-  Arg (ArgInfo h1 _ cs1) x1 == Arg (ArgInfo h2 _ cs2) x2 = (h1, cs1, x1) == (h2, cs2, x2)
+instance Eq a => Eq (Arg a) where
+  Arg (ArgInfo h1 _) x1 == Arg (ArgInfo h2 _) x2 = (h1, x1) == (h2, x2)
 
-instance (Show a, Show c) => Show (Arg c a) where
-    show (Arg (ArgInfo h r cs) x) = showC cs $ showR r $ showH h $ show x
+instance Show a => Show (Arg a) where
+    show (Arg (ArgInfo h r) x) = showR r $ showH h $ show x
       where
         showH Hidden     s = "{" ++ s ++ "}"
         showH NotHidden  s = "(" ++ s ++ ")"
@@ -426,61 +437,46 @@ instance (Show a, Show c) => Show (Arg c a) where
           Forced Small -> "!" ++ s
           UnusedArg    -> "k" ++ s -- constant
           Relevant     -> "r" ++ s -- Andreas: I want to see it explicitly
-        showC cs         s = show cs ++ s
 
-instance LensHiding (Arg c e) where
+instance NFData e => NFData (Arg e) where
+  rnf (Arg a b) = rnf a `seq` rnf b
+
+instance LensHiding (Arg e) where
   getHiding = getHiding . argInfo
   mapHiding = mapArgInfo . mapHiding
 
-instance LensRelevance (Arg c e) where
+instance LensRelevance (Arg e) where
   getRelevance = getRelevance . argInfo
   mapRelevance = mapArgInfo . mapRelevance
 
 {- RETIRED
-hide :: Arg c a -> Arg c a
+hide :: Arg a -> Arg a
 hide = setArgHiding Hidden
 
-makeInstance :: Arg c a -> Arg c a
+makeInstance :: Arg a -> Arg a
 makeInstance = setHiding Instance
 
-isHiddenArg :: Arg c a -> Bool
+isHiddenArg :: Arg a -> Bool
 isHiddenArg arg = argHiding arg /= NotHidden
 -}
 
-mapArgInfo :: (ArgInfo c -> ArgInfo c') -> Arg c a -> Arg c' a
-mapArgInfo f arg = arg { argInfo = f $ argInfo arg }
+instance LensArgInfo (Arg a) where
+  getArgInfo        = argInfo
+  mapArgInfo f arg  = arg { argInfo = f $ argInfo arg }
 
-argColors :: Arg c a -> [c]
-argColors = argInfoColors . argInfo
-
-mapArgColors :: ([c] -> [c']) -> Arg c a -> Arg c' a
-mapArgColors = mapArgInfo . mapArgInfoColors
-
-setArgColors :: [c] -> Arg c' a -> Arg c a
-setArgColors = mapArgColors . const
-
-defaultArg :: a -> Arg c a
+defaultArg :: a -> Arg a
 defaultArg = Arg defaultArgInfo
 
-defaultColoredArg :: ([c],a) -> Arg c a
-defaultColoredArg (cs,a) = setArgColors cs $ defaultArg a
-
-noColorArg :: Hiding -> Relevance -> a -> Arg c a
-noColorArg h r = Arg ArgInfo { argInfoHiding    = h
-                             , argInfoRelevance = r
-                             , argInfoColors    = []
-                             }
-
 -- | @xs \`withArgsFrom\` args@ translates @xs@ into a list of 'Arg's,
 -- using the elements in @args@ to fill in the non-'unArg' fields.
 --
 -- Precondition: The two lists should have equal length.
 
-withArgsFrom :: [a] -> [Arg c b] -> [Arg c a]
+withArgsFrom :: [a] -> [Arg b] -> [Arg a]
 xs `withArgsFrom` args =
   zipWith (\x arg -> fmap (const x) arg) xs args
 
-withNamedArgsFrom :: [a] -> [NamedArg c b] -> [NamedArg c a]
+withNamedArgsFrom :: [a] -> [NamedArg b] -> [NamedArg a]
 xs `withNamedArgsFrom` args =
   zipWith (\x -> fmap (x <$)) xs args
 
@@ -514,44 +510,42 @@ instance Underscore Doc where
 --   @Dom@ is used in 'Pi' of internal syntax, in 'Context' and 'Telescope'.
 --   'Arg' is used for actual arguments ('Var', 'Con', 'Def' etc.)
 --   and in 'Abstract' syntax and other situations.
-data Dom c e = Dom
-  { domInfo   :: ArgInfo c
+data Dom e = Dom
+  { domInfo   :: ArgInfo
   , unDom     :: e
   } deriving (Typeable, Eq, Ord, Functor, Foldable, Traversable)
 
-instance Decoration (Dom c) where
+instance Decoration Dom where
   traverseF f (Dom ai a) = Dom ai <$> f a
 
-instance HasRange a => HasRange (Dom c a) where
+instance HasRange a => HasRange (Dom a) where
   getRange = getRange . unDom
 
-instance (KillRange c, KillRange a) => KillRange (Dom c a) where
+instance KillRange a => KillRange (Dom a) where
   killRange (Dom info a) = killRange2 Dom info a
 
-instance (Show a, Show c) => Show (Dom c a) where
+instance Show a => Show (Dom a) where
   show = show . argFromDom
 
-instance LensHiding (Dom c e) where
+instance LensHiding (Dom e) where
   getHiding = getHiding . domInfo
-  mapHiding = mapDomInfo . mapHiding
+  mapHiding = mapArgInfo . mapHiding
 
-instance LensRelevance (Dom c e) where
+instance LensRelevance (Dom e) where
   getRelevance = getRelevance . domInfo
-  mapRelevance = mapDomInfo . mapRelevance
+  mapRelevance = mapArgInfo . mapRelevance
 
-mapDomInfo :: (ArgInfo c -> ArgInfo c') -> Dom c a -> Dom c' a
-mapDomInfo f arg = arg { domInfo = f $ domInfo arg }
+instance LensArgInfo (Dom e) where
+  getArgInfo = domInfo
+  mapArgInfo f arg = arg { domInfo = f $ domInfo arg }
 
-domColors :: Dom c a -> [c]
-domColors = argInfoColors . domInfo
-
-argFromDom :: Dom c a -> Arg c a
+argFromDom :: Dom a -> Arg a
 argFromDom (Dom i a) = Arg i a
 
-domFromArg :: Arg c a -> Dom c a
+domFromArg :: Arg a -> Dom a
 domFromArg (Arg i a) = Dom i a
 
-defaultDom :: a -> Dom c a
+defaultDom :: a -> Dom a
 defaultDom = Dom defaultArgInfo
 
 ---------------------------------------------------------------------------
@@ -590,19 +584,22 @@ instance Show a => Show (Named_ a) where
     show (Named Nothing x)  = show x
     show (Named (Just n) x) = rawNameToString (rangedThing n) ++ " = " ++ show x
 
+instance (NFData name, NFData a) => NFData (Named name a) where
+  rnf (Named a b) = rnf a `seq` rnf b
+
 -- | Only 'Hidden' arguments can have names.
-type NamedArg c a = Arg c (Named_ a)
+type NamedArg a = Arg (Named_ a)
 
 -- | Get the content of a 'NamedArg'.
-namedArg :: NamedArg c a -> a
+namedArg :: NamedArg a -> a
 namedArg = namedThing . unArg
 
-defaultNamedArg :: a -> NamedArg c a
+defaultNamedArg :: a -> NamedArg a
 defaultNamedArg = defaultArg . unnamed
 
 -- | The functor instance for 'NamedArg' would be ambiguous,
 --   so we give it another name here.
-updateNamedArg :: (a -> b) -> NamedArg c a -> NamedArg c b
+updateNamedArg :: (a -> b) -> NamedArg a -> NamedArg b
 updateNamedArg = fmap . fmap
 
 ---------------------------------------------------------------------------
@@ -638,6 +635,11 @@ instance KillRange (Ranged a) where
 instance Decoration Ranged where
   traverseF f (Ranged r x) = Ranged r <$> f x
 
+-- | Ranges are not forced.
+
+instance NFData a => NFData (Ranged a) where
+  rnf (Ranged _ a) = rnf a
+
 ---------------------------------------------------------------------------
 -- * Raw names (before parsing into name parts).
 ---------------------------------------------------------------------------
@@ -691,12 +693,33 @@ instance KillRange IsAbstract where
 data IsInstance = InstanceDef | NotInstanceDef
     deriving (Typeable, Show, Eq, Ord)
 
+instance KillRange IsInstance where
+  killRange = id
+
+instance HasRange IsInstance where
+  getRange _ = noRange
+
+instance NFData IsInstance where
+  rnf InstanceDef    = ()
+  rnf NotInstanceDef = ()
+
+-- | Is this a macro definition?
+data IsMacro = MacroDef | NotMacroDef
+  deriving (Typeable, Show, Eq, Ord)
+
+instance KillRange IsMacro where killRange = id
+instance HasRange  IsMacro where getRange _ = noRange
+
 type Nat    = Int
 type Arity  = Nat
 
+---------------------------------------------------------------------------
+-- * NameId
+---------------------------------------------------------------------------
+
 -- | The unique identifier of a name. Second argument is the top-level module
 --   identifier.
-data NameId = NameId Integer Integer
+data NameId = NameId !Integer !Integer
     deriving (Eq, Ord, Typeable, Generic)
 
 instance KillRange NameId where
@@ -711,6 +734,9 @@ instance Enum NameId where
   toEnum n              = __IMPOSSIBLE__  -- should not be used
   fromEnum (NameId n _) = fromIntegral n
 
+instance NFData NameId where
+  rnf (NameId _ _) = ()
+
 instance Hashable NameId where
   {-# INLINE hashWithSalt #-}
   hashWithSalt salt (NameId n m) = hashWithSalt salt (n, m)
@@ -720,6 +746,10 @@ instance Arbitrary NameId where
 
 instance CoArbitrary NameId
 
+---------------------------------------------------------------------------
+-- * Meta variables
+---------------------------------------------------------------------------
+
 -- | A meta variable identifier is just a natural number.
 --
 newtype MetaId = MetaId { metaId :: Nat }
@@ -733,8 +763,38 @@ instance Show MetaId where
   showsPrec p (MetaId n) = showParen (p > 0) $
     showString "MetaId " . shows n
 
+instance NFData MetaId where
+  rnf (MetaId x) = rnf x
+
 newtype Constr a = Constr a
 
+data Placeholder
+  = Beginning
+    -- ^ @_foo@.
+  | Middle
+    -- ^ @foo_bar@.
+  | End
+    -- ^ @foo_@.
+  deriving (Show, Eq, Ord)
+
+-- | Placeholders are used to represent the underscores in a section.
+data MaybePlaceholder e
+  = Placeholder !Placeholder
+  | NoPlaceholder e
+  deriving (Typeable, Eq, Ord, Functor, Foldable, Traversable, Show)
+
+instance HasRange a => HasRange (MaybePlaceholder a) where
+  getRange Placeholder{}     = noRange
+  getRange (NoPlaceholder e) = getRange e
+
+instance KillRange a => KillRange (MaybePlaceholder a) where
+  killRange p@Placeholder{}   = p
+  killRange (NoPlaceholder e) = killRange1 NoPlaceholder e
+
+instance NFData a => NFData (MaybePlaceholder a) where
+  rnf (Placeholder _)   = ()
+  rnf (NoPlaceholder a) = rnf a
+
 ---------------------------------------------------------------------------
 -- * Interaction meta variables
 ---------------------------------------------------------------------------
@@ -747,11 +807,120 @@ instance Show InteractionId where
 
 instance KillRange InteractionId where killRange = id
 
+-----------------------------------------------------------------------------
+-- * Import directive
+-----------------------------------------------------------------------------
+
+-- | The things you are allowed to say when you shuffle names between name
+--   spaces (i.e. in @import@, @namespace@, or @open@ declarations).
+data ImportDirective' a b = ImportDirective
+  { importDirRange :: Range
+  , using          :: Using' a b
+  , hiding         :: [ImportedName' a b]
+  , impRenaming    :: [Renaming' a b]
+  , publicOpen     :: Bool -- ^ Only for @open@. Exports the opened names from the current module.
+  }
+  deriving (Typeable, Eq)
+
+data Using' a b = UseEverything | Using [ImportedName' a b]
+  deriving (Typeable, Eq)
+
+instance Monoid (Using' a b) where
+  mempty = UseEverything
+  mappend UseEverything u = u
+  mappend u UseEverything = u
+  mappend (Using xs) (Using ys) = Using (xs ++ ys)
+
+-- | Default is directive is @private@ (use everything, but do not export).
+defaultImportDir :: ImportDirective' a b
+defaultImportDir = ImportDirective noRange UseEverything [] [] False
+
+isDefaultImportDir :: ImportDirective' a b -> Bool
+isDefaultImportDir (ImportDirective _ UseEverything [] [] False) = True
+isDefaultImportDir _                                             = False
+
+-- | An imported name can be a module or a defined name
+data ImportedName' a b
+  = ImportedModule  b
+  | ImportedName    a
+  deriving (Typeable, Eq, Ord)
+
+setImportedName :: ImportedName' a a -> a -> ImportedName' a a
+setImportedName (ImportedName   x) y = ImportedName   y
+setImportedName (ImportedModule x) y = ImportedModule y
+
+instance (Show a, Show b) => Show (ImportedName' a b) where
+  show (ImportedModule x) = "module " ++ show x
+  show (ImportedName   x) = show x
+
+data Renaming' a b = Renaming
+  { renFrom    :: ImportedName' a b
+    -- ^ Rename from this name.
+  , renTo      :: ImportedName' a b
+    -- ^ To this one.  Must be same kind as 'renFrom'.
+  , renToRange :: Range
+    -- ^ The range of the \"to\" keyword.  Retained for highlighting purposes.
+  }
+  deriving (Typeable, Eq)
+
+-- ** HasRange instances
+
+instance (HasRange a, HasRange b) => HasRange (ImportDirective' a b) where
+  getRange = importDirRange
+
+instance (HasRange a, HasRange b) => HasRange (Using' a b) where
+  getRange (Using  xs) = getRange xs
+  getRange UseEverything = noRange
+
+instance (HasRange a, HasRange b) => HasRange (Renaming' a b) where
+  getRange r = getRange (renFrom r, renTo r)
+
+instance (HasRange a, HasRange b) => HasRange (ImportedName' a b) where
+  getRange (ImportedName   x) = getRange x
+  getRange (ImportedModule x) = getRange x
+
+-- ** KillRange instances
+
+instance (KillRange a, KillRange b) => KillRange (ImportDirective' a b) where
+  killRange (ImportDirective _ u h r p) =
+    killRange3 (\u h r -> ImportDirective noRange u h r p) u h r
+
+instance (KillRange a, KillRange b) => KillRange (Using' a b) where
+  killRange (Using  i) = killRange1 Using  i
+  killRange UseEverything = UseEverything
+
+instance (KillRange a, KillRange b) => KillRange (Renaming' a b) where
+  killRange (Renaming i n _) = killRange2 (\i n -> Renaming i n noRange) i n
+
+instance (KillRange a, KillRange b) => KillRange (ImportedName' a b) where
+  killRange (ImportedModule n) = killRange1 ImportedModule n
+  killRange (ImportedName   n) = killRange1 ImportedName   n
+
+-- ** NFData instances
+
+-- | Ranges are not forced.
+
+instance (NFData a, NFData b) => NFData (ImportDirective' a b) where
+  rnf (ImportDirective _ a b c _) = rnf a `seq` rnf b `seq` rnf c
+
+instance (NFData a, NFData b) => NFData (Using' a b) where
+  rnf UseEverything = ()
+  rnf (Using a)     = rnf a
+
+-- | Ranges are not forced.
+
+instance (NFData a, NFData b) => NFData (Renaming' a b) where
+  rnf (Renaming a b _) = rnf a `seq` rnf b
+
+instance (NFData a, NFData b) => NFData (ImportedName' a b) where
+  rnf (ImportedModule a) = rnf a
+  rnf (ImportedName a)   = rnf a
+
 -----------------------------------------------------------------------------
 -- * Termination
 -----------------------------------------------------------------------------
 
--- | Termination check? (Default = True).
+-- | Termination check? (Default = TerminationCheck).
 data TerminationCheck m
   = TerminationCheck
     -- ^ Run the termination checker.
@@ -761,10 +930,24 @@ data TerminationCheck m
     -- ^ Treat as non-terminating.
   | Terminating
     -- ^ Treat as terminating (unsafe).  Same effect as 'NoTerminationCheck'.
-  | TerminationMeasure !Range m
+  | TerminationMeasure Range m
     -- ^ Skip termination checking but use measure instead.
     deriving (Typeable, Show, Eq, Functor)
 
 instance KillRange m => KillRange (TerminationCheck m) where
   killRange (TerminationMeasure _ m) = TerminationMeasure noRange (killRange m)
   killRange t                        = t
+
+instance NFData a => NFData (TerminationCheck a) where
+  rnf TerminationCheck         = ()
+  rnf NoTerminationCheck       = ()
+  rnf NonTerminating           = ()
+  rnf Terminating              = ()
+  rnf (TerminationMeasure _ a) = rnf a
+
+-----------------------------------------------------------------------------
+-- * Positivity
+-----------------------------------------------------------------------------
+
+-- | Positivity check? (Default = True).
+type PositivityCheck = Bool
