@@ -14,7 +14,6 @@ import Control.DeepSeq
 import Control.Applicative
 
 import Data.ByteString.Char8 (ByteString)
-import qualified Data.ByteString.Char8 as ByteString
 import Data.List
 import Data.Typeable (Typeable)
 
@@ -41,12 +40,10 @@ import Agda.Utils.Impossible
     in different locations are equal.
 -}
 data Name
-  = Name !Range [NamePart]  -- ^ A (mixfix) identifier.
-  | NoName !Range NameId    -- ^ @_@.
+  = Name Range [NamePart]  -- ^ A (mixfix) identifier.
+  | NoName Range NameId    -- ^ @_@.
   deriving (Typeable)
 
-instance NFData Name where rnf x = seq x ()
-
 instance Underscore Name where
   underscore = NoName noRange __IMPOSSIBLE__
   isUnderscore NoName{}        = True
@@ -125,17 +122,36 @@ nameToRawName = prettyShow
 
 nameParts :: Name -> [NamePart]
 nameParts (Name _ ps)  = ps
-nameParts (NoName _ _) = [Hole]
+nameParts (NoName _ _) = [Id "_"] -- To not return an empty list
 
 nameStringParts :: Name -> [RawName]
 nameStringParts n = [ s | Id s <- nameParts n ]
 
 -- | Parse a string to parts of a concrete name.
+--
+--   Note: @stringNameParts "_" == [Id "_"] == nameParts NoName{}@
 
 stringNameParts :: String -> [NamePart]
-stringNameParts ""                              = []
-stringNameParts ('_':s)                         = Hole : stringNameParts s
-stringNameParts s | (x, s') <- break (== '_') s = Id (stringToRawName x) : stringNameParts s'
+stringNameParts "_" = [Id "_"]   -- NoName
+stringNameParts s = loop s where
+  loop ""                              = []
+  loop ('_':s)                         = Hole : loop s
+  loop s | (x, s') <- break (== '_') s = Id (stringToRawName x) : loop s'
+
+-- | Number of holes in a 'Name' (i.e., arity of a mixfix-operator).
+class NumHoles a where
+  numHoles :: a -> Int
+
+instance NumHoles [NamePart] where
+  numHoles = length . filter (== Hole)
+
+instance NumHoles Name where
+  numHoles NoName{}       = 0
+  numHoles (Name _ parts) = numHoles parts
+
+instance NumHoles QName where
+  numHoles (QName x)  = numHoles x
+  numHoles (Qual _ x) = numHoles x
 
 -- | Is the name an operator?
 
@@ -217,7 +233,6 @@ projectRoot file (TopLevelModuleName m) =
 noName_ :: Name
 noName_ = noName noRange
 
--- | @noName r = 'Name' r ['Hole']@
 noName :: Range -> Name
 noName r = NoName r (NameId 0 0)
 
@@ -341,3 +356,21 @@ instance KillRange QName where
 instance KillRange Name where
   killRange (Name r ps)  = Name (killRange r) ps
   killRange (NoName r i) = NoName (killRange r) i
+
+------------------------------------------------------------------------
+-- * NFData instances
+------------------------------------------------------------------------
+
+-- | Ranges are not forced.
+
+instance NFData Name where
+  rnf (Name _ ns)  = rnf ns
+  rnf (NoName _ n) = rnf n
+
+instance NFData NamePart where
+  rnf Hole   = ()
+  rnf (Id s) = rnf s
+
+instance NFData QName where
+  rnf (Qual a b) = rnf a `seq` rnf b
+  rnf (QName a)  = rnf a
