@@ -1,6 +1,7 @@
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE FlexibleInstances #-}
 
 {-| Pretty printer for the concrete syntax.
@@ -9,15 +10,12 @@ module Agda.Syntax.Concrete.Pretty where
 
 import Prelude hiding (null)
 
-import Data.Char
 import Data.Functor
 import Data.Maybe
 
-import qualified Agda.Syntax.Common as Common
-import Agda.Syntax.Common hiding (Arg, Dom, NamedArg, ArgInfo)
+import Agda.Syntax.Common
 import Agda.Syntax.Concrete
 import Agda.Syntax.Fixity
-import Agda.Syntax.Literal
 import Agda.Syntax.Notation
 import Agda.Syntax.Position
 
@@ -35,7 +33,8 @@ instance Show Pattern         where show = show . pretty
 instance Show TypedBinding    where show = show . pretty
 instance Show TypedBindings   where show = show . pretty
 instance Show LamBinding      where show = show . pretty
-instance Show ImportDirective where show = show . pretty
+instance (Pretty a, Pretty b) => Show (ImportDirective' a b)
+                              where show = show . pretty
 instance Show Pragma          where show = show . pretty
 instance Show RHS             where show = show . pretty
 
@@ -105,6 +104,10 @@ instance Pretty (OpApp Expr) where
   pretty (Ordinary e) = pretty e
   pretty (SyntaxBindingLambda r bs e) = pretty (Lam r bs e)
 
+instance Pretty a => Pretty (MaybePlaceholder a) where
+  pretty Placeholder{}     = text "_"
+  pretty (NoPlaceholder e) = pretty e
+
 instance Pretty Expr where
     pretty e =
         case e of
@@ -136,7 +139,7 @@ instance Pretty Expr where
             AbsurdLam _ Instance -> lambda <+> text "{{}}"
             AbsurdLam _ Hidden -> lambda <+> text "{}"
             ExtendedLam _ pes ->
-              lambda <+> bracesAndSemicolons (map (\(x,y,z) -> prettyClause x y z) pes)
+              lambda <+> bracesAndSemicolons (map (\(x,y,z,_) -> prettyClause x y z) pes)
                    where prettyClause lhs rhs wh = sep [ pretty lhs
                                                        , nest 2 $ pretty' rhs
                                                        ] $$ nest 2 (pretty wh)
@@ -161,15 +164,14 @@ instance Pretty Expr where
             As _ x e  -> pretty x <> text "@" <> pretty e
             Dot _ e   -> text "." <> pretty e
             Absurd _  -> text "()"
-            Rec _ xs  -> sep [text "record", bracesAndSemicolons (map recPr xs)]
+            Rec _ xs  -> sep [text "record", bracesAndSemicolons (map pretty xs)]
             RecUpdate _ e xs ->
-              sep [text "record" <+> pretty e, bracesAndSemicolons (map recPr xs)]
+              sep [text "record" <+> pretty e, bracesAndSemicolons (map pretty xs)]
             ETel []  -> text "()"
             ETel tel -> fsep $ map pretty tel
             QuoteGoal _ x e -> sep [text "quoteGoal" <+> pretty x <+> text "in",
                                     nest 2 $ pretty e]
-            QuoteContext _ x e -> sep [text "quoteContext" <+> pretty x <+> text "in",
-                                    nest 2 $ pretty e]
+            QuoteContext _ -> text "quoteContext"
             Quote _ -> text "quote"
             QuoteTerm _ -> text "quoteTerm"
             Unquote _ -> text "unquote"
@@ -180,8 +182,15 @@ instance Pretty Expr where
             -- Andreas, 2011-10-03 print irrelevant things as .(e)
             DontCare e -> text "." <> parens (pretty e)
             Equal _ a b -> pretty a <+> text "=" <+> pretty b
-        where
-          recPr (x, e) = sep [ pretty x <+> text "=" , nest 2 $ pretty e ]
+
+instance (Pretty a, Pretty b) => Pretty (Either a b) where
+  pretty = either pretty pretty
+
+instance Pretty a => Pretty (FieldAssignment' a) where
+  pretty (FieldAssignment x e) = sep [ pretty x <+> text "=" , nest 2 $ pretty e ]
+
+instance Pretty ModuleAssignment where
+  pretty (ModuleAssignment m es i) = (fsep $ pretty m : map pretty es) <+> pretty i
 
 instance Pretty BoundName where
   pretty BName{ boundName = x, boundLabel = l }
@@ -195,7 +204,7 @@ instance Pretty LamBinding where
 
 instance Pretty TypedBindings where
   pretty (TypedBindings _ a) = prettyRelevance a $ prettyHiding a p $
-    pretty $ WithColors (argColors a) $ unArg a
+    pretty $ unArg a
       where
         p | isUnderscore (unArg a) = id
           | otherwise        = parens
@@ -210,27 +219,13 @@ instance Pretty Tel where
       | any isMeta tel = text "∀" <+> fsep (map pretty tel)
       | otherwise      = fsep (map pretty tel)
       where
-        isMeta (TypedBindings _ (Common.Arg _ (TBind _ _ (Underscore _ Nothing)))) = True
+        isMeta (TypedBindings _ (Arg _ (TBind _ _ (Underscore _ Nothing)))) = True
         isMeta _ = False
 
 
-instance Pretty ColoredTypedBinding where
-                -- (x y :{ i j } A) -> ...
-    pretty (WithColors [] (TBind _ xs (Underscore _ Nothing))) =
-        fsep (map pretty xs)
-    pretty (WithColors [] (TLet _ ds)) =
-        text "let" <+> vcat (map pretty ds)
-    pretty (WithColors _ (TLet _ _)) = __IMPOSSIBLE__
-    pretty (WithColors cs (TBind _ xs e)) =
-        sep [ fsep (map pretty xs)
-            , pColors ":" cs <+> pretty e
-            ]
-
-pColors :: String -> [Color] -> Doc
-pColors s [] = text s
-pColors s cs = text (s ++ "{") <+> fsep (map pretty cs) <+> text "}"
-
 instance Pretty TypedBinding where
+    pretty (TBind _ xs (Underscore _ Nothing)) =
+      fsep (map pretty xs)
     pretty (TBind _ xs e) =
         sep [ fsep (map pretty xs)
             , text ":" <+> pretty e
@@ -239,10 +234,10 @@ instance Pretty TypedBinding where
         text "let" <+> vcat (map pretty ds)
 
 smashTel :: Telescope -> Telescope
-smashTel (TypedBindings r (Common.Arg i  (TBind r' xs e)) :
-          TypedBindings _ (Common.Arg i' (TBind _  ys e')) : tel)
+smashTel (TypedBindings r (Arg i  (TBind r' xs e)) :
+          TypedBindings _ (Arg i' (TBind _  ys e')) : tel)
   | show i == show i' && show e == show e' && all (isUnnamed . dget) (xs ++ ys) =
-    smashTel (TypedBindings r (Common.Arg i (TBind r' (xs ++ ys) e)) : tel)
+    smashTel (TypedBindings r (Arg i (TBind r' (xs ++ ys) e)) : tel)
   where
     isUnnamed x = boundLabel x == boundName x
 smashTel (b : tel) = b : smashTel tel
@@ -299,15 +294,19 @@ instance Pretty Declaration where
     pretty d =
         case d of
             TypeSig i x e ->
-                sep [ prettyRelevance i $ pretty x <+> pColors ":" (argInfoColors i)
+                sep [ prettyRelevance i $ pretty x <+> text ":"
                     , nest 2 $ pretty e
                     ]
-            Field x (Common.Arg i e) ->
+            Field inst x (Arg i e) ->
                 sep [ text "field"
-                    , nest 2 $ prettyRelevance i $ prettyHiding i id $
+                    , nest 2 $ mkInst inst $
+                      prettyRelevance i $ prettyHiding i id $
                         pretty $ TypeSig (i {argInfoRelevance = Relevant}) x e
                     ]
-            FunClause lhs rhs wh ->
+                where
+                  mkInst InstanceDef    d = sep [ text "instance", nest 2 d ]
+                  mkInst NotInstanceDef d = d
+            FunClause lhs rhs wh _ ->
                 sep [ pretty lhs
                     , nest 2 $ pretty rhs
                     ] $$ nest 2 (pretty wh)
@@ -349,13 +348,14 @@ instance Pretty Declaration where
                             , pretty e
                             ]
                     ]
-            Record _ x ind con tel me cs ->
+            Record _ x ind eta con tel me cs ->
                 sep [ hsep  [ text "record"
                             , pretty x
                             , fcat (map pretty tel)
                             ]
                     , nest 2 $ pType me
                     ] $$ nest 2 (vcat $ pInd ++
+                                        pEta ++
                                         pCon ++
                                         map pretty cs)
               where pType (Just e) = hsep
@@ -366,7 +366,8 @@ instance Pretty Declaration where
                     pType Nothing  =
                               text "where"
                     pInd = maybeToList $ text . show . rangedThing <$> ind
-                    pCon = maybeToList $ (text "constructor" <+>) . pretty <$> con
+                    pEta = maybeToList $ (\x -> if x then text "eta-equality" else text "no-eta-equality") <$> eta
+                    pCon = maybeToList $ (text "constructor" <+>) . pretty <$> fst <$> con
             Infix f xs  ->
                 pretty f <+> (fsep $ punctuate comma $ map pretty xs)
             Syntax n xs -> text "syntax" <+> pretty n <+> text "..."
@@ -376,6 +377,7 @@ instance Pretty Declaration where
             Abstract _ ds   -> namedBlock "abstract" ds
             Private _ ds    -> namedBlock "private" ds
             InstanceB _ ds  -> namedBlock "instance" ds
+            Macro _ ds      -> namedBlock "macro" ds
             Postulate _ ds  -> namedBlock "postulate" ds
             Primitive _ ds  -> namedBlock "primitive" ds
             Module _ x tel ds ->
@@ -403,8 +405,10 @@ instance Pretty Declaration where
                 where
                     as Nothing  = empty
                     as (Just x) = text "as" <+> pretty (asName x)
-            UnquoteDecl _ x t ->
-              sep [ text "unquoteDecl" <+> pretty x <+> text "=", nest 2 $ pretty t ]
+            UnquoteDecl _ xs t ->
+              sep [ text "unquoteDecl" <+> fsep (map pretty xs) <+> text "=", nest 2 $ pretty t ]
+            UnquoteDef _ xs t ->
+              sep [ text "unquoteDef" <+> fsep (map pretty xs) <+> text "=", nest 2 $ pretty t ]
             Pragma pr   -> sep [ text "{-#" <+> pretty pr, text "#-}" ]
         where
             namedBlock s ds =
@@ -427,20 +431,32 @@ instance Pretty Pragma where
       hsep [ text "COMPILED_EXPORT", pretty x, text hs ]
     pretty (CompiledTypePragma _ x hs) =
       hsep [ text "COMPILED_TYPE", pretty x, text hs ]
+    pretty (CompiledDeclareDataPragma _ x hs) =
+      hsep [ text "COMPILED_DECLARE_DATA", pretty x, text hs ]
     pretty (CompiledDataPragma _ x hs hcs) =
       hsep $ [text "COMPILED_DATA", pretty x] ++ map text (hs : hcs)
     pretty (CompiledEpicPragma _ x e) =
       hsep [ text "COMPILED_EPIC", pretty x, text e ]
     pretty (CompiledJSPragma _ x e) =
       hsep [ text "COMPILED_JS", pretty x, text e ]
+    pretty (CompiledUHCPragma _ x e) =
+      hsep [ text "COMPILED_UHC", pretty x, text e ]
+    pretty (CompiledDataUHCPragma _ x crd crcs) =
+      hsep $ [ text "COMPILED_DATA_UHC", pretty x] ++ map text (crd : crcs)
+    pretty (HaskellCodePragma _ s) =
+      vcat (text "HASKELL" : map text (lines s))
+    pretty (NoSmashingPragma _ i) =
+      hsep $ [text "NO_SMASHING", pretty i]
     pretty (StaticPragma _ i) =
       hsep $ [text "STATIC", pretty i]
+    pretty (InlinePragma _ i) =
+      hsep $ [text "INLINE", pretty i]
     pretty (ImportPragma _ i) =
       hsep $ [text "IMPORT", text i]
+    pretty (ImportUHCPragma _ i) =
+      hsep $ [text "IMPORT_UHC", text i]
     pretty (ImpossiblePragma _) =
       hsep $ [text "IMPOSSIBLE"]
-    pretty (EtaPragma _ x) =
-      hsep $ [text "ETA", pretty x]
     pretty (TerminationCheckPragma _ tc) =
       case tc of
         TerminationCheck       -> __IMPOSSIBLE__
@@ -448,9 +464,13 @@ instance Pretty Pragma where
         NonTerminating         -> text "NON_TERMINATING"
         Terminating            -> text "TERMINATING"
         TerminationMeasure _ x -> hsep $ [text "MEASURE", pretty x]
+    pretty (CatchallPragma _) = text "CATCHALL"
+    pretty (DisplayPragma _ lhs rhs) = text "DISPLAY" <+> sep [ pretty lhs <+> text "=", nest 2 $ pretty rhs ]
+    pretty (NoPositivityCheckPragma _) = text "NO_POSITIVITY_CHECK"
 
 instance Pretty Fixity where
-    pretty (Fixity _ n ass) = text s <+> text (show n)
+    pretty (Fixity _ Unrelated   _)   = __IMPOSSIBLE__
+    pretty (Fixity _ (Related n) ass) = text s <+> text (show n)
       where
       s = case ass of
             LeftAssoc  -> "infixl"
@@ -461,14 +481,15 @@ instance Pretty GenPart where
     pretty (IdPart x)   = text x
     pretty BindHole{}   = underscore
     pretty NormalHole{} = underscore
+    pretty WildHole{}   = underscore
 
 instance Pretty Notation where
     pretty = hcat . map pretty
 
 instance Pretty Fixity' where
     pretty (Fixity' fix nota)
-      | nota == defaultNotation = pretty fix
-      | otherwise               = text "syntax" <+> pretty nota
+      | nota == noNotation = pretty fix
+      | otherwise          = text "syntax" <+> pretty nota
 
 instance Pretty e => Pretty (Arg e) where
  -- Andreas 2010-09-21: do not print relevance in general, only in function types!
@@ -490,7 +511,7 @@ instance Pretty Pattern where
             IdentP x        -> pretty x
             AppP p1 p2      -> sep [ pretty p1, nest 2 $ pretty p2 ]
             RawAppP _ ps    -> fsep $ map pretty ps
-            OpAppP _ q _ ps -> fsep $ prettyOpApp q ps
+            OpAppP _ q _ ps -> fsep $ prettyOpApp q (fmap (fmap (fmap NoPlaceholder)) ps)
             HiddenP _ p     -> braces' $ pretty p
             InstanceP _ p   -> dbraces $ pretty p
             ParenP _ p      -> parens $ pretty p
@@ -500,30 +521,63 @@ instance Pretty Pattern where
             AbsurdP _       -> text "()"
             LitP l          -> pretty l
             QuoteP _        -> text "quote"
+            RecP _ fs       -> sep [ text "record", bracesAndSemicolons (map pretty fs) ]
 
-prettyOpApp :: Pretty a => QName -> [a] -> [Doc]
-prettyOpApp q es = prOp ms xs es
+prettyOpApp ::
+  Pretty a => QName -> [NamedArg (MaybePlaceholder a)] -> [Doc]
+prettyOpApp q es = merge [] $ prOp ms xs es
   where
     ms = init (qnameParts q)
     xs = case unqualify q of
            Name _ xs -> xs
            NoName{}  -> __IMPOSSIBLE__
-    prOp ms (Hole : xs) (e : es) = pretty e : prOp ms xs es
-    prOp _  (Hole : _)  []       = __IMPOSSIBLE__
-    prOp ms (Id x : xs) es       = pretty (foldr Qual (QName (Name noRange $ [Id x])) ms) : prOp [] xs es
-    prOp _  []       []          = []
-    prOp _  []       es          = map pretty es
 
-instance Pretty ImportDirective where
+    prOp ms (Hole : xs) (e : es) = (pretty e, case namedArg e of
+                                                Placeholder p -> Just p
+                                                _             -> Nothing) :
+                                   prOp ms xs es
+    prOp _  (Hole : _)  []       = __IMPOSSIBLE__
+    prOp ms (Id x : xs) es       = ( pretty (foldr Qual (QName (Name noRange $ [Id x])) ms)
+                                   , Nothing
+                                   ) : prOp [] xs es
+    prOp _  []       []          = []
+    prOp _  []       es          = map (\e -> (pretty e, Nothing)) es
+
+    -- Section underscores should be printed without surrounding
+    -- whitespace. This function takes care of that.
+    merge :: [Doc] -> [(Doc, Maybe Placeholder)] -> [Doc]
+    merge before []                            = reverse before
+    merge before ((d, Nothing) : after)        = merge (d : before) after
+    merge before ((d, Just Beginning) : after) = mergeRight before d after
+    merge before ((d, Just End)       : after) = case mergeLeft d before of
+                                                   (d, bs) -> merge (d : bs) after
+    merge before ((d, Just Middle)    : after) = case mergeLeft d before of
+                                                   (d, bs) -> mergeRight bs d after
+
+    mergeRight before d after =
+      reverse before ++
+      case merge [] after of
+        []     -> [d]
+        a : as -> (d <> a) : as
+
+    mergeLeft d before = case before of
+      []     -> (d,      [])
+      b : bs -> (b <> d, bs)
+
+instance (Pretty a, Pretty b) => Pretty (ImportDirective' a b) where
     pretty i =
         sep [ public (publicOpen i)
-            , pretty $ usingOrHiding i
-            , rename $ renaming i
+            , pretty $ using i
+            , prettyHiding $ hiding i
+            , rename $ impRenaming i
             ]
         where
             public True  = text "public"
             public False = empty
 
+            prettyHiding [] = empty
+            prettyHiding xs = text "hiding" <+> parens (fsep $ punctuate (text ";") $ map pretty xs)
+
             rename [] = empty
             rename xs = hsep [ text "renaming"
                              , parens $ fsep $ punctuate (text ";") $ map pr xs
@@ -531,13 +585,11 @@ instance Pretty ImportDirective where
 
             pr r = hsep [ pretty (renFrom r), text "to", pretty (renTo r) ]
 
-instance Pretty UsingOrHiding where
-    pretty (Hiding [])  = empty
-    pretty (Hiding xs)  =
-        text "hiding" <+> parens (fsep $ punctuate (text ";") $ map pretty xs)
+instance (Pretty a, Pretty b) => Pretty (Using' a b) where
+    pretty UseEverything = empty
     pretty (Using xs)    =
         text "using" <+> parens (fsep $ punctuate (text ";") $ map pretty xs)
 
-instance Pretty ImportedName where
+instance (Pretty a, Pretty b) => Pretty (ImportedName' a b) where
     pretty (ImportedName x)     = pretty x
     pretty (ImportedModule x)   = text "module" <+> pretty x
