@@ -3,14 +3,20 @@
 {-# LANGUAGE DeriveFoldable #-}
 {-# LANGUAGE DeriveFunctor #-}
 {-# LANGUAGE DeriveTraversable #-}
+{-# LANGUAGE PatternGuards #-}
 
 {-| Definitions for fixity, precedence levels, and declared syntax.
 -}
 module Agda.Syntax.Fixity where
 
+import Prelude hiding (concatMap)
+
+import Control.DeepSeq
+
 import Data.Foldable
 import Data.Function
-import Data.List as List
+import qualified Data.List as List
+import Data.Maybe
 import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Traversable
@@ -22,6 +28,7 @@ import qualified Agda.Syntax.Abstract.Name as A
 import Agda.Syntax.Concrete.Name
 import Agda.Syntax.Notation
 
+import Agda.Utils.Lens
 import Agda.Utils.List
 
 #include "undefined.h"
@@ -32,7 +39,7 @@ import Agda.Utils.Impossible
 -- | The notation is handled as the fixity in the renamer.
 --   Hence, they are grouped together in this type.
 data Fixity' = Fixity'
-    { theFixity   :: Fixity
+    { theFixity   :: !Fixity
     , theNotation :: Notation
     }
   deriving (Typeable, Show, Eq)
@@ -43,35 +50,45 @@ data ThingWithFixity x = ThingWithFixity x Fixity'
 
 -- | All the notation information related to a name.
 data NewNotation = NewNotation
-  { notaName   :: QName
-  , notaNames  :: Set A.Name
+  { notaName  :: QName
+  , notaNames :: Set A.Name
     -- ^ The names the syntax and/or fixity belong to.
     --
     -- Invariant: The set is non-empty. Every name in the list matches
     -- 'notaName'.
   , notaFixity :: Fixity
     -- ^ Associativity and precedence (fixity) of the names.
-  , notation   :: Notation
+  , notation :: Notation
     -- ^ Syntax associated with the names.
+  , notaIsOperator :: Bool
+    -- ^ True if the notation comes from an operator (rather than a
+    -- syntax declaration).
   } deriving (Typeable, Show)
 
 -- | If an operator has no specific notation, then it is computed from
 -- its name.
 namesToNotation :: QName -> A.Name -> NewNotation
 namesToNotation q n = NewNotation
-  { notaName   = q
-  , notaNames  = Set.singleton n
-  , notaFixity = f
-  , notation   = if null syn then syntaxOf $ unqualify q else syn
+  { notaName       = q
+  , notaNames      = Set.singleton n
+  , notaFixity     = f
+  , notation       = if null syn then syntaxOf (unqualify q) else syn
+  , notaIsOperator = null syn
   }
   where Fixity' f syn = A.nameFixity n
 
+-- | Replace 'noFixity' by 'defaultFixity'.
+useDefaultFixity :: NewNotation -> NewNotation
+useDefaultFixity n
+  | notaFixity n == noFixity = n { notaFixity = defaultFixity }
+  | otherwise                = n
+
 -- | Return the 'IdPart's of a notation, the first part qualified,
 --   the other parts unqualified.
 --   This allows for qualified use of operators, e.g.,
 --   @M.for x ∈ xs return e@, or @x ℕ.+ y@.
 notationNames :: NewNotation -> [QName]
-notationNames (NewNotation q _ _ parts) =
+notationNames (NewNotation q _ _ parts _) =
   zipWith ($) (reQualify : repeat QName) [Name noRange [Id x] | IdPart x <- parts ]
   where
     -- The qualification of @q@.
@@ -96,36 +113,104 @@ syntaxOf (Name _ xs)  = mkSyn 0 xs
     mkSyn n (Hole : xs) = NormalHole (defaultNamedArg n) : mkSyn (1 + n) xs
     mkSyn n (Id x : xs) = IdPart x : mkSyn n xs
 
-defaultFixity' :: Fixity'
-defaultFixity' = Fixity' defaultFixity defaultNotation
-
 noFixity' :: Fixity'
 noFixity' = Fixity' noFixity noNotation
 
--- | Merges all 'NewNotation's that have the same notation.
+-- | Merges 'NewNotation's that have the same precedence level and
+-- notation, with two exceptions:
 --
--- If all 'NewNotation's with a given notation have the same fixity,
--- then this fixity is preserved, and otherwise it is replaced by
--- 'defaultFixity'.
+-- * Operators and notations coming from syntax declarations are kept
+--   separate.
+--
+-- * If /all/ instances of a given 'NewNotation' have the same
+--   precedence level or are \"unrelated\", then they are merged. They
+--   get the given precedence level, if any, and otherwise they become
+--   unrelated (but related to each other).
+--
+-- If 'NewNotation's that are merged have distinct associativities,
+-- then they get 'NonAssoc' as their associativity.
 --
 -- Precondition: No 'A.Name' may occur in more than one list element.
 -- Every 'NewNotation' must have the same 'notaName'.
 --
 -- Postcondition: No 'A.Name' occurs in more than one list element.
 mergeNotations :: [NewNotation] -> [NewNotation]
-mergeNotations = map (merge . fixFixities) . groupOn notation
+mergeNotations =
+  map merge .
+  concatMap groupIfLevelsMatch .
+  groupOn (\n -> ( notation n
+                 , notaIsOperator n
+                 ))
   where
-  fixFixities ns
-    | allEqual (map notaFixity ns) = ns
-    | otherwise                    =
-        map (\n -> n { notaFixity = defaultFixity }) ns
+  groupIfLevelsMatch :: [NewNotation] -> [[NewNotation]]
+  groupIfLevelsMatch []         = __IMPOSSIBLE__
+  groupIfLevelsMatch ns@(n : _) =
+    if allEqual (map fixityLevel related)
+    then [sameAssoc (sameLevel ns)]
+    else map (: []) ns
+    where
+    -- Fixities of operators whose precedence level is not Unrelated.
+    related = mapMaybe (\f -> case fixityLevel f of
+                                Unrelated  -> Nothing
+                                Related {} -> Just f)
+                       (map notaFixity ns)
+
+    -- Precondition: All related operators have the same precedence
+    -- level.
+    --
+    -- Gives all unrelated operators the same level.
+    sameLevel = map (set (_notaFixity . _fixityLevel) level)
+      where
+      level = case related of
+        f : _ -> fixityLevel f
+        []    -> Unrelated
+
+    -- If all related operators have the same associativity, then the
+    -- unrelated operators get the same associativity, and otherwise
+    -- all operators get the associativity NonAssoc.
+    sameAssoc = map (set (_notaFixity . _fixityAssoc) assoc)
+      where
+      assoc = case related of
+        f : _ | allEqual (map fixityAssoc related) -> fixityAssoc f
+        _                                          -> NonAssoc
 
   merge :: [NewNotation] -> NewNotation
   merge []         = __IMPOSSIBLE__
   merge ns@(n : _) = n { notaNames = Set.unions $ map notaNames ns }
 
+-- * Sections
+
+-- | Sections, as well as non-sectioned operators.
+
+data NotationSection = NotationSection
+  { sectNotation  :: NewNotation
+  , sectKind      :: NotationKind
+    -- ^ For non-sectioned operators this should match the notation's
+    -- 'notationKind'.
+  , sectLevel     :: Maybe PrecedenceLevel
+    -- ^ Effective precedence level. 'Nothing' for closed notations.
+  , sectIsSection :: Bool
+    -- ^ 'False' for non-sectioned operators.
+  }
+  deriving Show
+
+-- | Converts a notation to a (non-)section.
+
+noSection :: NewNotation -> NotationSection
+noSection n = NotationSection
+  { sectNotation  = n
+  , sectKind      = notationKind (notation n)
+  , sectLevel     = Just (fixityLevel (notaFixity n))
+  , sectIsSection = False
+  }
+
 -- * Fixity
 
+-- | Precedence levels for operators.
+
+data PrecedenceLevel = Unrelated | Related !Integer
+  deriving (Eq, Ord, Show, Typeable)
+
 -- | Associativity.
 
 data Associativity = NonAssoc | LeftAssoc | RightAssoc
@@ -135,8 +220,8 @@ data Associativity = NonAssoc | LeftAssoc | RightAssoc
 
 data Fixity =
   Fixity { fixityRange :: Range
-         , fixityLevel :: Integer
-         , fixityAssoc :: Associativity
+         , fixityLevel :: !PrecedenceLevel
+         , fixityAssoc :: !Associativity
          }
   deriving (Typeable, Show)
 
@@ -148,21 +233,11 @@ instance Ord Fixity where
 
 -- For @instance Pretty Fixity@, see Agda.Syntax.Concrete.Pretty
 
--- | The default fixity. Currently defined to be @'NonAssoc' 20@.
-defaultFixity :: Fixity
-defaultFixity = Fixity noRange 20 NonAssoc
-
--- | Hack used for @syntax@ facility.
 noFixity :: Fixity
-noFixity = Fixity noRange (negate 666) NonAssoc
-  -- Ts,ts,ts, why the number of the beast?  Revelation 13, 18
-  --
-  -- It's not the number of the beast, it's the negation of the
-  -- number of the beast, which must be a divine number, right?
-  --
-  -- The divine is not the negation of evil.
-  -- Evil is only the absense of the good and divine.
+noFixity = Fixity noRange Unrelated NonAssoc
 
+defaultFixity :: Fixity
+defaultFixity = Fixity noRange (Related 20) NonAssoc
 
 -- * Precendence
 
@@ -171,8 +246,7 @@ data Precedence = TopCtx | FunctionSpaceDomainCtx
                 | LeftOperandCtx Fixity | RightOperandCtx Fixity
                 | FunctionCtx | ArgumentCtx | InsideOperandCtx
                 | WithFunCtx | WithArgCtx | DotPatternCtx
-    deriving (Show,Typeable)
-
+    deriving (Show, Typeable, Eq)
 
 -- | The precedence corresponding to a possibly hidden argument.
 hiddenArgumentCtx :: Hiding -> Precedence
@@ -183,14 +257,18 @@ hiddenArgumentCtx Instance  = TopCtx
 -- | Do we need to bracket an operator application of the given fixity
 --   in a context with the given precedence.
 opBrackets :: Fixity -> Precedence -> Bool
-opBrackets (Fixity _ n1 LeftAssoc)
-           (LeftOperandCtx   (Fixity _ n2 LeftAssoc))  | n1 >= n2       = False
-opBrackets (Fixity _ n1 RightAssoc)
-           (RightOperandCtx  (Fixity _ n2 RightAssoc)) | n1 >= n2       = False
+opBrackets                   (Fixity _ (Related n1) LeftAssoc)
+           (LeftOperandCtx   (Fixity _ (Related n2) LeftAssoc))  | n1 >= n2 = False
+opBrackets                   (Fixity _ (Related n1) RightAssoc)
+           (RightOperandCtx  (Fixity _ (Related n2) RightAssoc)) | n1 >= n2 = False
 opBrackets f1
-           (LeftOperandCtx  f2) | fixityLevel f1 > fixityLevel f2 = False
+           (LeftOperandCtx  f2) | Related f1 <- fixityLevel f1
+                                , Related f2 <- fixityLevel f2
+                                , f1 > f2 = False
 opBrackets f1
-           (RightOperandCtx f2) | fixityLevel f1 > fixityLevel f2 = False
+           (RightOperandCtx f2) | Related f1 <- fixityLevel f1
+                                , Related f2 <- fixityLevel f2
+                                , f1 > f2 = False
 opBrackets _ TopCtx = False
 opBrackets _ FunctionSpaceDomainCtx = False
 opBrackets _ InsideOperandCtx       = False
@@ -245,8 +323,31 @@ instance KillRange Fixity' where
 instance KillRange x => KillRange (ThingWithFixity x) where
   killRange (ThingWithFixity c f) = ThingWithFixity (killRange c) f
 
+-- * Some lenses
+
+_notaFixity :: Lens' Fixity NewNotation
+_notaFixity f r = f (notaFixity r) <&> \x -> r { notaFixity = x }
+
+_fixityAssoc :: Lens' Associativity Fixity
+_fixityAssoc f r = f (fixityAssoc r) <&> \x -> r { fixityAssoc = x }
+
+_fixityLevel :: Lens' PrecedenceLevel Fixity
+_fixityLevel f r = f (fixityLevel r) <&> \x -> r { fixityLevel = x }
+
 ------------------------------------------------------------------------
 -- * Printing
 ------------------------------------------------------------------------
 
 -- deriving instance Show Fixity'
+
+------------------------------------------------------------------------
+-- * NFData instances
+------------------------------------------------------------------------
+
+instance NFData Fixity' where
+  rnf (Fixity' _ a) = rnf a
+
+-- | Ranges are not forced.
+
+instance NFData Fixity where
+  rnf (Fixity _ _ _) = ()
