@@ -2,6 +2,7 @@
 {-# LANGUAGE DeriveDataTypeable         #-}
 {-# LANGUAGE DeriveFoldable             #-}
 {-# LANGUAGE DeriveFunctor              #-}
+{-# LANGUAGE DeriveGeneric              #-}
 {-# LANGUAGE DeriveTraversable          #-}
 {-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
@@ -19,7 +20,9 @@
 module Agda.Syntax.Position
   ( -- * Positions
     Position
+  , PositionWithoutFile
   , Position'(..)
+  , SrcFile
   , positionInvariant
   , startPos
   , movePos
@@ -28,20 +31,26 @@ module Agda.Syntax.Position
 
     -- * Intervals
   , Interval
+  , IntervalWithoutFile
   , Interval'(..)
   , intervalInvariant
+  , posToInterval
   , takeI
   , dropI
 
     -- * Ranges
   , Range
-  , Range'(..)
+  , Range'
   , rangeInvariant
+  , consecutiveAndSeparated
+  , intervalsToRange
+  , rangeIntervals
+  , rangeFile
   , rightMargin
   , noRange
-  , posToRange
-  , rStart
-  , rEnd
+  , posToRange, posToRange'
+  , rStart, rStart'
+  , rEnd, rEnd'
   , rangeToInterval
   , continuous
   , continuousPerLine
@@ -70,6 +79,7 @@ import Control.Applicative hiding (empty)
 import Control.Monad
 
 import Data.Foldable (Foldable)
+import qualified Data.Foldable as Fold
 import Data.Function
 import Data.Int
 import Data.List hiding (null)
@@ -79,15 +89,19 @@ import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Traversable (Traversable)
 import Data.Typeable (Typeable)
+import Data.Sequence (Seq)
+import qualified Data.Sequence as Seq
+import Data.Void
+
+import GHC.Generics (Generic)
 
 import Test.QuickCheck.All
 
 import Agda.Utils.FileName hiding (tests)
 import Agda.Utils.List hiding (tests)
-import Agda.Utils.Maybe
+import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Null
 import Agda.Utils.Pretty
-import Agda.Utils.TestHelpers
 import Agda.Utils.QuickCheck
 
 #include "undefined.h"
@@ -107,7 +121,7 @@ import Agda.Utils.Impossible
 --
 -- Note the invariant which positions have to satisfy: 'positionInvariant'.
 data Position' a = Pn
-  { srcFile :: a
+  { srcFile :: !a
     -- ^ File.
   , posPos  :: !Int32
     -- ^ Position, counting from 1.
@@ -116,7 +130,7 @@ data Position' a = Pn
   , posCol  :: !Int32
     -- ^ Column number, counting from 1.
   }
-    deriving (Typeable, Functor, Foldable, Traversable)
+  deriving (Typeable, Functor, Foldable, Traversable, Generic)
 
 positionInvariant :: Position' a -> Bool
 positionInvariant p =
@@ -131,48 +145,111 @@ instance Eq a => Eq (Position' a) where
 instance Ord a => Ord (Position' a) where
   compare = compare `on` importantPart
 
-type SrcFile = Maybe AbsolutePath
+type SrcFile = Strict.Maybe AbsolutePath
 
-type Position = Position' SrcFile
+type Position            = Position' SrcFile
+type PositionWithoutFile = Position' ()
 
 -- | An interval. The @iEnd@ position is not included in the interval.
 --
 -- Note the invariant which intervals have to satisfy: 'intervalInvariant'.
 data Interval' a = Interval { iStart, iEnd :: !(Position' a) }
-    deriving (Typeable, Eq, Ord, Functor, Foldable, Traversable)
+  deriving (Typeable, Eq, Ord, Functor, Foldable, Traversable, Generic)
 
-type Interval = Interval' SrcFile
+type Interval            = Interval' SrcFile
+type IntervalWithoutFile = Interval' ()
 
 intervalInvariant :: Ord a => Interval' a -> Bool
 intervalInvariant i =
-  all positionInvariant [iStart i, iEnd i] &&
+  all positionInvariant [iStart i, iEnd i]
+    &&
   iStart i <= iEnd i
+    &&
+  srcFile (iStart i) == srcFile (iEnd i)
 
--- | The length of an interval, assuming that the start and end
--- positions are in the same file.
+-- | Sets the 'srcFile' components of the interval.
+
+setIntervalFile :: a -> Interval' b -> Interval' a
+setIntervalFile f (Interval p1 p2) =
+  Interval (p1 { srcFile = f }) (p2 { srcFile = f })
+
+-- | Converts a file name and two positions to an interval.
+posToInterval ::
+  a -> PositionWithoutFile -> PositionWithoutFile -> Interval' a
+posToInterval f p1 p2 = setIntervalFile f $
+  if p1 < p2
+  then Interval p1 p2
+  else Interval p2 p1
+
+-- | The length of an interval.
 iLength :: Interval' a -> Int32
 iLength i = posPos (iEnd i) - posPos (iStart i)
 
--- | A range is a list of intervals. The intervals should be
--- consecutive and separated.
+-- | A range is a file name, plus a sequence of intervals, assumed to
+-- point to the given file. The intervals should be consecutive and
+-- separated.
 --
 -- Note the invariant which ranges have to satisfy: 'rangeInvariant'.
-newtype Range' a = Range [Interval' a]
-  deriving (Typeable, Eq, Ord, Functor, Foldable, Traversable, Null)
+data Range' a
+  = NoRange
+  | Range !a (Seq IntervalWithoutFile)
+  deriving
+    (Typeable, Eq, Ord, Functor, Foldable, Traversable, Generic)
 
 type Range = Range' SrcFile
 
-rangeInvariant :: Range -> Bool
-rangeInvariant (Range []) = True
-rangeInvariant (Range is) =
-  all intervalInvariant is &&
-  and (zipWith (<) (map iEnd $ init is) (map iStart $ tail is))
+instance Null (Range' a) where
+  null NoRange = True
+  null Range{} = False
+
+  empty = NoRange
+
+-- | The intervals that make up the range. The intervals are
+-- consecutive and separated ('consecutiveAndSeparated').
+rangeIntervals :: Range' a -> [IntervalWithoutFile]
+rangeIntervals NoRange      = []
+rangeIntervals (Range _ is) = Fold.toList is
+
+-- | Turns a file name plus a list of intervals into a range.
+--
+-- Precondition: 'consecutiveAndSeparated'.
+intervalsToRange :: a -> [IntervalWithoutFile] -> Range' a
+intervalsToRange _ [] = NoRange
+intervalsToRange f is = Range f (Seq.fromList is)
+
+-- | Are the intervals consecutive and separated, do they all point to
+-- the same file, and do they satisfy the interval invariant?
+consecutiveAndSeparated :: Ord a => [Interval' a] -> Bool
+consecutiveAndSeparated is =
+  all intervalInvariant is
+    &&
+  allEqual (map (srcFile . iStart) is)
+    &&
+  (null is
+     ||
+   and (zipWith (<) (map iEnd   (init is))
+                    (map iStart (tail is))))
+
+-- | Range invariant.
+rangeInvariant :: Ord a => Range' a -> Bool
+rangeInvariant r =
+  consecutiveAndSeparated (rangeIntervals r)
+    &&
+  case r of
+    Range _ is -> not (null is)
+    NoRange    -> True
+
+-- | The file the range is pointing to.
+rangeFile :: Range -> SrcFile
+rangeFile NoRange     = Strict.Nothing
+rangeFile (Range f _) = f
 
 -- | Conflate a range to its right margin.
 rightMargin :: Range -> Range
-rightMargin r@(Range is) =
-  if null is then r else
-  let i = last is in Range [ i { iStart = iEnd i } ]
+rightMargin r@NoRange      = r
+rightMargin r@(Range f is) = case Seq.viewr is of
+  Seq.EmptyR -> __IMPOSSIBLE__
+  _ Seq.:> i -> intervalToRange f (i { iStart = iEnd i })
 
 -- | Wrapper to indicate that range should be printed.
 newtype PrintRange a = PrintRange a
@@ -183,29 +260,48 @@ class HasRange t where
     getRange :: t -> Range
 
 instance HasRange Interval where
-    getRange i = Range [i]
+    getRange i =
+      intervalToRange (srcFile (iStart i))
+                      (setIntervalFile () i)
 
 instance HasRange Range where
     getRange = id
 
+instance HasRange Bool where
+    getRange _ = noRange
+
+-- | Precondition: The ranges of the list elements must point to the
+-- same file (or be empty).
 instance HasRange a => HasRange [a] where
     getRange = foldr fuseRange noRange
 
+-- | Precondition: The ranges of the tuple elements must point to the
+-- same file (or be empty).
 instance (HasRange a, HasRange b) => HasRange (a,b) where
     getRange = uncurry fuseRange
 
+-- | Precondition: The ranges of the tuple elements must point to the
+-- same file (or be empty).
 instance (HasRange a, HasRange b, HasRange c) => HasRange (a,b,c) where
     getRange (x,y,z) = getRange (x,(y,z))
 
+-- | Precondition: The ranges of the tuple elements must point to the
+-- same file (or be empty).
 instance (HasRange a, HasRange b, HasRange c, HasRange d) => HasRange (a,b,c,d) where
     getRange (x,y,z,w) = getRange (x,(y,(z,w)))
 
+-- | Precondition: The ranges of the tuple elements must point to the
+-- same file (or be empty).
 instance (HasRange a, HasRange b, HasRange c, HasRange d, HasRange e) => HasRange (a,b,c,d,e) where
     getRange (x,y,z,w,v) = getRange (x,(y,(z,(w,v))))
 
+-- | Precondition: The ranges of the tuple elements must point to the
+-- same file (or be empty).
 instance (HasRange a, HasRange b, HasRange c, HasRange d, HasRange e, HasRange f) => HasRange (a,b,c,d,e,f) where
     getRange (x,y,z,w,v,u) = getRange (x,(y,(z,(w,(v,u)))))
 
+-- | Precondition: The ranges of the tuple elements must point to the
+-- same file (or be empty).
 instance (HasRange a, HasRange b, HasRange c, HasRange d, HasRange e, HasRange f, HasRange g) => HasRange (a,b,c,d,e,f,g) where
     getRange (x,y,z,w,v,u,t) = getRange (x,(y,(z,(w,(v,(u,t))))))
 
@@ -213,6 +309,9 @@ instance HasRange a => HasRange (Maybe a) where
     getRange Nothing  = noRange
     getRange (Just a) = getRange a
 
+instance (HasRange a, HasRange b) => HasRange (Either a b) where
+    getRange = either getRange getRange
+
 -- | If it is also possible to set the range, this is the class.
 --
 --   Instances should satisfy @'getRange' ('setRange' r x) == r@.
@@ -361,6 +460,9 @@ killRange19 f a = killRange18 (f $ killRange a)
 instance KillRange Range where
   killRange _ = noRange
 
+instance KillRange Void where
+  killRange = id
+
 instance KillRange () where
   killRange = id
 
@@ -430,24 +532,22 @@ instance (KillRange a, KillRange b) => KillRange (Either a b) where
 
 -- ASR (02 December 2014). This instance is not used anymore (module
 -- the test suite) when reporting errors. See Issue 1293.
-instance Show a => Show (Position' (Maybe a)) where
-  show (Pn Nothing  _ l c) = show l ++ "," ++ show c
-  show (Pn (Just f) _ l c) = show f ++ ":" ++ show l ++ "," ++ show c
+instance Show a => Show (Position' (Strict.Maybe a)) where
+  show (Pn (Strict.Just f) _ l c) = show f ++ ":" ++
+                                    show l ++ "," ++ show c
+  show (Pn Strict.Nothing  _ l c) = show l ++ "," ++ show c
 
-instance Show a => Show (Interval' (Maybe a)) where
-  show (Interval s e) = file ++ start ++ "-" ++ end
+instance Show PositionWithoutFile where
+  show p = show (p { srcFile = Strict.Nothing } :: Position)
+
+instance Show IntervalWithoutFile where
+  show (Interval s e) = start ++ "-" ++ end
     where
-      f  = srcFile s
       sl = posLine s
       el = posLine e
       sc = posCol s
       ec = posCol e
 
-      file :: String
-      file = case f of
-               Nothing -> ""
-               Just f  -> show f ++ ":"
-
       start :: String
       start = show sl ++ "," ++ show sc
 
@@ -455,34 +555,42 @@ instance Show a => Show (Interval' (Maybe a)) where
       end | sl == el  = show ec
           | otherwise = show el ++ "," ++ show ec
 
-instance Show a => Show (Range' (Maybe a)) where
-  show r = case rangeToInterval r of
+instance Show a => Show (Interval' (Strict.Maybe a)) where
+  show i@(Interval s _) = file ++ show (setIntervalFile () i)
+    where
+      file :: String
+      file = case srcFile s of
+               Strict.Nothing -> ""
+               Strict.Just f  -> show f ++ ":"
+
+instance Show a => Show (Range' (Strict.Maybe a)) where
+  show r = case rangeToIntervalWithFile r of
     Nothing -> ""
     Just i  -> show i
 
+instance Show a => Show (Range' (Maybe a)) where
+  show = show . fmap Strict.toStrict
+
 ------------------------------------------------------------------------
 -- Printing
 ------------------------------------------------------------------------
 
-instance Pretty a => Pretty (Position' (Maybe a)) where
-  pretty (Pn Nothing  _ l c) = pretty l <> pretty "," <> pretty c
-  pretty (Pn (Just f) _ l c) =
+instance Pretty a => Pretty (Position' (Strict.Maybe a)) where
+  pretty (Pn Strict.Nothing  _ l c) = pretty l <> pretty "," <> pretty c
+  pretty (Pn (Strict.Just f) _ l c) =
     pretty f <> pretty ":" <> pretty l <> pretty "," <> pretty c
 
-instance Pretty a => Pretty (Interval' (Maybe a)) where
-  pretty (Interval s e) = file <> start <> pretty "-" <> end
+instance Pretty PositionWithoutFile where
+  pretty p = pretty (p { srcFile = Strict.Nothing } :: Position)
+
+instance Pretty IntervalWithoutFile where
+  pretty (Interval s e) = start <> pretty "-" <> end
     where
-      f  = srcFile s
       sl = posLine s
       el = posLine e
       sc = posCol s
       ec = posCol e
 
-      file :: Doc
-      file = case f of
-               Nothing -> empty
-               Just f  -> pretty f <> colon
-
       start :: Doc
       start = pretty sl <> comma <> pretty sc
 
@@ -490,8 +598,16 @@ instance Pretty a => Pretty (Interval' (Maybe a)) where
         | sl == el  = pretty ec
         | otherwise = pretty el <> comma <> pretty ec
 
-instance Pretty a => Pretty (Range' (Maybe a)) where
-  pretty r = case rangeToInterval r of
+instance Pretty a => Pretty (Interval' (Strict.Maybe a)) where
+  pretty i@(Interval s _) = file <> pretty (setIntervalFile () i)
+    where
+      file :: Doc
+      file = case srcFile s of
+               Strict.Nothing -> empty
+               Strict.Just f  -> pretty f <> colon
+
+instance Pretty a => Pretty (Range' (Strict.Maybe a)) where
+  pretty r = case rangeToIntervalWithFile r of
     Nothing -> empty
     Just i  -> pretty i
 
@@ -502,13 +618,22 @@ instance (Pretty a, HasRange a) => Pretty (PrintRange a) where
     Functions on postitions and ranges
  --------------------------------------------------------------------------}
 
+-- | The first position in a file: position 1, line 1, column 1.
+startPos' :: a -> Position' a
+startPos' f = Pn
+  { srcFile = f
+  , posPos  = 1
+  , posLine = 1
+  , posCol  = 1
+  }
+
 -- | The first position in a file: position 1, line 1, column 1.
 startPos :: Maybe AbsolutePath -> Position
-startPos f = Pn { srcFile = f, posPos = 1, posLine = 1, posCol = 1 }
+startPos = startPos' . Strict.toStrict
 
 -- | Ranges between two unknown positions
 noRange :: Range' a
-noRange = Range []
+noRange = NoRange
 
 -- | Advance the position by one character.
 --   A newline character (@'\n'@) moves the position to the first
@@ -547,62 +672,135 @@ dropI :: String -> Interval' a -> Interval' a
 dropI s i | genericLength s > iLength i = __IMPOSSIBLE__
           | otherwise = i { iStart = movePosByString (iStart i) s }
 
+-- | Converts a file name and two positions to a range.
+posToRange' ::
+  a -> PositionWithoutFile -> PositionWithoutFile -> Range' a
+posToRange' f p1 p2 = intervalToRange f (posToInterval () p1 p2)
+
 -- | Converts two positions to a range.
-posToRange :: Ord a => Position' a -> Position' a -> Range' a
-posToRange p1 p2 | p1 < p2   = Range [Interval p1 p2]
-                 | otherwise = Range [Interval p2 p1]
+--
+-- Precondition: The positions have to point to the same file.
+posToRange :: Position' a -> Position' a -> Range' a
+posToRange p1 p2 =
+  posToRange' (srcFile p1) (p1 { srcFile = () }) (p2 { srcFile = () })
+
+-- | Converts a file name and an interval to a range.
+intervalToRange :: a -> IntervalWithoutFile -> Range' a
+intervalToRange f i = Range f (Seq.singleton i)
 
 -- | Converts a range to an interval, if possible.
-rangeToInterval :: Range' a -> Maybe (Interval' a)
-rangeToInterval (Range [])   = Nothing
-rangeToInterval (Range is)   = Just $ Interval { iStart = iStart (head is)
-                                               , iEnd   = iEnd   (last is)
+rangeToIntervalWithFile :: Range' a -> Maybe (Interval' a)
+rangeToIntervalWithFile NoRange      = Nothing
+rangeToIntervalWithFile (Range f is) = case (Seq.viewl is, Seq.viewr is) of
+  (head Seq.:< _, _ Seq.:> last) -> Just $ setIntervalFile f $
+                                      Interval { iStart = iStart head
+                                               , iEnd   = iEnd   last
                                                }
+  _                              -> __IMPOSSIBLE__
+
+-- | Converts a range to an interval, if possible. Note that the
+-- information about the source file is lost.
+rangeToInterval :: Range' a -> Maybe IntervalWithoutFile
+rangeToInterval NoRange      = Nothing
+rangeToInterval (Range _ is) = case (Seq.viewl is, Seq.viewr is) of
+  (head Seq.:< _, _ Seq.:> last) -> Just $
+                                      Interval { iStart = iStart head
+                                               , iEnd   = iEnd   last
+                                               }
+  _                              -> __IMPOSSIBLE__
 
 -- | Returns the shortest continuous range containing the given one.
 continuous :: Range' a -> Range' a
-continuous r = case rangeToInterval r of
-  Nothing -> Range []
-  Just i  -> Range [i]
+continuous NoRange = NoRange
+continuous r@(Range f _) = case rangeToInterval r of
+  Nothing -> __IMPOSSIBLE__
+  Just i  -> intervalToRange f i
 
 -- | Removes gaps between intervals on the same line.
 continuousPerLine :: Ord a => Range' a -> Range' a
-continuousPerLine (Range [])     = Range []
-continuousPerLine (Range (i:is)) = Range $ fuse i is
+continuousPerLine r@NoRange     = r
+continuousPerLine r@(Range f _) =
+  Range f (Seq.unfoldr step (rangeIntervals r))
   where
-    fuse i [] = [i]
-    fuse i (j:is)
-      | sameLine i j = fuse (fuseIntervals i j) is
-      | otherwise    = i : fuse j is
-    sameLine i j = posLine (iEnd i) == posLine (iStart j)
+  step []  = Nothing
+  step [i] = Just (i, [])
+  step (i : is@(j : js))
+    | sameLine  = step (fuseIntervals i j : js)
+    | otherwise = Just (i, is)
+    where
+    sameLine = posLine (iEnd i) == posLine (iStart j)
+
+-- | The initial position in the range, if any.
+rStart' :: Range' a -> Maybe PositionWithoutFile
+rStart' r = iStart <$> rangeToInterval r
 
 -- | The initial position in the range, if any.
 rStart :: Range' a -> Maybe (Position' a)
-rStart r = iStart <$> rangeToInterval r
+rStart NoRange       = Nothing
+rStart r@(Range f _) = (\p -> p { srcFile = f }) <$> rStart' r
+
+-- | The position after the final position in the range, if any.
+rEnd' :: Range' a -> Maybe PositionWithoutFile
+rEnd' r = iEnd <$> rangeToInterval r
 
 -- | The position after the final position in the range, if any.
 rEnd :: Range' a -> Maybe (Position' a)
-rEnd r = iEnd <$> rangeToInterval r
+rEnd NoRange       = Nothing
+rEnd r@(Range f _) = (\p -> p { srcFile = f }) <$> rEnd' r
 
 -- | Finds the least interval which covers the arguments.
+--
+-- Precondition: The intervals must point to the same file.
 fuseIntervals :: Ord a => Interval' a -> Interval' a -> Interval' a
-fuseIntervals x y = Interval { iStart = head ps, iEnd = last ps }
-    where ps = sort [iStart x, iStart y, iEnd x, iEnd y]
+fuseIntervals x y = Interval { iStart = head ss, iEnd = last es }
+    where
+    ss = sort [iStart x, iStart y]
+    es = sort [iEnd   x, iEnd   y]
 
 -- | @fuseRanges r r'@ unions the ranges @r@ and @r'@.
 --
 --   Meaning it finds the least range @r0@ that covers @r@ and @r'@.
+--
+-- Precondition: The ranges must point to the same file (or be empty).
 fuseRanges :: (Ord a) => Range' a -> Range' a -> Range' a
-fuseRanges (Range is) (Range js) = Range (helper is js)
+fuseRanges NoRange       is2           = is2
+fuseRanges is1           NoRange       = is1
+fuseRanges (Range f is1) (Range _ is2) = Range f (fuse is1 is2)
   where
-  helper []     js  = js
-  helper is     []  = is
-  helper (i:is) (j:js)
-    | iEnd i < iStart j = i : helper is     (j:js)
-    | iEnd j < iStart i = j : helper (i:is) js
-    | iEnd i < iEnd j   = helper is (fuseIntervals i j : js)
-    | otherwise         = helper (fuseIntervals i j : is) js
+  fuse is1 is2 = case (Seq.viewl is1, Seq.viewr is1,
+                       Seq.viewl is2, Seq.viewr is2) of
+    (Seq.EmptyL, _, _, _) -> is2
+    (_, _, Seq.EmptyL, _) -> is1
+    (s1 Seq.:< r1, l1 Seq.:> e1, s2 Seq.:< r2, l2 Seq.:> e2)
+        -- Special cases.
+      | iEnd e1 <  iStart s2 -> is1 Seq.>< is2
+      | iEnd e2 <  iStart s1 -> is2 Seq.>< is1
+      | iEnd e1 == iStart s2 -> mergeTouching l1 e1 s2 r2
+      | iEnd e2 == iStart s1 -> mergeTouching l2 e2 s1 r1
+        -- General cases.
+      | iEnd s1 <  iStart s2 -> outputLeftPrefix s1 r1 s2 is2
+      | iEnd s2 <  iStart s1 -> outputLeftPrefix s2 r2 s1 is1
+      | iEnd s1 <  iEnd   s2 -> fuseSome s1 r1 s2 r2
+      | otherwise            -> fuseSome s2 r2 s1 r1
+    _ -> __IMPOSSIBLE__
 
+  mergeTouching l e s r = l Seq.>< i Seq.<| r
+    where
+    i = Interval { iStart = iStart e, iEnd = iEnd s }
+
+  -- The following two functions could use binary search instead of
+  -- linear.
+
+  outputLeftPrefix s1 r1 s2 is2 = s1 Seq.<| r1' Seq.>< fuse r1'' is2
+    where
+    (r1', r1'') = Seq.spanl (\s -> iEnd s < iStart s2) r1
+
+  fuseSome s1 r1 s2 r2 = fuse r1' (fuseIntervals s1 s2 Seq.<| r2)
+    where
+    r1' = Seq.dropWhileL (\s -> iEnd s <= iEnd s2) r1
+
+-- | Precondition: The ranges must point to the same file (or be
+-- empty).
 fuseRange :: (HasRange u, HasRange t) => u -> t -> Range
 fuseRange x y = fuseRanges (getRange x) (getRange y)
 
@@ -610,18 +808,18 @@ fuseRange x y = fuseRanges (getRange x) (getRange y)
 -- positioned at the beginning of @r@. If @r@ does not have a
 -- beginning, then 'noRange' is returned.
 beginningOf :: Range -> Range
-beginningOf r = case rStart r of
-  Nothing  -> noRange
-  Just pos -> posToRange pos pos
+beginningOf NoRange       = NoRange
+beginningOf r@(Range f _) = case rStart' r of
+  Nothing  -> __IMPOSSIBLE__
+  Just pos -> posToRange' f pos pos
 
 -- | @beginningOfFile r@ is an empty range (a single, empty interval)
 -- at the beginning of @r@'s starting position's file. If there is no
 -- such position, then an empty range is returned.
 beginningOfFile :: Range -> Range
-beginningOfFile r = case rStart r of
-  Nothing                   -> noRange
-  Just (Pn { srcFile = f }) -> posToRange p p
-    where p = startPos f
+beginningOfFile NoRange     = NoRange
+beginningOfFile (Range f _) = posToRange' f p p
+  where p = startPos' ()
 
 -- | @x \`withRangeOf\` y@ sets the range of @x@ to the range of @y@.
 withRangeOf :: (SetRange t, HasRange u) => t -> u -> t
@@ -630,17 +828,16 @@ x `withRangeOf` y = setRange (getRange y) x
 ------------------------------------------------------------------------
 -- Test suite
 
--- | The positions corresponding to the interval, /including/ the
--- end-point. This function assumes that the two end points belong to
--- the same file. Note that the 'Arbitrary' instance for 'Position's
--- uses a single, hard-wired file name.
+-- | The positions corresponding to the interval. The positions do not
+-- refer to characters, but to the positions between characters, with
+-- zero pointing to the position before the first character.
 iPositions :: Interval' a -> Set Int32
 iPositions i = Set.fromList [posPos (iStart i) .. posPos (iEnd i)]
 
 -- | The positions corresponding to the range, including the
--- end-points. All ranges are assumed to belong to a single file.
+-- end-points.
 rPositions :: Range' a -> Set Int32
-rPositions (Range is) = Set.unions (map iPositions is)
+rPositions r = Set.unions (map iPositions $ rangeIntervals r)
 
 -- | Constructs the least interval containing all the elements in the
 -- set.
@@ -652,11 +849,14 @@ makeInterval s
 prop_iLength :: Interval' Integer -> Bool
 prop_iLength i = iLength i >= 0
 
+prop_startPos' :: Bool
+prop_startPos' = positionInvariant (startPos' ())
+
 prop_startPos :: Maybe AbsolutePath -> Bool
 prop_startPos = positionInvariant . startPos
 
 prop_noRange :: Bool
-prop_noRange = rangeInvariant noRange
+prop_noRange = rangeInvariant (noRange :: Range)
 
 prop_takeI_dropI :: Interval' Integer -> Property
 prop_takeI_dropI i =
@@ -669,12 +869,38 @@ prop_takeI_dropI i =
     intervalInvariant d &&
     fuseIntervals t d == i
 
+prop_posToRange' ::
+  Integer -> PositionWithoutFile -> PositionWithoutFile -> Bool
+prop_posToRange' f p1 p2 =
+  rangeInvariant (posToRange' f p1 p2)
+
+prop_posToRange :: Position' Integer -> Position' Integer -> Bool
+prop_posToRange p1 p2 =
+  rangeInvariant (posToRange p1 (p2 { srcFile = srcFile p1 }))
+
+prop_intervalToRange :: Integer -> IntervalWithoutFile -> Bool
+prop_intervalToRange f i = rangeInvariant (intervalToRange f i)
+
+rangeToIntervalPropertyTemplate ::
+  Ord a =>
+  (Range' Integer -> Maybe (Interval' a)) ->
+  Range' Integer -> Bool
+rangeToIntervalPropertyTemplate r2i r = case r2i r of
+  Nothing -> r == noRange
+  Just i  ->
+    r /= noRange
+      &&
+    intervalInvariant i
+      &&
+    iPositions i == makeInterval (rPositions r)
+
+prop_rangeToIntervalWithFile :: Range' Integer -> Bool
+prop_rangeToIntervalWithFile =
+  rangeToIntervalPropertyTemplate rangeToIntervalWithFile
+
 prop_rangeToInterval :: Range' Integer -> Bool
-prop_rangeToInterval (Range []) = True
-prop_rangeToInterval r =
-  intervalInvariant i &&
-  iPositions i == makeInterval (rPositions r)
-  where Just i = rangeToInterval r
+prop_rangeToInterval =
+  rangeToIntervalPropertyTemplate rangeToInterval
 
 prop_continuous :: Range -> Bool
 prop_continuous r =
@@ -687,10 +913,12 @@ prop_continuousPerLine r =
   rangeInvariant r'
     &&
   distinct lineNumbers
+    &&
+  rangeFile r' == rangeFile r
   where
-  r'@(Range is') = continuousPerLine r
+  r' = continuousPerLine r
 
-  lineNumbers = concatMap lines is'
+  lineNumbers = concatMap lines (rangeIntervals r')
     where
     lines i | s == e    = [s]
             | otherwise = [s, e]
@@ -706,11 +934,13 @@ prop_fuseIntervals i1 =
     iPositions i ==
       makeInterval (Set.union (iPositions i1) (iPositions i2))
 
-prop_fuseRanges :: Range -> Range -> Bool
-prop_fuseRanges r1 r2 =
-  rangeInvariant r &&
-  rPositions r == Set.union (rPositions r1) (rPositions r2)
-  where r = fuseRanges r1 r2
+prop_fuseRanges :: Range -> Property
+prop_fuseRanges r1 =
+  forAll (rangeInSameFileAs r1) $ \r2 ->
+    let r = fuseRanges r1 r2 in
+    rangeInvariant r
+      &&
+    rPositions r == Set.union (rPositions r1) (rPositions r2)
 
 prop_beginningOf :: Range -> Bool
 prop_beginningOf r = rangeInvariant (beginningOf r)
@@ -728,17 +958,14 @@ instance Arbitrary a => Arbitrary (Position' a) where
     return (Pn {srcFile = srcFile, posPos = pos,
                 posLine = line, posCol = col })
 
--- | Sets the 'srcFile' components of the interval.
-
-setFile :: a -> Interval' a -> Interval' a
-setFile f (Interval p1 p2) =
-  Interval (p1 { srcFile = f }) (p2 { srcFile = f })
-
 -- | Generates an interval located in the same file as the given
 -- interval.
 
-intervalInSameFileAs :: Interval' Integer -> Gen (Interval' Integer)
-intervalInSameFileAs i = setFile (srcFile $ iStart i) <$> arbitrary
+intervalInSameFileAs ::
+  (Arbitrary a, Ord a) => Interval' a -> Gen (Interval' a)
+intervalInSameFileAs i =
+  setIntervalFile (srcFile $ iStart i) <$>
+    (arbitrary :: Gen IntervalWithoutFile)
 
 prop_intervalInSameFileAs :: Interval' Integer -> Property
 prop_intervalInSameFileAs i =
@@ -746,6 +973,25 @@ prop_intervalInSameFileAs i =
     intervalInvariant i' &&
     srcFile (iStart i) == srcFile (iStart i')
 
+-- | Generates a range located in the same file as the given
+-- range (if possible).
+
+rangeInSameFileAs :: (Arbitrary a, Ord a) => Range' a -> Gen (Range' a)
+rangeInSameFileAs NoRange      = arbitrary
+rangeInSameFileAs (Range f is) = do
+  Range _f is <- arbitrary `suchThat` (not . null)
+  return $ Range (f `asTypeOf` _f) is
+
+prop_rangeInSameFileAs :: Range' Integer -> Property
+prop_rangeInSameFileAs r =
+  forAll (rangeInSameFileAs r) $ \r' ->
+    rangeInvariant r'
+      &&
+    case (r, r') of
+      (NoRange, _)            -> True
+      (Range f _, Range f' _) -> f == f'
+      (Range _ _, NoRange)    -> False
+
 instance (Arbitrary a, Ord a) => Arbitrary (Interval' a) where
   arbitrary = do
     (p1, p2 :: Position' a) <- liftM2 (,) arbitrary arbitrary
@@ -753,16 +999,19 @@ instance (Arbitrary a, Ord a) => Arbitrary (Interval' a) where
     return (Interval p1' p2')
 
 instance (Ord a, Arbitrary a) => Arbitrary (Range' a) where
-  arbitrary = Range . fuse . sort . fixFiles <$> arbitrary
+  arbitrary = do
+    f <- arbitrary
+    intervalsToRange f . fuse . sort <$> arbitrary
     where
-    fixFiles []       = []
-    fixFiles (i : is) = i : map (setFile $ srcFile $ iStart i) is
-
     fuse (i1 : i2 : is)
       | iEnd i1 >= iStart i2 = fuse (fuseIntervals i1 i2 : is)
       | otherwise            = i1 : fuse (i2 : is)
     fuse is = is
 
+instance CoArbitrary a => CoArbitrary (Position' a)
+instance CoArbitrary a => CoArbitrary (Interval' a)
+instance CoArbitrary a => CoArbitrary (Range' a)
+
 prop_positionInvariant :: Position' Integer -> Bool
 prop_positionInvariant = positionInvariant
 
@@ -772,9 +1021,9 @@ prop_intervalInvariant = intervalInvariant
 prop_rangeInvariant :: Range -> Bool
 prop_rangeInvariant = rangeInvariant
 
-instance Show (Position' Integer) where show = show . fmap Just
-instance Show (Interval' Integer) where show = show . fmap Just
-instance Show (Range'    Integer) where show = show . fmap Just
+instance Show (Position' Integer) where show = show . fmap Strict.Just
+instance Show (Interval' Integer) where show = show . fmap Strict.Just
+instance Show (Range'    Integer) where show = show . fmap Strict.Just
 
 ------------------------------------------------------------------------
 -- * All tests
@@ -789,20 +1038,3 @@ tests :: IO Bool
 tests = do
   putStrLn "Agda.Syntax.Position"
   $quickCheckAll
-
--- tests = runTests "Agda.Syntax.Position"
---   [ quickCheck' (positionInvariant :: Position -> Bool)
---   , quickCheck' intervalInvariant
---   , quickCheck' rangeInvariant
---   , quickCheck' prop_iLength
---   , quickCheck' prop_startPos
---   , quickCheck' prop_noRange
---   , quickCheck' prop_takeI_dropI
---   , quickCheck' prop_rangeToInterval
---   , quickCheck' prop_continuous
---   , quickCheck' prop_fuseIntervals
---   , quickCheck' prop_fuseRanges
---   , quickCheck' prop_beginningOf
---   , quickCheck' prop_beginningOfFile
---   , quickCheck' prop_intervalInSameFileAs
---   ]
