@@ -1,10 +1,15 @@
-{-# LANGUAGE CPP #-}
-{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE CPP                    #-}
+{-# LANGUAGE FlexibleInstances      #-}
 {-# LANGUAGE FunctionalDependencies #-}
-{-# LANGUAGE MultiParamTypeClasses #-}
-{-# LANGUAGE PatternGuards #-}
-{-# LANGUAGE TupleSections #-}
-{-# LANGUAGE UndecidableInstances #-}
+{-# LANGUAGE LambdaCase             #-}
+{-# LANGUAGE MultiParamTypeClasses  #-}
+{-# LANGUAGE PatternGuards          #-}
+{-# LANGUAGE TupleSections          #-}
+{-# LANGUAGE UndecidableInstances   #-}
+
+#if __GLASGOW_HASKELL__ >= 710
+{-# LANGUAGE FlexibleContexts #-}
+#endif
 
 -- {-# OPTIONS -fwarn-unused-binds #-}
 
@@ -33,16 +38,15 @@ import Prelude hiding (null)
 import Control.Applicative hiding (empty)
 import Control.Monad.Reader
 
-import Data.List as List hiding (null)
 import qualified Data.Map as Map
 import Data.Maybe
 import Data.Monoid
 import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Traversable (traverse)
+import Data.Void
 
-import Agda.Syntax.Common hiding (Arg, Dom, NamedArg)
-import qualified Agda.Syntax.Common as Common
+import Agda.Syntax.Common
 import Agda.Syntax.Position
 import Agda.Syntax.Literal
 import Agda.Syntax.Info
@@ -58,7 +62,9 @@ import Agda.TypeChecking.Monad.Base  (TCM, NamedMeta(..))
 import Agda.TypeChecking.Monad.Options
 
 import qualified Agda.Utils.AssocList as AssocList
+import Agda.Utils.Either
 import Agda.Utils.Functor
+import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
 import Agda.Utils.Tuple
@@ -139,7 +145,7 @@ lookupName x = do
       Just y  -> return y
       Nothing -> return $ nameConcrete x
 
-lookupQName :: AllowAmbiguousConstructors -> A.QName -> AbsToCon C.QName
+lookupQName :: AllowAmbiguousNames -> A.QName -> AbsToCon C.QName
 lookupQName ambCon x = do
   ys <- inverseScopeLookupName' ambCon x <$> asks currentScope
   lift $ reportSLn "scope.inverse" 100 $
@@ -217,7 +223,7 @@ withInfixDecl :: DefInfo -> C.Name -> AbsToCon [C.Declaration] -> AbsToCon [C.De
 withInfixDecl i x m = do
   ds <- m
   return $ fixDecl ++ synDecl ++ ds
- where fixDecl = [C.Infix (theFixity $ defFixity i) [x] | theFixity (defFixity i) /= defaultFixity]
+ where fixDecl = [C.Infix (theFixity $ defFixity i) [x] | theFixity (defFixity i) /= noFixity]
        synDecl = [C.Syntax x (theNotation (defFixity i))]
 
 {- UNUSED
@@ -287,6 +293,15 @@ instance ToConcrete a c => ToConcrete [a] [c] where
     toConcrete     = mapM toConcrete
     bindToConcrete = thread bindToConcrete
 
+instance (ToConcrete a1 c1, ToConcrete a2 c2) => ToConcrete (Either a1 a2) (Either c1 c2) where
+    toConcrete = traverseEither toConcrete toConcrete
+    bindToConcrete (Left x) ret =
+        bindToConcrete x $ \x ->
+        ret (Left x)
+    bindToConcrete (Right y) ret =
+        bindToConcrete y $ \y ->
+        ret (Right y)
+
 instance (ToConcrete a1 c1, ToConcrete a2 c2) => ToConcrete (a1,a2) (c1,c2) where
     toConcrete (x,y) = liftM2 (,) (toConcrete x) (toConcrete y)
     bindToConcrete (x,y) ret =
@@ -304,19 +319,12 @@ instance (ToConcrete a1 c1, ToConcrete a2 c2, ToConcrete a3 c3) =>
         where
             reorder (x,(y,z)) = (x,y,z)
 
-instance ToConcrete (Common.ArgInfo ac) C.ArgInfo where
-  toConcrete info = -- do cs <- mapM toConcrete $ argInfoColors info
-                    return $ info { argInfoColors = [] } -- TODO: zapping ignoring colours
+instance ToConcrete a c => ToConcrete (Arg a) (Arg c) where
+    toConcrete (Arg i a) = Arg i <$> toConcreteHiding i a
 
-instance ToConcrete a c => ToConcrete (Common.Arg ac a) (C.Arg c) where
-    toConcrete (Common.Arg info x) = Common.Arg
-      <$> toConcrete info
-      <*> toConcreteHiding info x
-
-    bindToConcrete (Common.Arg info x) ret = do
-      info <- toConcrete info
+    bindToConcrete (Arg info x) ret =
       bindToConcreteCtx (hiddenArgumentCtx $ getHiding info) x $
-        ret . Common.Arg info
+        ret . Arg info
 
 instance ToConcrete a c => ToConcrete (WithHiding a) (WithHiding c) where
   toConcrete     (WithHiding h a) = WithHiding h <$> toConcreteHiding h a
@@ -339,7 +347,7 @@ instance ToConcrete A.Name C.Name where
   bindToConcrete x = bindName x
 
 instance ToConcrete A.QName C.QName where
-  toConcrete = lookupQName AllowAmbiguousConstructors
+  toConcrete = lookupQName AmbiguousConstructors
 
 instance ToConcrete A.ModuleName C.QName where
   toConcrete = lookupModule
@@ -350,13 +358,14 @@ instance ToConcrete A.Expr C.Expr where
     toConcrete (Var x)            = Ident . C.QName <$> toConcrete x
     toConcrete (Def x)            = Ident <$> toConcrete x
     toConcrete (Proj x)           = Ident <$> toConcrete x
+    toConcrete (A.Macro x)        = Ident <$> toConcrete x
     toConcrete (Con (AmbQ (x:_))) = Ident <$> toConcrete x
     toConcrete (Con (AmbQ []))    = __IMPOSSIBLE__
         -- for names we have to use the name from the info, since the abstract
         -- name has been resolved to a fully qualified name (except for
         -- variables)
     toConcrete (A.Lit (LitQName r x)) = do
-      x <- lookupQName NoAmbiguousConstructors x
+      x <- lookupQName AmbiguousNothing x
       bracket appBrackets $ return $
         C.App r (C.Quote r) (defaultNamedArg $ C.Ident x)
     toConcrete (A.Lit l)            = return $ C.Lit l
@@ -423,12 +432,12 @@ instance ToConcrete A.Expr C.Expr where
               removeApp p = do
                 lift $ reportSLn "extendedlambda" 50 $ "abstractToConcrete removeApp p = " ++ show p
                 return p -- __IMPOSSIBLE__ -- Andreas, this is actually not impossible, my strictification exposed this sleeping bug
-          let decl2clause (C.FunClause lhs rhs wh) = do
+          let decl2clause (C.FunClause lhs rhs wh ca) = do
                 let p = lhsOriginalPattern lhs
                 lift $ reportSLn "extendedlambda" 50 $ "abstractToConcrete extended lambda pattern p = " ++ show p
                 p' <- removeApp p
                 lift $ reportSLn "extendedlambda" 50 $ "abstractToConcrete extended lambda pattern p' = " ++ show p'
-                return (lhs{ lhsOriginalPattern = p' }, rhs, wh)
+                return (lhs{ lhsOriginalPattern = p' }, rhs, wh, ca)
               decl2clause _ = __IMPOSSIBLE__
           C.ExtendedLam (getRange i) <$> mapM decl2clause decls
     toConcrete (A.Pi _ [] e) = toConcrete e
@@ -454,7 +463,7 @@ instance ToConcrete A.Expr C.Expr where
                            NonStrict  -> addDot a (addDot a e)
                            _          -> e
             addDot a e = Dot (getRange a) e
-            mkArg (Common.Arg info e) = case getHiding info of
+            mkArg (Arg info e) = case getHiding info of
                                           Hidden    -> HiddenArg   (getRange e) (unnamed e)
                                           Instance  -> InstanceArg (getRange e) (unnamed e)
                                           NotHidden -> e
@@ -471,16 +480,11 @@ instance ToConcrete A.Expr C.Expr where
 
     toConcrete (A.Rec i fs) =
       bracket appBrackets $ do
-        let (xs, es) = unzip fs
-        es <- toConcreteTop es
-        return $ C.Rec (getRange i) $ zip xs es
+        C.Rec (getRange i) . map (fmap (\x -> ModuleAssignment x [] defaultImportDir)) <$> toConcreteTop fs
 
     toConcrete (A.RecUpdate i e fs) =
       bracket appBrackets $ do
-        let (xs, es) = unzip fs
-        e <- toConcrete e
-        es <- toConcreteTop es
-        return $ C.RecUpdate (getRange i) e $ zip xs es
+        C.RecUpdate (getRange i) <$> toConcrete e <*> toConcreteTop fs
 
     toConcrete (A.ETel tel) = do
       tel <- concat <$> toConcrete tel
@@ -493,38 +497,38 @@ instance ToConcrete A.Expr C.Expr where
         bindToConcrete x $ \ x' -> do
             e' <- toConcrete e
             return $ C.QuoteGoal (getRange i) x' e'
-    toConcrete (A.QuoteContext i x e) =
-      bracket lamBrackets $
-        bindToConcrete x $ \ x' -> do
-            e' <- toConcrete e
-            return $ C.QuoteContext (getRange i) x' e'
+    toConcrete (A.QuoteContext i) = return $ C.QuoteContext (getRange i)
     toConcrete (A.Quote i) = return $ C.Quote (getRange i)
     toConcrete (A.QuoteTerm i) = return $ C.QuoteTerm (getRange i)
     toConcrete (A.Unquote i) = return $ C.Unquote (getRange i)
+    toConcrete (A.Tactic i e xs ys) = do
+      e' <- toConcrete e
+      xs' <- toConcrete xs
+      ys' <- toConcrete ys
+      let r      = getRange i
+          rawtac = foldl (C.App r) e' xs'
+      return $ C.Tactic (getRange i) rawtac (map namedArg ys')
 
     -- Andreas, 2012-04-02: TODO!  print DontCare as irrAxiom
     -- Andreas, 2010-10-05 print irrelevant things as ordinary things
     toConcrete (A.DontCare e) = C.Dot r . C.Paren r  <$> toConcrete e
        where r = getRange e
---    toConcrete (A.DontCare e) = C.DontCare <$> toConcreteTop e
-{-
-    -- Andreas, 2010-09-21 abuse C.Underscore to print irrelevant things
-    toConcrete (A.DontCare) = return $ C.Underscore noRange Nothing
--}
     toConcrete (A.PatternSyn n) = C.Ident <$> toConcrete n
 
 makeDomainFree :: A.LamBinding -> A.LamBinding
-makeDomainFree b@(A.DomainFull (A.TypedBindings r (Common.Arg info (A.TBind _ [WithHiding h x] t)))) =
+makeDomainFree b@(A.DomainFull (A.TypedBindings r (Arg info (A.TBind _ [WithHiding h x] t)))) =
   case unScope t of
     A.Underscore MetaInfo{metaNumber = Nothing} -> A.DomainFree (mapHiding (mappend h) info) x
     _ -> b
 makeDomainFree b = b
 
+instance ToConcrete a c => ToConcrete (FieldAssignment' a) (FieldAssignment' c) where
+    toConcrete = traverse toConcrete
+
 -- Binder instances -------------------------------------------------------
 
 instance ToConcrete A.LamBinding [C.LamBinding] where
-    bindToConcrete (A.DomainFree info x) ret = do info <- toConcrete info
-                                                  bindToConcrete x $ ret . (:[]) . C.DomainFree info . mkBoundName_
+    bindToConcrete (A.DomainFree info x) ret = bindToConcrete x $ ret . (:[]) . C.DomainFree info . mkBoundName_
     bindToConcrete (A.DomainFull b)      ret = bindToConcrete b $ ret . map C.DomainFull
 
 instance ToConcrete A.TypedBindings [C.TypedBindings] where
@@ -532,7 +536,7 @@ instance ToConcrete A.TypedBindings [C.TypedBindings] where
     bindToConcrete bs $ \cbs ->
     ret (map (C.TypedBindings r) $ recoverLabels bs cbs)
     where
-      recoverLabels :: A.Arg A.TypedBinding -> C.Arg C.TypedBinding -> [C.Arg C.TypedBinding]
+      recoverLabels :: Arg A.TypedBinding -> Arg C.TypedBinding -> [Arg C.TypedBinding]
       recoverLabels b cb
         | getHiding b == NotHidden = [cb]   -- We don't care about labels for explicit args
         | otherwise = traverse (recover (unArg b)) cb
@@ -564,17 +568,16 @@ instance ToConcrete LetBinding [C.Declaration] where
     bindToConcrete (LetBind i info x t e) ret =
         bindToConcrete x $ \x ->
         do (t,(e, [], [], [])) <- toConcrete (t, A.RHS e)
-           info <- toConcrete info
            ret [ C.TypeSig info x t
                , C.FunClause (C.LHS (C.IdentP $ C.QName x) [] [] [])
-                             e C.NoWhere
+                             e C.NoWhere False
                ]
     -- TODO: bind variables
     bindToConcrete (LetPatBind i p e) ret = do
         p <- toConcrete p
         e <- toConcrete e
-        ret [ C.FunClause (C.LHS p [] [] []) (C.RHS e) NoWhere ]
-    bindToConcrete (LetApply i x modapp _ _) ret = do
+        ret [ C.FunClause (C.LHS p [] [] []) (C.RHS e) NoWhere False ]
+    bindToConcrete (LetApply i x modapp _ _ _) ret = do
       x' <- unqualify <$> toConcrete x
       modapp <- toConcrete modapp
       let r = getRange modapp
@@ -583,11 +586,14 @@ instance ToConcrete LetBinding [C.Declaration] where
       -- This is no use since toAbstract LetDefs is in localToAbstract.
       local (openModule' x dir id) $
         ret [ C.ModuleMacro (getRange i) x' modapp open dir ]
-    bindToConcrete (LetOpen i x) ret = do
+    bindToConcrete (LetOpen i x _) ret = do
       x' <- toConcrete x
       let dir = maybe defaultImportDir id $ minfoDirective i
       local (openModule' x dir restrictPrivate) $
             ret [ C.Open (getRange i) x' dir ]
+    bindToConcrete (LetDeclaredVariable _) ret =
+      -- Note that the range of the declaration site is dropped.
+      ret []
 
 data AsWhereDecls = AsWhereDecls [A.Declaration]
 
@@ -602,14 +608,14 @@ instance ToConcrete AsWhereDecls WhereClause where
     ret . AnyWhere =<< declsToConcrete ds
 
 mergeSigAndDef :: [C.Declaration] -> [C.Declaration]
-mergeSigAndDef (C.RecordSig _ x bs e : C.Record r y ind c _ Nothing fs : ds)
-  | x == y = C.Record r y ind c bs (Just e) fs : mergeSigAndDef ds
+mergeSigAndDef (C.RecordSig _ x bs e : C.Record r y ind eta c _ Nothing fs : ds)
+  | x == y = C.Record r y ind eta c bs (Just e) fs : mergeSigAndDef ds
 mergeSigAndDef (C.DataSig _ _ x bs e : C.Data r i y _ Nothing cs : ds)
   | x == y = C.Data r i y bs (Just e) cs : mergeSigAndDef ds
 mergeSigAndDef (d : ds) = d : mergeSigAndDef ds
 mergeSigAndDef [] = []
 
-openModule' :: A.ModuleName -> ImportDirective -> (Scope -> Scope) -> Env -> Env
+openModule' :: A.ModuleName -> C.ImportDirective -> (Scope -> Scope) -> Env -> Env
 openModule' x dir restrict env = env{currentScope = sInfo{scopeModules = mods'}}
   where sInfo = currentScope env
         amod  = scopeCurrent sInfo
@@ -649,37 +655,23 @@ instance ToConcrete (Maybe A.QName) (Maybe C.Name) where
     x' <- toConcrete (qnameName x)
     return $ Just x'
 
-{- UNUSED
--- | Helper function used in instance @ToConcrete Definition@.
-telToTypedBindingss :: [C.LamBinding] -> [C.TypedBindings]
-telToTypedBindingss = map lamBindingToTypedBindings where
-
-  lamBindingToTypedBindings :: C.LamBinding -> C.TypedBindings
-  lamBindingToTypedBindings b =
-    case b of
-      C.DomainFull t     -> t
-      C.DomainFree info n -> C.TypedBindings noRange $
-        Common.Arg info $ C.TBind noRange [n] $ C.Underscore noRange Nothing
--}
-
 instance ToConcrete (Constr A.Constructor) C.Declaration where
   toConcrete (Constr (A.ScopedDecl scope [d])) =
     withScope scope $ toConcrete (Constr d)
   toConcrete (Constr (A.Axiom _ i info x t)) = do
     x' <- unsafeQNameToName <$> toConcrete x
     t' <- toConcreteTop t
-    info <- toConcrete info
     return $ C.TypeSig info x' t'
   toConcrete (Constr d) = head <$> toConcrete d
 
 instance ToConcrete a C.LHS => ToConcrete (A.Clause' a) [C.Declaration] where
-  toConcrete (A.Clause lhs rhs wh) =
+  toConcrete (A.Clause lhs rhs wh catchall) =
       bindToConcrete lhs $ \lhs ->
         case lhs of
           C.LHS p wps _ _ -> do
             bindToConcrete (AsWhereDecls wh)  $ \wh' -> do
                 (rhs', eqs, with, wcs) <- toConcreteTop rhs
-                return $ FunClause (C.LHS p wps eqs with) rhs' wh' : wcs
+                return $ FunClause (C.LHS p wps eqs with) rhs' wh' catchall : wcs
           C.Ellipsis {} -> __IMPOSSIBLE__
           -- TODO: Is the case above impossible? Previously there was
           -- no code for it, but GHC 7's completeness checker spotted
@@ -706,7 +698,6 @@ instance ToConcrete A.Declaration [C.Declaration] where
     withAbstractPrivate i $
       withInfixDecl i x'  $ do
       t' <- toConcreteTop t
-      info <- toConcrete info
       return [C.Postulate (getRange i) [C.TypeSig info x' t']]
 
   toConcrete (A.Field i x t) = do
@@ -714,7 +705,7 @@ instance ToConcrete A.Declaration [C.Declaration] where
     withAbstractPrivate i $
       withInfixDecl i x'  $ do
       t' <- toConcreteTop t
-      return [C.Field x' t']
+      return [C.Field (defInstance i) x' t']
 
   toConcrete (A.Primitive i x t) = do
     x' <- unsafeQNameToName <$> toConcrete x
@@ -747,11 +738,11 @@ instance ToConcrete A.Declaration [C.Declaration] where
       t' <- toConcreteTop t
       return [ C.RecordSig (getRange i) x' (map C.DomainFull $ concat tel') t' ]
 
-  toConcrete (A.RecDef  i x ind c bs t cs) =
+  toConcrete (A.RecDef  i x ind eta c bs t cs) =
     withAbstractPrivate i $
     bindToConcrete (map makeDomainFree bs) $ \tel' -> do
       (x',cs') <- (unsafeQNameToName -*- id) <$> toConcrete (x, map Constr cs)
-      return [ C.Record (getRange i) x' ind Nothing (concat tel') Nothing cs' ]
+      return [ C.Record (getRange i) x' ind eta Nothing (concat tel') Nothing cs' ]
 
   toConcrete (A.Mutual i ds) = declsToConcrete ds
 
@@ -761,7 +752,7 @@ instance ToConcrete A.Declaration [C.Declaration] where
       ds <- declsToConcrete ds
       return [ C.Module (getRange i) x (concat tel) ds ]
 
-  toConcrete (A.Apply i x modapp _ _) = do
+  toConcrete (A.Apply i x modapp _ _ _) = do
     x  <- unsafeQNameToName <$> toConcrete x
     modapp <- toConcrete modapp
     let r = getRange modapp
@@ -769,7 +760,7 @@ instance ToConcrete A.Declaration [C.Declaration] where
         dir  = fromMaybe defaultImportDir{ importDirRange = r } $ minfoDirective i
     return [ C.ModuleMacro (getRange i) x modapp open dir ]
 
-  toConcrete (A.Import i x) = do
+  toConcrete (A.Import i x _) = do
     x <- toConcrete x
     let open = maybe DontOpen id $ minfoOpenShort i
         dir  = maybe defaultImportDir id $ minfoDirective i
@@ -779,17 +770,25 @@ instance ToConcrete A.Declaration [C.Declaration] where
     p <- toConcrete $ RangeAndPragma (getRange i) p
     return [C.Pragma p]
 
-  toConcrete (A.Open i x) = do
+  toConcrete (A.Open i x _) = do
     x <- toConcrete x
     return [C.Open (getRange i) x defaultImportDir]
 
   toConcrete (A.PatternSynDef x xs p) = do
     C.QName x <- toConcrete x
-    bindToConcrete xs $ \xs -> (:[]) . C.PatternSyn (getRange x) x xs <$> toConcrete p
+    bindToConcrete xs $ \xs -> (:[]) . C.PatternSyn (getRange x) x xs <$> toConcrete (vacuous p :: A.Pattern)
 
-  toConcrete (A.UnquoteDecl _ i x e) = do
-    C.QName x <- toConcrete x
-    (:[]) . C.UnquoteDecl (getRange i) x <$> toConcrete e
+  toConcrete (A.UnquoteDecl _ i xs e) = do
+    let unqual (C.QName x) = return x
+        unqual _           = __IMPOSSIBLE__
+    xs <- mapM (unqual <=< toConcrete) xs
+    (:[]) . C.UnquoteDecl (getRange i) xs <$> toConcrete e
+
+  toConcrete (A.UnquoteDef i xs e) = do
+    let unqual (C.QName x) = return x
+        unqual _           = __IMPOSSIBLE__
+    xs <- mapM (unqual <=< toConcrete) xs
+    (:[]) . C.UnquoteDef (getRange i) xs <$> toConcrete e
 
 
 data RangeAndPragma = RangeAndPragma Range A.Pragma
@@ -807,6 +806,9 @@ instance ToConcrete RangeAndPragma C.Pragma where
     A.CompiledDataPragma x hs hcs -> do
       x <- toConcrete x
       return $ C.CompiledDataPragma r x hs hcs
+    A.CompiledDeclareDataPragma x hs -> do
+      x <- toConcrete x
+      return $ C.CompiledDeclareDataPragma r x hs
     A.CompiledPragma x hs -> do
       x <- toConcrete x
       return $ C.CompiledPragma r x hs
@@ -819,28 +821,27 @@ instance ToConcrete RangeAndPragma C.Pragma where
     A.CompiledJSPragma x e -> do
       x <- toConcrete x
       return $ C.CompiledJSPragma r x e
+    A.CompiledUHCPragma x cr -> do
+      x <- toConcrete x
+      return $ C.CompiledUHCPragma r x cr
+    A.CompiledDataUHCPragma x crd crcs -> do
+      x <- toConcrete x
+      return $ C.CompiledDataUHCPragma r x crd crcs
+    A.NoSmashingPragma x -> C.NoSmashingPragma r <$> toConcrete x
     A.StaticPragma x -> C.StaticPragma r <$> toConcrete x
-    A.EtaPragma x    -> C.EtaPragma    r <$> toConcrete x
+    A.InlinePragma x -> C.InlinePragma r <$> toConcrete x
+    A.DisplayPragma f ps rhs ->
+      C.DisplayPragma r <$> toConcrete (A.DefP (PatRange noRange) f ps) <*> toConcrete rhs
 
 -- Left hand sides --------------------------------------------------------
 
-noImplicitArgs :: A.Patterns -> A.Patterns
-noImplicitArgs = filter (noImplicit . namedArg)
-
-noImplicitPats :: [A.Pattern] -> [A.Pattern]
-noImplicitPats = filter noImplicit
-
-noImplicit :: A.Pattern -> Bool
-noImplicit (A.WildP _) = False
-noImplicit _           = True
-
 instance ToConcrete A.SpineLHS C.LHS where
   bindToConcrete lhs = bindToConcrete (A.spineToLhs lhs :: A.LHS)
 
 instance ToConcrete A.LHS C.LHS where
     bindToConcrete (A.LHS i lhscore wps) ret = do
       bindToConcreteCtx TopCtx lhscore $ \lhs ->
-        bindToConcreteCtx TopCtx (noImplicitPats wps) $ \wps ->
+        bindToConcreteCtx TopCtx wps $ \wps ->
           ret $ C.LHS lhs wps [] []
 
 instance ToConcrete A.LHSCore C.Pattern where
@@ -852,34 +853,58 @@ appBrackets' (_:_) ctx = appBrackets ctx
 
 -- TODO: bind variables properly
 instance ToConcrete A.Pattern C.Pattern where
-    toConcrete (VarP x)    = toConcrete x >>= return . IdentP . C.QName
-    toConcrete (A.WildP i)         =
+  toConcrete p =
+    case p of
+      A.VarP x ->
+        C.IdentP . C.QName <$> toConcrete x
+
+      A.WildP i ->
         return $ C.WildP (getRange i)
-    toConcrete (ConP i (AmbQ []) args) = __IMPOSSIBLE__
-    toConcrete p@(ConP i (AmbQ (x:_)) args) =
-      tryToRecoverOpAppP p $
-        bracketP_ (appBrackets' args) $ do
-            x <- toConcrete x
-            args <- toConcreteCtx ArgumentCtx (noImplicitArgs args)
-            return $ foldl AppP (C.IdentP x) args
-    toConcrete p@(DefP i x args) =
-      tryToRecoverOpAppP p $
-        bracketP_ (appBrackets' args) $ do
-            x <- toConcrete x
-            args <- toConcreteCtx ArgumentCtx (noImplicitArgs args)
-            return $ foldl AppP (C.IdentP x) args
-    toConcrete (A.AsP i x p)   = do
-      (x, p) <- toConcreteCtx ArgumentCtx (x,p)
-      return $ C.AsP (getRange i) x p
-    toConcrete (A.AbsurdP i) = return $ C.AbsurdP (getRange i)
-    toConcrete (A.LitP (LitQName r x)) = do
-      x <- lookupQName NoAmbiguousConstructors x
-      bracketP_ appBrackets $ return $ AppP (C.QuoteP r) (defaultNamedArg (C.IdentP x))
-    toConcrete (A.LitP l)    = return $ C.LitP l
-    toConcrete (A.DotP i e)  = do
-        e <- toConcreteCtx DotPatternCtx e
-        return $ C.DotP (getRange i) e
-    toConcrete (A.PatternSynP i n _) = IdentP <$> toConcrete n
+
+      A.ConP i (AmbQ []) args        -> __IMPOSSIBLE__
+      p@(A.ConP i xs@(AmbQ (x:_)) args) -> tryOp x (A.ConP i xs) args
+      p@(A.DefP i x args)               -> tryOp x (A.DefP i x)  args
+
+      A.AsP i x p -> do
+        (x, p) <- toConcreteCtx ArgumentCtx (x,p)
+        return $ C.AsP (getRange i) x p
+
+      A.AbsurdP i ->
+        return $ C.AbsurdP (getRange i)
+
+      A.LitP (LitQName r x) -> do
+        x <- lookupQName AmbiguousNothing x
+        bracketP_ appBrackets $ return $ C.AppP (C.QuoteP r) (defaultNamedArg (C.IdentP x))
+      A.LitP l ->
+        return $ C.LitP l
+
+      A.DotP i e  -> do
+        c <- toConcreteCtx DotPatternCtx e
+        case c of
+          -- Andreas, 2016-02-04 print ._ pattern as _ pattern,
+          -- following the fusing of WildP and ImplicitP.
+          C.Underscore{} -> return $ C.WildP $ getRange i
+          _ -> return $ C.DotP (getRange i) c
+
+      A.PatternSynP i n _ ->
+        C.IdentP <$> toConcrete n
+
+      A.RecP i as ->
+        C.RecP (getRange i) <$> mapM (traverse toConcrete) as
+    where
+    tryOp x f args = do
+      -- Andreas, 2016-02-04, Issue #1792
+      -- To prevent failing of tryToRecoverOpAppP for overapplied operators,
+      -- we take off the exceeding arguments first
+      -- and apply them pointwise with C.AppP later.
+      let (args1, args2) = splitAt (numHoles x) args
+      let funCtx = if null args2 then id else withPrecedence FunctionCtx
+      funCtx (tryToRecoverOpAppP $ f args1) >>= \case
+        Just c  -> applyTo args2 c
+        Nothing -> applyTo args . C.IdentP =<< toConcrete x
+    -- Note: applyTo [] c = return c
+    applyTo args c = bracketP_ (appBrackets' args) $ do
+      foldl C.AppP c <$> toConcreteCtx ArgumentCtx args
 
 
 -- Helpers for recovering C.OpApp ------------------------------------------
@@ -888,10 +913,11 @@ data Hd = HdVar A.Name | HdCon A.QName | HdDef A.QName
 
 cOpApp :: Range -> C.QName -> A.Name -> [C.Expr] -> C.Expr
 cOpApp r x n es =
-  C.OpApp r x (Set.singleton n) (map (defaultNamedArg . Ordinary) es)
+  C.OpApp r x (Set.singleton n)
+          (map (defaultNamedArg . NoPlaceholder . Ordinary) es)
 
 tryToRecoverOpApp :: A.Expr -> AbsToCon C.Expr -> AbsToCon C.Expr
-tryToRecoverOpApp e def = recoverOpApp bracket cOpApp view e def
+tryToRecoverOpApp e def = caseMaybeM (recoverOpApp bracket cOpApp view e) def return
   where
     view e = do
       let Application hd args = AV.appView e
@@ -902,8 +928,8 @@ tryToRecoverOpApp e def = recoverOpApp bracket cOpApp view e def
         Con (AmbQ [])    -> __IMPOSSIBLE__
         _                -> Nothing
 
-tryToRecoverOpAppP :: A.Pattern -> AbsToCon C.Pattern -> AbsToCon C.Pattern
-tryToRecoverOpAppP p def = recoverOpApp bracketP_ opApp view p def
+tryToRecoverOpAppP :: A.Pattern -> AbsToCon (Maybe C.Pattern)
+tryToRecoverOpAppP = recoverOpApp bracketP_ opApp view
   where
     opApp r x n ps =
       C.OpAppP r x (Set.singleton n) (map defaultNamedArg ps)
@@ -916,11 +942,10 @@ tryToRecoverOpAppP p def = recoverOpApp bracketP_ opApp view p def
 recoverOpApp :: (ToConcrete a c, HasRange c)
   => ((Precedence -> Bool) -> AbsToCon c -> AbsToCon c)
   -> (Range -> C.QName -> A.Name -> [c] -> c)
-  -> (a -> Maybe (Hd, [A.NamedArg a]))
+  -> (a -> Maybe (Hd, [NamedArg a]))
   -> a
-  -> AbsToCon c
-  -> AbsToCon c
-recoverOpApp bracket opApp view e mDefault = case view e of
+  -> AbsToCon (Maybe c)
+recoverOpApp bracket opApp view e = case view e of
   Nothing -> mDefault
   Just (hd, args)
     | all notHidden args  -> do
@@ -933,23 +958,21 @@ recoverOpApp bracket opApp view e mDefault = case view e of
         HdCon qn          -> doQNameHelper qnameName id      qn args'
     | otherwise           -> mDefault
   where
+  mDefault = return Nothing
 
   doQNameHelper fixityHelper conHelper n as = do
-    x <- toConcrete n
-    doQName (theFixity $ nameFixity n') (conHelper x) n' as
-    where n' = fixityHelper n
+    x <- conHelper <$> toConcrete n
+    doQName (theFixity $ nameFixity n') x n' as (C.nameParts $ C.unqualify x)
+    where
+      n' = fixityHelper n
 
   -- fall-back (wrong number of arguments or no holes)
-  doQName _ x _ es
-    | length xs == 1        = mDefault
-    | length es /= numHoles = mDefault
-    | null es               = mDefault
-    where
-      xs       = C.nameParts $ C.unqualify x
-      numHoles = length (filter (== Hole) xs)
+  doQName _ x _ es xs
+    | null es                 = mDefault
+    | length es /= numHoles x = mDefault
 
   -- binary case
-  doQName fixity x n as
+  doQName fixity x n as xs
     | Hole <- head xs
     , Hole <- last xs = do
         let a1  = head as
@@ -960,13 +983,12 @@ recoverOpApp bracket opApp view e mDefault = case view e of
         e1 <- toConcreteCtx (LeftOperandCtx fixity) a1
         es <- mapM (toConcreteCtx InsideOperandCtx) as'
         en <- toConcreteCtx (RightOperandCtx fixity) an
-        bracket (opBrackets fixity)
-          $ return $ opApp (getRange (e1, en)) x n ([e1] ++ es ++ [en])
-    where
-      xs = C.nameParts $ C.unqualify x
+        Just <$> do
+          bracket (opBrackets fixity) $
+            return $ opApp (getRange (e1, en)) x n ([e1] ++ es ++ [en])
 
   -- prefix
-  doQName fixity x n as
+  doQName fixity x n as xs
     | Hole <- last xs = do
         let an  = last as
             as' = case as of
@@ -974,28 +996,27 @@ recoverOpApp bracket opApp view e mDefault = case view e of
                     _          -> __IMPOSSIBLE__
         es <- mapM (toConcreteCtx InsideOperandCtx) as'
         en <- toConcreteCtx (RightOperandCtx fixity) an
-        bracket (opBrackets fixity)
-          $ return $ opApp (getRange (n, en)) x n (es ++ [en])
-    where
-      xs = C.nameParts $ C.unqualify x
+        Just <$> do
+          bracket (opBrackets fixity) $
+            return $ opApp (getRange (n, en)) x n (es ++ [en])
 
   -- postfix
-  doQName fixity x n as
+  doQName fixity x n as xs
     | Hole <- head xs = do
         let a1  = head as
             as' = tail as
         e1 <- toConcreteCtx (LeftOperandCtx fixity) a1
         es <- mapM (toConcreteCtx InsideOperandCtx) as'
-        bracket (opBrackets fixity)
-          $ return $ opApp (getRange (e1, n)) x n ([e1] ++ es)
-    where
-      xs = C.nameParts $ C.unqualify x
+        Just <$> do
+          bracket (opBrackets fixity) $
+            return $ opApp (getRange (e1, n)) x n ([e1] ++ es)
 
   -- roundfix
-  doQName _ x n as = do
+  doQName _ x n as xs = do
     es <- mapM (toConcreteCtx InsideOperandCtx) as
-    bracket roundFixBrackets
-      $ return $ opApp (getRange x) x n es
+    Just <$> do
+      bracket roundFixBrackets $
+        return $ opApp (getRange x) x n es
 
 -- Some instances that are related to interaction with users -----------
 
