@@ -2,6 +2,7 @@
 {-# LANGUAGE FlexibleContexts       #-}
 {-# LANGUAGE FlexibleInstances      #-}
 {-# LANGUAGE FunctionalDependencies #-}
+{-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE MultiParamTypeClasses  #-}
 {-# LANGUAGE ScopedTypeVariables    #-}
 {-# LANGUAGE TupleSections          #-}
@@ -26,7 +27,6 @@ module Agda.Syntax.Translation.InternalToAbstract
 
 import Prelude hiding (mapM_, mapM)
 import Control.Applicative
-import Control.Arrow
 import Control.Monad.State hiding (mapM_, mapM)
 import Control.Monad.Reader hiding (mapM_, mapM)
 
@@ -41,14 +41,16 @@ import Data.Traversable as Trav
 
 import Agda.Syntax.Literal
 import Agda.Syntax.Position
-import Agda.Syntax.Common hiding (Arg, Dom, NamedArg, ArgInfo)
+import Agda.Syntax.Common
 import Agda.Syntax.Fixity
-import qualified Agda.Syntax.Common as Common
+import Agda.Syntax.Concrete (FieldAssignment'(..), exprFieldA)
 import Agda.Syntax.Info as Info
 import Agda.Syntax.Abstract as A
 import Agda.Syntax.Internal as I
+import Agda.Syntax.Internal.Pattern as I
+import Agda.Syntax.Scope.Base (isNameInScope, inverseScopeLookupName)
 
-import Agda.TypeChecking.Monad as M hiding (MetaInfo)
+import Agda.TypeChecking.Monad as M hiding (MetaInfo, tick)
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Reduce
 import {-# SOURCE #-} Agda.TypeChecking.Records
@@ -62,6 +64,7 @@ import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.DropArgs
 
 import Agda.Utils.Except ( MonadError(catchError) )
+import Agda.Utils.Lens
 import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
@@ -74,37 +77,30 @@ import Agda.Utils.Impossible
 
 -- Composition of reified applications ------------------------------------
 
-napps :: Expr -> [I.NamedArg Expr] -> TCM Expr
+napps :: Expr -> [NamedArg Expr] -> TCM Expr
 napps e args = do
   dontShowImp <- not <$> showImplicitArguments
   let apply1 e arg | notVisible arg && dontShowImp = e
                    | otherwise = App exprInfo e arg
   foldl' apply1 e <$> reify args
 
-apps :: Expr -> [I.Arg Expr] -> TCM Expr
+apps :: Expr -> [Arg Expr] -> TCM Expr
 apps e args = napps e $ map (fmap unnamed) args
 
-reifyApp :: Expr -> [I.Arg Term] -> TCM Expr
+reifyApp :: Expr -> [Arg Term] -> TCM Expr
 reifyApp e vs = apps e =<< reifyIArgs vs
 
-reifyIArg :: Reify i a => I.Arg i -> TCM (I.Arg a)
-reifyIArg i = Common.Arg (argInfo i) <$> reify (unArg i)
+reifyIArg :: Reify i a => Arg i -> TCM (Arg a)
+reifyIArg i = Arg (argInfo i) <$> reify (unArg i)
 
-reifyIArgs :: Reify i a => [I.Arg i] -> TCM [I.Arg a]
+reifyIArgs :: Reify i a => [Arg i] -> TCM [Arg a]
 reifyIArgs = mapM reifyIArg
 
-reifyIArg' :: I.Arg e -> TCM (A.Arg e)
-reifyIArg' e = flip Common.Arg (unArg e) <$> reify (argInfo e)
-
-reifyIArgs' :: [I.Arg e] -> TCM [A.Arg e]
-reifyIArgs' = mapM reifyIArg'
-
 -- Composition of reified eliminations ------------------------------------
 
 elims :: Expr -> [I.Elim' Expr] -> TCM Expr
 elims e [] = return e
-elims e (I.Apply arg : es) = do
-  arg <- reifyIArg' arg
+elims e (I.Apply arg : es) =
   elims (A.App exprInfo e $ fmap unnamed arg) es
 elims e (I.Proj d    : es) = elims (A.App exprInfo (A.Proj d) $ defaultNamedArg e) es
 
@@ -201,8 +197,7 @@ reifyDisplayFormP lhs@(A.SpineLHS i f ps wps) =
     -- But apparently, it has no influence...
     -- Ulf, can you add an explanation?
     md <- liftTCM $ -- addContext (replicate (length ps) "x") $
-      displayForm f vs `catchError` \_ -> return Nothing
-        -- unquoted extended lambdas use fake names, so catch errors here
+      displayForm f vs
     reportSLn "reify.display" 20 $
       "display form of " ++ show f ++ " " ++ show ps ++ " " ++ show wps ++ ":\n  " ++ show md
     case md of
@@ -216,32 +211,36 @@ reifyDisplayFormP lhs@(A.SpineLHS i f ps wps) =
   where
     -- Andreas, 2015-05-03: Ulf, please comment on what
     -- is the idea behind okDisplayForm.
-    okDisplayForm (DWithApp d ds []) =
-      okDisplayForm d && all okDisplayTerm ds
+    -- Ulf, 2016-04-15: okDisplayForm should return True if the display form
+    -- can serve as a valid left-hand side. That means checking that it is a
+    -- defined name applied to valid lhs eliminators (projections or
+    -- applications to constructor patterns).
+    okDisplayForm (DWithApp d ds [])   = okDisplayForm d && all okDisplayTerm ds
     okDisplayForm (DTerm (I.Def f vs)) = all okElim vs
-    okDisplayForm (DDef f es) = all okDElim es
-    okDisplayForm DDot{} = False
-    okDisplayForm DCon{} = False
-    okDisplayForm DTerm{} = True -- False?
-    okDisplayForm DWithApp{} = True -- False?
+    okDisplayForm (DDef f es)          = all okDElim es
+    okDisplayForm DDot{}               = False
+    okDisplayForm DCon{}               = False
+    okDisplayForm DTerm{}              = False
+    okDisplayForm DWithApp{}           = False
 
     okDisplayTerm (DTerm v) = okTerm v
-    okDisplayTerm DDot{} = True
-    okDisplayTerm DCon{} = True
-    okDisplayTerm DDef{} = False
-    okDisplayTerm _ = False
+    okDisplayTerm DDot{}    = True
+    okDisplayTerm DCon{}    = True
+    okDisplayTerm DDef{}    = False
+    okDisplayTerm _         = False
 
     okDElim (I.Apply v) = okDisplayTerm $ unArg v
-    okDElim I.Proj{}    = True  -- True, man, or False?  No clue what I am implementing here --Andreas, 2015-05-03
+    okDElim I.Proj{}    = True
+
     okArg = okTerm . unArg
 
     okElim (I.Apply a) = okArg a
-    okElim (I.Proj{})  = False
+    okElim (I.Proj{})  = True
 
     okTerm (I.Var _ []) = True
     okTerm (I.Con c vs) = all okArg vs
     okTerm (I.Def x []) = isNoName $ qnameToConcrete x -- Handling wildcards in display forms
-    okTerm _            = True -- False
+    okTerm _            = False
 
     -- Flatten a dt into (parentName, parentElims, withArgs).
     flattenWith :: DisplayTerm -> (QName, [I.Elim' DisplayTerm], [DisplayTerm])
@@ -256,7 +255,6 @@ reifyDisplayFormP lhs@(A.SpineLHS i f ps wps) =
       (f, vs, ds) -> do
         ds <- mapM termToPat ds
         vs <- mapM elimToPat vs
-        vs <- reifyIArgs' vs
         return $ SpineLHS i f vs (ds ++ wps)
       where
         ci   = ConPatInfo ConPCon patNoRange
@@ -271,11 +269,11 @@ reifyDisplayFormP lhs@(A.SpineLHS i f ps wps) =
                                            Nothing -> __IMPOSSIBLE__
                                            Just p  -> p
 
-        termToPat (DCon c vs)          = A.ConP ci (AmbQ [conName c]) <$> do
-          mapM argToPat =<< reifyIArgs' vs
+        termToPat (DCon c vs)          = tryRecPFromConP =<< do
+           A.ConP ci (AmbQ [conName c]) <$> mapM argToPat vs
 
-        termToPat (DTerm (I.Con c vs)) = A.ConP ci (AmbQ [conName c]) <$> do
-          mapM (argToPat . fmap DTerm) =<< reifyIArgs' vs
+        termToPat (DTerm (I.Con c vs)) = tryRecPFromConP =<< do
+           A.ConP ci (AmbQ [conName c]) <$> mapM (argToPat . fmap DTerm) vs
 
         termToPat (DTerm (I.Def _ [])) = return $ A.WildP patNoRange
         termToPat (DDef _ [])          = return $ A.WildP patNoRange
@@ -312,11 +310,7 @@ reifyDisplayFormP lhs@(A.SpineLHS i f ps wps) =
 
 instance Reify Literal Expr where
   reifyWhen = reifyWhenE
-  reify l@(LitInt    {}) = return (A.Lit l)
-  reify l@(LitFloat  {}) = return (A.Lit l)
-  reify l@(LitString {}) = return (A.Lit l)
-  reify l@(LitChar   {}) = return (A.Lit l)
-  reify l@(LitQName  {}) = return (A.Lit l)
+  reify l = return (A.Lit l)
 
 instance Reify Term Expr where
   reifyWhen = reifyWhenE
@@ -329,7 +323,6 @@ reifyTerm expandAnonDefs0 v = do
   expandAnonDefs <- return expandAnonDefs0 `and2M` displayFormsEnabled
   v <- unSpine <$> instantiate v
   case v of
-    _ | isHackReifyToMeta v -> return $ A.Underscore emptyMetaInfo
     I.Var n es   -> do
         let vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
         x  <- liftTCM $ nameOfBV n `catchError` \_ -> freshName_ ("@" ++ show n)
@@ -347,7 +340,7 @@ reifyTerm expandAnonDefs0 v = do
           r  <- getConstructorData x
           xs <- getRecordFieldNames r
           vs <- map unArg <$> reifyIArgs vs
-          return $ A.Rec exprInfo $ map (unArg *** id) $ filter keep $ zip xs vs
+          return $ A.Rec exprInfo $ map (Left . uncurry FieldAssignment . mapFst unArg) $ filter keep $ zip xs vs
         False -> reifyDisplayForm x vs $ do
           ci <- getConstInfo x
           let Constructor{conPars = np} = theDef ci
@@ -370,10 +363,7 @@ reifyTerm expandAnonDefs0 v = do
             -- (see for example the parameter {i} to Data.Star.Star, which is also
             -- the first argument to the cons).
             -- @data Star {i}{I : Set i} ... where cons : {i :  I} ...@
-            -- Ulf, 2014-07-19: Don't do any of this if we're reifying an
-            -- unquoted term (issue 1237).
-            unquote <- isReifyingUnquoted
-            if np == 0 || unquote then apps h es else do
+            if np == 0 then apps h es else do
               -- Get name of first argument from type of constructor.
               -- Here, we need the reducing version of @telView@
               -- because target of constructor could be a definition
@@ -383,9 +373,9 @@ reifyTerm expandAnonDefs0 v = do
                 -- Andreas, 2012-09-18
                 -- If the first regular constructor argument is hidden,
                 -- we keep the parameters to avoid confusion.
-                (Common.Dom info _ : _) | isHidden info -> do
+                (Dom info _ : _) | isHidden info -> do
                   let us = genericReplicate (np - n) $
-                             setRelevance Relevant $ Common.Arg info underscore
+                             setRelevance Relevant $ Arg info underscore
                   apps h $ us ++ es
                 -- otherwise, we drop all parameters
                 _ -> apps h es
@@ -403,7 +393,6 @@ reifyTerm expandAnonDefs0 v = do
 --    I.Lam info b | isAbsurdBody b -> return $ A.AbsurdLam exprInfo $ getHiding info
     I.Lam info b    -> do
       (x,e) <- reify b
-      info <- reify info
       return $ A.Lam exprInfo (DomainFree info x) e
       -- Andreas, 2011-04-07 we do not need relevance information at internal Lambda
     I.Lit l        -> reify l
@@ -418,12 +407,12 @@ reifyTerm expandAnonDefs0 v = do
           | otherwise   -> mkPi b =<< reify a
         b               -> mkPi b =<< do
           ifM (domainFree a (absBody b))
-            {- then -} (Common.Arg <$> reify (domInfo a) <*> pure underscore)
+            {- then -} (pure $ Arg (domInfo a) underscore)
             {- else -} (reify a)
       where
-        mkPi b (Common.Arg info a) = do
+        mkPi b (Arg info a) = do
           (x, b) <- reify b
-          return $ A.Pi exprInfo [TypedBindings noRange $ Common.Arg info (TBind noRange [pure x] a)] b
+          return $ A.Pi exprInfo [TypedBindings noRange $ Arg info (TBind noRange [pure x] a)] b
         -- We can omit the domain type if it doesn't have any free variables
         -- and it's mentioned in the target type.
         domainFree a b = do
@@ -437,16 +426,14 @@ reifyTerm expandAnonDefs0 v = do
       apps x' =<< reifyIArgs vs
     I.DontCare v -> A.DontCare <$> reifyTerm expandAnonDefs v
     I.Shared p   -> reifyTerm expandAnonDefs $ derefPtr p
-    I.ExtLam cls args -> do
-      x <- freshName_ "extlam"
-      reifyExtLam (qnameFromList [x]) 0 cls (map (fmap unnamed) args)
   where
-    -- Andreas, 2012-10-20  expand a copy in an anonymous module
+    -- Andreas, 2012-10-20  expand a copy if not in scope
     -- to improve error messages.
     -- Don't do this if we have just expanded into a display form,
     -- otherwise we loop!
     reifyDef :: Bool -> QName -> I.Args -> TCM Expr
-    reifyDef True x@(QName m name) vs | A.isAnonymousModuleName m = do
+    reifyDef True x vs =
+      ifM (not . null . inverseScopeLookupName x <$> getScope) (reifyDef' x vs) $ do
       r <- reduceDefCopy x vs
       case r of
         YesReduction _ v -> do
@@ -480,7 +467,7 @@ reifyTerm expandAnonDefs0 v = do
                   apps (A.AbsurdLam exprInfo h) =<< reifyIArgs vs
               _ -> cont
       reifyAbsurdLambda $ do
-        (pad, vs :: [I.NamedArg Term]) <- do
+        (pad, vs :: [NamedArg Term]) <- do
           case mdefn of
             Nothing   -> return ([], map (fmap unnamed) $ genericDrop n vs)
             Just defn -> do
@@ -504,8 +491,7 @@ reifyTerm expandAnonDefs0 v = do
                         return (np, map (argFromDom . (fmap $ const whocares)) as, dom)
                       _ -> return (0, [], __IMPOSSIBLE__)
               -- Now pad' ++ vs' = drop n (pad ++ vs)
-              pad' <- reifyIArgs' $ genericDrop n pad
-              let vs'  :: [I.Arg Term]
+              let pad' = genericDrop n pad
                   vs'  = genericDrop (max 0 (n - size pad)) vs
               -- Andreas, 2012-04-21: get rid of hidden underscores {_}
               -- Keep non-hidden arguments of the padding
@@ -528,21 +514,21 @@ reifyTerm expandAnonDefs0 v = do
            let apps = foldl' (\e a -> A.App exprInfo e (fmap unnamed a))
            napps (A.Def x `apps` pad) =<< reifyIArgs vs
 
-    reifyExtLam :: QName -> Int -> [I.Clause] -> [I.NamedArg Term] -> TCM Expr
+    reifyExtLam :: QName -> Int -> [I.Clause] -> [NamedArg Term] -> TCM Expr
     reifyExtLam x n cls vs = do
       reportSLn "reify.def" 10 $ "reifying extended lambda with definition: x = " ++ show x
       -- drop lambda lifted arguments
       cls <- mapM (reify . QNamed x . dropArgs n) $ cls
       let cx    = nameConcrete $ qnameName x
-          dInfo = mkDefInfo cx defaultFixity' PublicAccess ConcreteDef (getRange x)
+          dInfo = mkDefInfo cx noFixity' PublicAccess ConcreteDef (getRange x)
       napps (A.ExtendedLam exprInfo dInfo x cls) =<< reifyIArgs vs
 
 -- | @nameFirstIfHidden n (a1->...an->{x:a}->b) ({e} es) = {x = e} es@
-nameFirstIfHidden :: [I.Dom (ArgName, t)] -> [I.Arg a] -> [I.NamedArg a]
+nameFirstIfHidden :: [Dom (ArgName, t)] -> [Arg a] -> [NamedArg a]
 nameFirstIfHidden _         []                    = []
 nameFirstIfHidden []        (_ : _)               = __IMPOSSIBLE__
-nameFirstIfHidden (dom : _) (Common.Arg info e : es) | isHidden info =
-  Common.Arg info (Named (Just $ unranged $ fst $ unDom dom) e) :
+nameFirstIfHidden (dom : _) (Arg info e : es) | isHidden info =
+  Arg info (Named (Just $ unranged $ fst $ unDom dom) e) :
   map (fmap unnamed) es
 nameFirstIfHidden _         es                    = map (fmap unnamed) es
 
@@ -551,21 +537,19 @@ instance Reify i a => Reify (Named n i) (Named n a) where
   reifyWhen b = traverse (reifyWhen b)
 
 -- | Skip reification of implicit and irrelevant args if option is off.
-instance (Reify i a) => Reify (I.Arg i) (A.Arg a) where
-  reify (Common.Arg info i) = liftM2 Common.Arg (reify info)
-                                                (flip reifyWhen i =<< condition)
+instance (Reify i a) => Reify (Arg i) (Arg a) where
+  reify (Arg info i) = Arg info <$> (flip reifyWhen i =<< condition)
     where condition = (return (argInfoHiding info /= Hidden) `or2M` showImplicitArguments)
               `and2M` (return (argInfoRelevance info /= Irrelevant) `or2M` showIrrelevantArguments)
-  reifyWhen b i = do info <- reify $ argInfo i
-                     traverse (reifyWhen b) $ i { argInfo = info }
+  reifyWhen b i = traverse (reifyWhen b) i
 
 instance Reify Elim Expr where
   reifyWhen = reifyWhenE
   reify e = case e of
     I.Apply v -> appl "apply" <$> reify v
-    I.Proj f  -> appl "proj"  <$> reify ((defaultArg $ I.Def f []) :: I.Arg Term)
+    I.Proj f  -> appl "proj"  <$> reify ((defaultArg $ I.Def f []) :: Arg Term)
     where
-      appl :: String -> A.Arg Expr -> Expr
+      appl :: String -> Arg Expr -> Expr
       appl s v = A.App exprInfo (A.Lit (LitString noRange s)) $ fmap unnamed v
 
 type NamedClause = QNamed I.Clause
@@ -591,7 +575,7 @@ instance (Ord k, Monoid v) => Monoid (MonoidMap k v) where
 -- | Move dots on variables so that each variable is bound at its first
 --   non-hidden occurrence (if any). If all occurrences are hidden it's bound
 --   at the first occurrence.
-shuffleDots :: ([A.NamedArg A.Pattern], [A.Pattern]) -> TCM ([A.NamedArg A.Pattern], [A.Pattern])
+shuffleDots :: ([NamedArg A.Pattern], [A.Pattern]) -> TCM ([NamedArg A.Pattern], [A.Pattern])
 shuffleDots (ps, wps) = do
   return $ (`evalState` xs)
          $ (`runReaderT` NotHidden)
@@ -622,6 +606,7 @@ shuffleDots (ps, wps) = do
       A.AbsurdP{}          -> mempty
       A.LitP{}             -> mempty
       A.AsP{}              -> __IMPOSSIBLE__
+      A.RecP _ as          -> foldMap (foldMap (patVars h)) as
 
     shouldBind x = do
       xs <- get
@@ -652,6 +637,7 @@ shuffleDots (ps, wps) = do
       A.AbsurdP{}          -> pure p
       A.LitP{}             -> pure p
       A.AsP{}              -> __IMPOSSIBLE__
+      A.RecP i as          -> A.RecP i <$> traverse (traverse redotPat) as
 
     redotVar p x = do
       b <- shouldBind x
@@ -662,8 +648,8 @@ shuffleDots (ps, wps) = do
 
 -- | Removes implicit arguments that are not needed, that is, that don't bind
 --   any variables that are actually used and doesn't do pattern matching.
-stripImplicits :: ([A.NamedArg A.Pattern], [A.Pattern]) ->
-                  TCM ([A.NamedArg A.Pattern], [A.Pattern])
+stripImplicits :: ([NamedArg A.Pattern], [A.Pattern]) ->
+                  TCM ([NamedArg A.Pattern], [A.Pattern])
 stripImplicits (ps, wps) = do          -- v if show-implicit we don't need the names
   ifM showImplicitArguments (return (map (unnamed . namedThing <$>) ps, wps)) $ do
     let vars = dotVars (ps, wps)
@@ -694,6 +680,7 @@ stripImplicits (ps, wps) = do          -- v if show-implicit we don't need the n
         A.LitP _      -> Set.empty
         A.AsP _ _ p   -> patVars p
         A.PatternSynP _ _ _ -> __IMPOSSIBLE__ -- Set.empty
+        A.RecP _ as -> foldMap (foldMap patVars) as
 
       -- Replace dot variables by ._ if they use implicitly bound variables. This
       -- is slightly nicer than making the implicts explicit.
@@ -744,6 +731,7 @@ stripImplicits (ps, wps) = do          -- v if show-implicit we don't need the n
             A.LitP _      -> p
             A.AsP i x p   -> A.AsP i x $ stripPat p
             A.PatternSynP _ _ _ -> __IMPOSSIBLE__ -- p
+            A.RecP i fs   -> A.RecP i $ map (fmap stripPat) fs  -- TODO Andreas: is this right?
 
           noInterestingBindings p =
             Set.null $ dvs `Set.intersection` patVars p
@@ -762,7 +750,7 @@ class DotVars a where
   isConPat :: a -> Bool
   isConPat _ = False
 
-instance DotVars a => DotVars (A.Arg a) where
+instance DotVars a => DotVars (Arg a) where
   dotVars a = if notVisible a && not (isConPat a)   -- Hidden constructor patterns are visible!
               then Set.empty
               else dotVars (unArg a)
@@ -778,9 +766,12 @@ instance DotVars a => DotVars [a] where
 instance (DotVars a, DotVars b) => DotVars (a, b) where
   dotVars (x, y) = Set.union (dotVars x) (dotVars y)
 
+instance (DotVars a, DotVars b) => DotVars (Either a b) where
+  dotVars = either dotVars dotVars
+
 instance DotVars A.Clause where
-  dotVars (A.Clause _ rhs []) = dotVars rhs
-  dotVars (A.Clause _ rhs (_:_)) = __IMPOSSIBLE__ -- cannot contain where clauses?
+  dotVars (A.Clause _ rhs [] _) = dotVars rhs
+  dotVars (A.Clause _ rhs (_:_) _) = __IMPOSSIBLE__ -- cannot contain where clauses?
 
 instance DotVars A.Pattern where
   dotVars p = case p of
@@ -793,6 +784,8 @@ instance DotVars A.Pattern where
     A.LitP _      -> Set.empty
     A.AsP _ _ p   -> dotVars p
     A.PatternSynP _ _ _ -> __IMPOSSIBLE__ -- Set.empty
+    A.RecP _ fs   -> dotVars fs
+
   isConPat A.ConP{} = True
   isConPat A.LitP{} = True
   isConPat _        = False
@@ -820,16 +813,24 @@ instance DotVars A.Expr where
     A.Set _ _              -> Set.empty
     A.Prop _               -> Set.empty
     A.Let _ _ _            -> __IMPOSSIBLE__
-    A.Rec _ es             -> dotVars $ map snd es
-    A.RecUpdate _ e es     -> dotVars (e, map snd es)
+    A.Rec _ es             -> dotVars es
+    A.RecUpdate _ e es     -> dotVars (e, es)
     A.ETel _               -> __IMPOSSIBLE__
     A.QuoteGoal {}         -> __IMPOSSIBLE__
     A.QuoteContext {}      -> __IMPOSSIBLE__
     A.Quote {}             -> __IMPOSSIBLE__
     A.QuoteTerm {}         -> __IMPOSSIBLE__
     A.Unquote {}           -> __IMPOSSIBLE__
+    A.Tactic {}            -> __IMPOSSIBLE__
     A.DontCare v           -> dotVars v
-    A.PatternSyn n         -> Set.empty
+    A.PatternSyn {}        -> Set.empty
+    A.Macro {}             -> Set.empty
+
+instance DotVars a => DotVars (FieldAssignment' a) where
+  dotVars a = dotVars (a ^. exprFieldA)
+
+instance DotVars A.ModuleName where
+  dotVars _ = Set.empty
 
 instance DotVars RHS where
   dotVars (RHS e) = dotVars e
@@ -844,48 +845,89 @@ instance DotVars TypedBinding where
   dotVars (TBind _ _ e) = dotVars e
   dotVars (TLet _ _)    = __IMPOSSIBLE__ -- Since the internal syntax has no let bindings left
 
+class MonadTick m where
+  tick :: m Nat
+
+newtype TickT m a = TickT { unTickT :: StateT Nat m a }
+  deriving (Functor, Applicative, Monad, MonadReader r, MonadTrans, MonadIO) -- MonadExcept e,
+
+instance Monad m => MonadTick (TickT m) where
+  tick = TickT $ do i <- get; put (i + 1); return i
+
+instance MonadState s m => MonadState s (TickT m) where
+  state f = lift $ state f -- TickT $ StateT $ \ n -> (,n) <$> state f
+
+instance MonadTCM tcm => MonadTCM (TickT tcm) where
+  liftTCM = lift . liftTCM
+
+runTickT :: Monad m => TickT m a -> m a
+runTickT m = evalStateT (unTickT m) 0
 
 -- TODO: implement reifyPatterns on de Bruijn patterns ( numberPatVars )
-reifyPatterns :: I.Telescope -> Permutation -> [I.NamedArg I.Pattern] -> TCM [A.NamedArg A.Pattern]
-reifyPatterns tel perm ps = evalStateT (reifyArgs ps) 0
+reifyPatterns :: I.Telescope -> Permutation -> [NamedArg I.Pattern] -> TCM [NamedArg A.Pattern]
+reifyPatterns tel perm ps = runTickT (reifyArgs ps)
   where
-    reifyArgs :: [I.NamedArg I.Pattern] -> StateT Nat TCM [A.NamedArg A.Pattern]
+    reifyArgs :: [NamedArg I.Pattern] -> TickT TCM [NamedArg A.Pattern]
     reifyArgs is = mapM reifyArg is
 
-    reifyArg :: I.NamedArg I.Pattern -> StateT Nat TCM (A.NamedArg A.Pattern)
+    reifyArg :: NamedArg I.Pattern -> TickT TCM (NamedArg A.Pattern)
     reifyArg i = stripNameFromExplicit <$>
-                 traverse (traverse reifyPat) (setArgColors [] i) -- TODO guilhem
+                 traverse (traverse reifyPat) i
 
     stripNameFromExplicit a
       | getHiding a == NotHidden = fmap (unnamed . namedThing) a
       | otherwise                = a
 
-    tick = do i <- get; put (i + 1); return i
-
     translate n = fromMaybe __IMPOSSIBLE__ $ vars !!! n
       where
         vars = permPicks $ invertP __IMPOSSIBLE__ perm
 
-    reifyPat :: I.Pattern -> StateT Nat TCM A.Pattern
-    reifyPat p = case p of
+    reifyPat :: I.Pattern -> TickT TCM A.Pattern
+    reifyPat p = do
+     liftTCM $ reportSLn "reify.pat" 80 $ "reifying pattern " ++ show p
+     case p of
       I.VarP "()" -> A.AbsurdP patNoRange <$ tick   -- HACK
       I.VarP s -> do
         i <- tick
         let j = translate i
-        lift $ A.VarP <$> nameOfBV (size tel - 1 - j)
+        liftTCM $ A.VarP <$> nameOfBV (size tel - 1 - j)
       I.DotP v -> do
-        t <- lift $ reify v
+        t <- liftTCM $ reify v
         _ <- tick
         let vars = Set.map show (dotVars t)
             t'   = if Set.member "()" vars then underscore else t
         return $ A.DotP patNoRange t'
       I.LitP l  -> return $ A.LitP l
       I.ProjP d -> return $ A.DefP patNoRange d []
-      I.ConP c cpi ps -> A.ConP ci (AmbQ [conName c]) <$> reifyArgs ps
-        where ci = flip ConPatInfo patNoRange $ fromMaybe ConPCon $ I.conPRecord cpi
+      I.ConP c cpi ps -> do
+        liftTCM $ reportSLn "reify.pat" 60 $ "reifying pattern " ++ show p
+        tryRecPFromConP =<< do A.ConP ci (AmbQ [conName c]) <$> reifyArgs ps
+        where
+          origin = fromMaybe ConPCon $ I.conPRecord cpi
+          ci = ConPatInfo origin patNoRange
+
+-- | If the record constructor is generated or the user wrote a record pattern,
+--   turn constructor pattern into record pattern.
+--   Otherwise, keep constructor pattern.
+tryRecPFromConP :: MonadTCM tcm => A.Pattern -> tcm A.Pattern
+tryRecPFromConP p = do
+  let fallback = return p
+  case p of
+    A.ConP ci (AmbQ [c]) ps -> do
+        caseMaybeM (liftTCM $ isRecordConstructor c) fallback $ \ (r, def) -> do
+          -- If the record constructor is generated or the user wrote a record pattern,
+          -- print record pattern.
+          -- Otherwise, print constructor pattern.
+          if recNamedCon def && patOrigin ci /= ConPRec then fallback else do
+            fs <- liftTCM $ getRecordFieldNames r
+            unless (length fs == length ps) __IMPOSSIBLE__
+            return $ A.RecP patNoRange $ zipWith mkFA fs ps
+        where
+          mkFA ax nap = FieldAssignment (unArg ax) (namedArg nap)
+    _ -> __IMPOSSIBLE__
 
 instance Reify NamedClause A.Clause where
-  reify (QNamed f (I.Clause _ tel perm ps body _)) = addCtxTel tel $ do
+  reify (QNamed f (I.Clause _ tel ps' body _ catchall)) = addCtxTel tel $ do
     ps  <- reifyPatterns tel perm ps
     lhs <- liftTCM $ reifyDisplayFormP $ SpineLHS info f ps [] -- LHS info (LHSHead f ps) []
     nfv <- getDefFreeVars f `catchError` \_ -> return 0
@@ -893,11 +935,13 @@ instance Reify NamedClause A.Clause where
     reportSLn "reify.clause" 60 $ "reifying NamedClause, lhs = " ++ show lhs
     rhs <- reify $ renameP (reverseP perm) <$> body
     reportSLn "reify.clause" 60 $ "reifying NamedClause, rhs = " ++ show rhs
-    let result = A.Clause (spineToLhs lhs) rhs []
+    let result = A.Clause (spineToLhs lhs) rhs [] catchall
     reportSLn "reify.clause" 60 $ "reified NamedClause, result = " ++ show result
     return result
     where
       info = LHSRange noRange
+      ps   = unnumberPatVars ps'
+      perm = dbPatPerm ps'
 
       dropParams n (SpineLHS i f ps wps) = SpineLHS i f (genericDrop n ps) wps
       stripImps (SpineLHS i f ps wps) = do
@@ -949,16 +993,13 @@ instance (Free i, Reify i a) => Reify (Abs i) (Name, a) where
 instance Reify I.Telescope A.Telescope where
   reify EmptyTel = return []
   reify (ExtendTel arg tel) = do
-    Common.Arg info e <- reify arg
+    Arg info e <- reify arg
     (x,bs)  <- reify tel
     let r = getRange e
-    return $ TypedBindings r (Common.Arg info (TBind r [pure x] e)) : bs
+    return $ TypedBindings r (Arg info (TBind r [pure x] e)) : bs
 
-instance Reify I.ArgInfo A.ArgInfo where
-    reify i = flip (mapArgInfoColors.const) i <$> reify (argInfoColors i)
-
-instance Reify i a => Reify (I.Dom i) (A.Arg a) where
-    reify (Common.Dom info i) = liftM2 Common.Arg (reify info) (reify i)
+instance Reify i a => Reify (Dom i) (Arg a) where
+    reify (Dom info i) = Arg info <$> reify i
 
 instance Reify i a => Reify [i] [a] where
     reify = traverse reify
