@@ -13,8 +13,6 @@ import Agda.Syntax.Parser.Parser              as SyntPars   (tests)
 import Agda.Syntax.Position                   as SyntPosi   (tests)
 import Agda.Termination.CallGraph             as TermCall   (tests)
 import Agda.Termination.CallMatrix            as TermCM     (tests)
--- import Agda.Termination.Lexicographic         as TermLex    (tests)
--- import Agda.Termination.Matrix                as TermMatrix (tests)
 import Agda.Termination.Order                 as TermOrd    (tests)
 import Agda.Termination.Semiring              as TermRing   (tests)
 import Agda.Termination.SparseMatrix          as TermSparse (tests)
@@ -36,6 +34,7 @@ import Agda.Utils.List                        as UtilList   (tests)
 import Agda.Utils.ListT.Tests                 as UtilListT  (tests)
 import Agda.Utils.PartialOrd                  as UtilPOrd   (tests)
 import Agda.Utils.Permutation.Tests           as UtilPerm   (tests)
+import Agda.Utils.Trie                        as UtilTrie   (tests)
 import Agda.Utils.Warshall                    as UtilWarsh  (tests)
 
 testSuite :: IO Bool
@@ -57,8 +56,6 @@ testSuite = runTests "QuickCheck test suite:"
   , SyntPosi.tests
   , TermCall.tests
   , TermCM.tests
---  , TermLex.tests
---  , TermMatrix.tests
   , TermOrd.tests
   , TermRing.tests
   , TermSparse.tests
@@ -73,5 +70,6 @@ testSuite = runTests "QuickCheck test suite:"
   , UtilFile.tests
   , UtilGraphUni.tests
   , UtilList.tests
+  , UtilTrie.tests
   , UtilWarsh.tests
   ]
