@@ -9,7 +9,6 @@ import Data.List as List
 
 import Agda.Syntax.Internal
 import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Errors
 import Agda.TypeChecking.InstanceArguments
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Reduce
@@ -58,13 +57,19 @@ addConstraint c = do
         reportSDoc "tc.constr.add" 20 $ text "  simplified:" <+> prettyTCM c'
         solveConstraint_ c'
       else addConstraint' c'
-    -- the added constraint can cause IFS constraints to be solved
+    -- the added constraint can cause IFS constraints to be solved (but only
+    -- the constraints which aren’t blocked on an uninstantiated meta)
     unless (isIFSConstraint c) $
-       wakeConstraints (isIFSConstraint . clValue . theConstraint)
+       wakeConstraints (isWakeableIFSConstraint . clValue . theConstraint)
   where
+    isWakeableIFSConstraint :: Constraint -> TCM Bool
+    isWakeableIFSConstraint (FindInScope _ b _) = caseMaybe b (return True) (\m -> isInstantiatedMeta m)
+    isWakeableIFSConstraint _ = return False
+
     isIFSConstraint :: Constraint -> Bool
     isIFSConstraint FindInScope{} = True
-    isIFSConstraint _ = False
+    isIFSConstraint _             = False
+
     simpl :: Constraint -> TCM Constraint
     simpl c = do
       n <- genericLength <$> getContext
@@ -121,13 +126,13 @@ whenConstraints action handler =
 -- | Wake up the constraints depending on the given meta.
 wakeupConstraints :: MetaId -> TCM ()
 wakeupConstraints x = do
-  wakeConstraints (const True) -- (mentionsMeta x) -- TODO: needs fixing to cope with shared updates
+  wakeConstraints (return . const True) -- (mentionsMeta x) -- TODO: needs fixing to cope with shared updates
   solveAwakeConstraints
 
 -- | Wake up all constraints.
 wakeupConstraints_ :: TCM ()
 wakeupConstraints_ = do
-  wakeConstraints (const True)
+  wakeConstraints (return . const True)
   solveAwakeConstraints
 
 solveAwakeConstraints :: TCM ()
@@ -192,10 +197,12 @@ solveConstraint_ (UnBlock m)                =
       -- Open (whatever that means)
       Open -> __IMPOSSIBLE__
       OpenIFS -> __IMPOSSIBLE__
-solveConstraint_ (FindInScope m cands)      = findInScope m cands
+solveConstraint_ (FindInScope m b cands)      = findInScope m cands
 
 checkTypeCheckingProblem :: TypeCheckingProblem -> TCM Term
 checkTypeCheckingProblem p = case p of
   CheckExpr e t                  -> checkExpr e t
-  CheckArgs eh ei r args t0 t1 k -> checkArguments' eh ei r args t0 t1 k
+  CheckArgs eh r args t0 t1 k    -> checkArguments' eh r args t0 t1 k
   CheckLambda args body target   -> checkPostponedLambda args body target
+  UnquoteTactic tac hole t       -> unquoteTactic tac hole t $ return hole
+
