@@ -13,7 +13,7 @@ import Data.Monoid
 import Data.Traversable (traverse)
 
 import Agda.Syntax.Abstract (IsProjP(..))
-import Agda.Syntax.Common hiding (Arg, Dom, NamedArg)
+import Agda.Syntax.Common
 import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Pattern ()
 import Agda.Syntax.Literal
@@ -28,7 +28,7 @@ import Agda.Utils.Impossible
 {-| Given
 
     1. the function clauses @cs@
-    2. the patterns @ps@ and permutation @perm@ of a split clause
+    2. the patterns @ps@
 
 we want to compute a variable index of the split clause to split on next.
 
@@ -40,10 +40,10 @@ We try to split on this column first.
 -}
 
 -- | Match the given patterns against a list of clauses
-match :: [Clause] -> [Arg Pattern] -> Permutation -> Match Nat
-match cs ps perm = foldr choice No $ zipWith matchIt [0..] cs
+match :: [Clause] -> [Arg DeBruijnPattern] -> Match (Nat,[MPat])
+match cs ps = foldr choice No $ zipWith matchIt [0..] cs
   where
-    mps = buildMPatterns perm ps
+    mps = buildMPatterns ps
 
     -- If liberal matching on literals fails or blocks we go with that.
     -- If it succeeds we use the result from conservative literal matching.
@@ -66,27 +66,35 @@ match cs ps perm = foldr choice No $ zipWith matchIt [0..] cs
 --   are blocking a match.
 data MPat
   = VarMP Nat    -- ^ De Bruijn index (usually, rightmost variable in patterns is 0).
-  | ConMP ConHead [Arg MPat]
+  | ConMP ConHead (Maybe ConPOrigin) [Arg MPat]
   | LitMP Literal
+  | DotMP MPat   -- ^ For keeping track of the original dot positions.
   | WildMP       -- ^ For dot patterns that cannot be turned into patterns.
   | ProjMP QName -- ^ Projection copattern.
+  deriving (Show)
 
-buildMPatterns :: Permutation -> [Arg Pattern] -> [Arg MPat]
-buildMPatterns perm ps = evalState (mapM (traverse build) ps) xs
+buildMPatterns :: [Arg DeBruijnPattern] -> [Arg MPat]
+buildMPatterns ps = map (fmap build) ps
   where
-    xs   = permPicks $ flipP $ invertP __IMPOSSIBLE__ perm
-    tick = do x : xs <- get; put xs; return x
+    build (VarP (i,_))    = VarMP i
+    build (ConP con i ps) = ConMP con (conPRecord i) $ buildMPatterns $ map (fmap namedThing) $ ps
+    build (DotP t)        = DotMP $ buildT t
+    build (LitP l)        = LitMP l
+    build (ProjP dest)    = ProjMP dest
 
-    build (VarP _)        = VarMP <$> tick
-    build (ConP con _ ps) = ConMP con <$> mapM (traverse build . fmap namedThing) ps
-    build (DotP t)        = tick *> buildT t
-    build (LitP l)        = return $ LitMP l
-    build (ProjP dest)    = return $ ProjMP dest
-
-    buildT (Con c args)   = ConMP c <$> mapM (traverse buildT) args
-    buildT (Var i [])     = return (VarMP i)
-    buildT _              = return WildMP
+    buildT (Con c args)   = ConMP c Nothing $ map (fmap buildT) args
+    buildT (Var i [])     = VarMP i
+    buildT (Shared p)     = buildT (derefPtr p)
+    buildT _              = WildMP
 
+isTrivialMPattern :: MPat -> Bool
+isTrivialMPattern VarMP{} = True
+isTrivialMPattern (ConMP c (Just _) ps) = all isTrivialMPattern $ map unArg ps
+isTrivialMPattern (ConMP c Nothing ps) = False
+isTrivialMPattern LitMP{} = False
+isTrivialMPattern DotMP{} = True
+isTrivialMPattern WildMP{} = True
+isTrivialMPattern ProjMP{} = False -- or True?
 
 -- | If matching is inconclusive (@Block@) we want to know which
 --   variables are blocking the match.
@@ -143,27 +151,27 @@ choice (Block x) _         = Block x
 choice BlockP    m         = BlockP
 choice No        m         = m
 
-type MatchLit = Literal -> MPat -> Match ()
+type MatchLit = Literal -> MPat -> Match [MPat]
 
 noMatchLit :: MatchLit
 noMatchLit _ _ = No
 
 yesMatchLit :: MatchLit
-yesMatchLit _ VarMP{}  = Yes ()
-yesMatchLit _ WildMP{} = Yes ()
+yesMatchLit _ q@VarMP{}  = Yes [q]
+yesMatchLit _ q@WildMP{} = Yes [q]
 yesMatchLit _ _        = No
 
 -- | Check if a clause could match given generously chosen literals
-matchLits :: Clause -> [Arg Pattern] -> Permutation -> Bool
-matchLits c ps perm =
-  case matchClause yesMatchLit (buildMPatterns perm ps) 0 c of
+matchLits :: Clause -> [Arg DeBruijnPattern] -> Bool
+matchLits c ps =
+  case matchClause yesMatchLit (buildMPatterns ps) 0 c of
     Yes _ -> True
     _     -> False
 
 -- | @matchClause mlit qs i c@ checks whether clause @c@ number @i@
 --   covers a split clause with patterns @qs@.
-matchClause :: MatchLit -> [Arg MPat] -> Nat -> Clause -> Match Nat
-matchClause mlit qs i c = i <$ matchPats mlit (clausePats c) qs
+matchClause :: MatchLit -> [Arg MPat] -> Nat -> Clause -> Match (Nat,[MPat])
+matchClause mlit qs i c = (\q -> (i,q)) <$> matchPats mlit (clausePats c) qs
 
 -- | @matchPats mlit ps qs@ checks whether a function clause with patterns
 --   @ps@ covers a split clause with patterns @qs@.
@@ -178,7 +186,7 @@ matchClause mlit qs i c = i <$ matchPats mlit (clausePats c) qs
 --     F true = Set
 --     F      = \ x -> Set
 --   @
-matchPats :: MatchLit -> [Arg Pattern] -> [Arg MPat] -> Match ()
+matchPats :: MatchLit -> [Arg (Pattern' a)] -> [Arg MPat] -> Match [MPat]
 matchPats mlit ps qs = mconcat $ properMatchesLeft :
     zipWith (matchPat mlit) (map unArg ps) (map unArg qs) ++
     [ projPatternsLeft ]
@@ -186,11 +194,11 @@ matchPats mlit ps qs = mconcat $ properMatchesLeft :
     projPatternsLeft =
       let psrest = map unArg $ drop (length qs) ps
       in  if null $ mapMaybe isProjP psrest -- not $ any properlyMatching psrest
-            then Yes ()  -- no proj. patterns left
+            then Yes []  -- no proj. patterns left
             else BlockP  -- proj. patterns left
     properMatchesLeft =
       if any (properMatch . unArg) $ drop (length ps) qs
-      then No else Yes ()
+      then No else Yes []
     properMatch ConMP{} = True
     properMatch LitMP{} = True
     properMatch _       = False
@@ -229,21 +237,27 @@ instance Monoid a => Monoid (Match a) where
 --   @No@ means it does not cover.
 --   @Block [x]@ means @p@ is a proper instance of @q@ and could become
 --   a cover if @q@ was split on variable @x@.
-matchPat :: MatchLit -> Pattern -> MPat -> Match ()
-matchPat _    (VarP _) _ = Yes ()
-matchPat _    (DotP _) _ = Yes ()
+matchPat :: MatchLit -> Pattern' a -> MPat -> Match [MPat]
+matchPat _    (VarP _) q = Yes [q]
+matchPat _    (DotP _) q = Yes []
+-- Jesper, 2014-11-04: putting 'Yes [q]' here triggers issue 1333.
+-- Not checking for trivial MPats should be safe here, as dot patterns are
+-- guaranteed to match if the rest of the pattern does, so some extra splitting
+-- on them doesn't change the reduction behaviour.
+matchPat mlit p (DotMP q) = matchPat mlit p q
 matchPat mlit (LitP l) q = mlit l q
-matchPat _    (ProjP d) (ProjMP d') = if d == d' then Yes () else No
+matchPat _    (ProjP d) (ProjMP d') = if d == d' then Yes [] else No
 matchPat _    (ProjP d) _ = __IMPOSSIBLE__
 -- matchPat mlit (ConP c (Just _) ps) q | recordPattern ps = Yes ()  -- Andreas, 2012-07-25 record patterns always match!
 matchPat mlit (ConP c _ ps) q = case q of
   VarMP x -> Block [BlockingVar x (Just [c])]
   WildMP{} -> No -- Andreas, 2013-05-15 this was "Yes()" triggering issue 849
-  ConMP c' qs
+  ConMP c' i qs
     | c == c'   -> matchPats mlit (map (fmap namedThing) ps) qs
     | otherwise -> No
   LitMP _  -> __IMPOSSIBLE__
   ProjMP _ -> __IMPOSSIBLE__
+  DotMP _  -> __IMPOSSIBLE__
 
 {- UNUSED
 class RecordPattern a where
