@@ -1,6 +1,6 @@
-{-# LANGUAGE CPP #-}
+{-# LANGUAGE CPP               #-}
 {-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE TupleSections #-}
+{-# LANGUAGE TupleSections     #-}
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
@@ -16,33 +16,40 @@ import Prelude hiding (null)
 import Control.Monad.State
 
 import Data.Function
-import Data.List (nub, sortBy, intercalate)
+import Data.List (nub, sortBy, intersperse, isInfixOf)
 import Data.Maybe
+import Data.Char (toLower)
+import qualified Data.Set as Set
+import qualified Data.Map as Map
+import qualified Text.PrettyPrint.Boxes as Boxes
 
-import Agda.Syntax.Common hiding (Arg, Dom, NamedArg)
-import qualified Agda.Syntax.Common as Common
+import Agda.Syntax.Common
 import Agda.Syntax.Fixity
+import Agda.Syntax.Notation
 import Agda.Syntax.Position
 import qualified Agda.Syntax.Info as A
 import qualified Agda.Syntax.Concrete as C
 import qualified Agda.Syntax.Concrete.Definitions as D
 import Agda.Syntax.Abstract as A
+import Agda.Syntax.Abstract.Views (deepUnscope)
 import Agda.Syntax.Internal as I
-import qualified Agda.Syntax.Abstract.Pretty as AP
 import Agda.Syntax.Translation.InternalToAbstract
 import Agda.Syntax.Translation.AbstractToConcrete
 import Agda.Syntax.Scope.Monad (isDatatypeModule)
+import Agda.Syntax.Scope.Base
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.Closure
 import Agda.TypeChecking.Monad.Context
 import Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Monad.Builtin
+import Agda.TypeChecking.Monad.State
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Reduce (instantiate)
 
 import Agda.Utils.Except ( MonadError(catchError) )
 import Agda.Utils.FileName
 import Agda.Utils.Function
+import Agda.Utils.List
 import Agda.Utils.Monad
 import Agda.Utils.Null
 import Agda.Utils.Size
@@ -145,6 +152,7 @@ errorString err = case err of
   CoverageCantSplitOn{}                    -> "CoverageCantSplitOn"
   CoverageCantSplitIrrelevantType{}        -> "CoverageCantSplitIrrelevantType"
   CoverageCantSplitType{}                  -> "CoverageCantSplitType"
+  CoverageNoExactSplit{}                   -> "CoverageNoExactSplit"
   CyclicModuleDependency{}                 -> "CyclicModuleDependency"
   DataMustEndInSort{}                      -> "DataMustEndInSort"
 -- UNUSED:    DataTooManyParameters{}                  -> "DataTooManyParameters"
@@ -205,15 +213,17 @@ errorString err = case err of
   NothingAppliedToHiddenArg{}              -> "NothingAppliedToHiddenArg"
   NothingAppliedToInstanceArg{}            -> "NothingAppliedToInstanceArg"
   OverlappingProjects {}                   -> "OverlappingProjects"
+  OperatorChangeMessage {}                 -> "OperatorChangeMessage"
+  OperatorInformation {}                   -> "OperatorInformation"
   PatternShadowsConstructor {}             -> "PatternShadowsConstructor"
   PropMustBeSingleton                      -> "PropMustBeSingleton"
   RepeatedVariablesInPattern{}             -> "RepeatedVariablesInPattern"
   SafeFlagPostulate{}                      -> "SafeFlagPostulate"
   SafeFlagPragma{}                         -> "SafeFlagPragma"
-  SafeFlagNoTerminationCheck{}             -> "SafeFlagNoTerminationCheck"
   SafeFlagNonTerminating{}                 -> "SafeFlagNonTerminating"
   SafeFlagTerminating{}                    -> "SafeFlagTerminating"
   SafeFlagPrimTrustMe{}                    -> "SafeFlagPrimTrustMe"
+  SafeFlagNoPositivityCheck{}              -> "SafeNoPositivityCheck"
   ShadowedModule{}                         -> "ShadowedModule"
   ShouldBeASort{}                          -> "ShouldBeASort"
   ShouldBeApplicationOf{}                  -> "ShouldBeApplicationOf"
@@ -239,8 +249,12 @@ errorString err = case err of
   UnequalSorts{}                           -> "UnequalSorts"
   UnequalTerms{}                           -> "UnequalTerms"
   UnequalTypes{}                           -> "UnequalTypes"
+  UnifyConflict{}                          -> "UnifyConflict"
+  UnifyCycle{}                             -> "UnifyCycle"
+  UnifyIndicesNotVars{}                    -> "UnifyIndicesNotVars"
+  UnificationRecursiveEq{}                 -> "UnificationRecursiveEq"
+  UnificationStuck{}                       -> "UnificationStuck"
 --  UnequalTelescopes{}                      -> "UnequalTelescopes" -- UNUSED
-  UnequalColors{}                          -> "UnequalTelescopes"
   HeterogeneousEquality{}                  -> "HeterogeneousEquality"
   WithOnFreeVariable{}                     -> "WithOnFreeVariable"
   UnexpectedWithPatterns{}                 -> "UnexpectedWithPatterns"
@@ -257,12 +271,12 @@ errorString err = case err of
   WrongHidingInApplication{}               -> "WrongHidingInApplication"
   WrongHidingInLHS{}                       -> "WrongHidingInLHS"
   WrongHidingInLambda{}                    -> "WrongHidingInLambda"
+  WrongInstanceDeclaration{}               -> "WrongInstanceDeclaration"
   WrongIrrelevanceInLambda{}               -> "WrongIrrelevanceInLambda"
   WrongNamedArgument{}                     -> "WrongNamedArgument"
   WrongNumberOfConstructorArguments{}      -> "WrongNumberOfConstructorArguments"
   HidingMismatch{}                         -> "HidingMismatch"
   RelevanceMismatch{}                      -> "RelevanceMismatch"
-  ColorMismatch{}                          -> "ColorMismatch"
 
 instance PrettyTCM TCErr where
   prettyTCM err = case err of
@@ -363,6 +377,8 @@ instance PrettyTCM TypeError where
     WrongHidingInApplication t ->
       fwords "Found an implicit application where an explicit application was expected"
 
+    WrongInstanceDeclaration -> fwords "Terms marked as eligible for instance search should end with a name"
+
     HidingMismatch h h' -> fwords $
       "Expected " ++ verbalize (Indefinite h') ++ " argument, but found " ++
       verbalize (Indefinite h) ++ " argument"
@@ -371,10 +387,6 @@ instance PrettyTCM TypeError where
       "Expected " ++ verbalize (Indefinite r') ++ " argument, but found " ++
       verbalize (Indefinite r) ++ " argument"
 
-    ColorMismatch c c' -> fsep $   -- TODO guilhem
-      pwords "Expected argument color to be" ++ [prettyTCM c'] ++
-      pwords "but found color" ++ [prettyTCM c]
-
     NotInductive t -> fsep $
       [prettyTCM t] ++ pwords "is not an inductive data type"
 
@@ -525,8 +537,6 @@ instance PrettyTCM TypeError where
     UnequalTypes cmp a b -> prettyUnequal a (notCmp cmp) b
 --              fsep $ [prettyTCM a, notCmp cmp, prettyTCM b]
 
-    UnequalColors a b -> error "TODO guilhem 4"
-
     HeterogeneousEquality u a v b -> fsep $
       pwords "Refuse to solve heterogeneous constraint" ++
       [prettyTCM u] ++ pwords ":" ++ [prettyTCM a] ++ pwords "=?=" ++
@@ -693,19 +703,31 @@ instance PrettyTCM TypeError where
 
     BothWithAndRHS -> fsep $ pwords "Unexpected right hand side"
 
-    NotInScope xs ->
-      fsep (pwords "Not in scope:") $$ nest 2 (vcat $ map name xs)
+    NotInScope xs -> do
+      inscope <- Set.toList . concreteNamesInScope <$> getScope
+      fsep (pwords "Not in scope:") $$ nest 2 (vcat $ map (name inscope) xs)
       where
-      name x = fsep [ pretty x
-                    , text "at" <+> prettyTCM (getRange x)
-                    , suggestion (P.prettyShow x)
-                    ]
-      suggestion s
-        | elem ':' s    = parens $ text "did you forget space around the ':'?"
-        | elem "->" two = parens $ text "did you forget space around the '->'?"
-        | otherwise     = empty
+      name inscope x =
+        fsep [ pretty x
+             , text "at" <+> prettyTCM (getRange x)
+             , suggestion inscope x
+             ]
+      suggestion inscope x = nest 2 $ par $
+        [ text "did you forget space around the ':'?"  | elem ':' s ] ++
+        [ text "did you forget space around the '->'?" | isInfixOf "->" s ] ++
+        [ sep [ text "did you mean"
+              , nest 2 $ vcat (punctuate (text " or") $ map (\ y -> text $ "'" ++ y ++ "'") ys) <> text "?" ]
+          | not $ null ys ]
         where
-          two = zipWith (\a b -> [a,b]) s (tail s)
+          s = P.prettyShow x
+          par []  = empty
+          par [d] = parens d
+          par ds  = parens $ vcat ds
+
+          strip x = map toLower $ filter (/= '_') $ P.prettyShow $ C.unqualify x
+          maxDist n = div n 3
+          close a b = editDistance a b <= maxDist (length a)
+          ys = map P.prettyShow $ filter (close (strip x) . strip) inscope
 
     NoSuchModule x -> fsep $ pwords "No such module" ++ [pretty x]
 
@@ -789,8 +811,8 @@ instance PrettyTCM TypeError where
       [pretty e] ++ pwords "cannot appear by itself. It needs to be the argument to" ++
       pwords "a function expecting an instance argument."
 
-    NoParseForApplication es -> fsep $
-      pwords "Could not parse the application" ++ [pretty $ C.RawApp noRange es]
+    NoParseForApplication es -> fsep (
+      pwords "Could not parse the application" ++ [pretty $ C.RawApp noRange es])
 
     AmbiguousParseForApplication es es' -> fsep (
       pwords "Don't know how to parse" ++ [pretty_es <> (text ".")] ++
@@ -804,20 +826,29 @@ instance PrettyTCM TypeError where
         pretty' e = do
           p1 <- pretty_es
           p2 <- pretty e
-          pretty $ if show p1 == show p2 then unambiguous e else e
+          if show p1 == show p2 then unambiguous e else pretty e
 
-        unambiguous :: C.Expr -> C.Expr
-        unambiguous (C.OpApp r op _ xs) | all (isOrdinary . namedArg) xs
-            = foldl (C.App r) (C.Ident op) $ (map . fmap . fmap) fromOrdinary xs
-        unambiguous e = e
+        unambiguous :: C.Expr -> TCM Doc
+        unambiguous e@(C.OpApp r op _ xs)
+          | all (isOrdinary . namedArg) xs =
+            pretty $
+              foldl (C.App r) (C.Ident op) $
+                (map . fmap . fmap) fromOrdinary xs
+          | any (isPlaceholder . namedArg) xs =
+              pretty e <+> text "(section)"
+        unambiguous e = pretty e
 
-        isOrdinary :: C.OpApp e -> Bool
-        isOrdinary (C.Ordinary _) = True
-        isOrdinary _              = False
+        isOrdinary :: MaybePlaceholder (C.OpApp e) -> Bool
+        isOrdinary (NoPlaceholder (C.Ordinary _)) = True
+        isOrdinary _                              = False
 
-        fromOrdinary :: C.OpApp e -> e
-        fromOrdinary (C.Ordinary e) = e
-        fromOrdinary _              = __IMPOSSIBLE__
+        fromOrdinary :: MaybePlaceholder (C.OpApp e) -> e
+        fromOrdinary (NoPlaceholder (C.Ordinary e)) = e
+        fromOrdinary _                              = __IMPOSSIBLE__
+
+        isPlaceholder :: MaybePlaceholder a -> Bool
+        isPlaceholder Placeholder{}   = True
+        isPlaceholder NoPlaceholder{} = False
 
     BadArgumentsToPatternSynonym x -> fsep $
       pwords "Bad arguments to pattern synonym " ++ [prettyTCM x]
@@ -828,11 +859,11 @@ instance PrettyTCM TypeError where
     UnusedVariableInPatternSynonym -> fsep $
       pwords "Unused variable in pattern synonym."
 
-    NoParseForLHS IsLHS p -> fsep $
-      pwords "Could not parse the left-hand side" ++ [pretty p]
+    NoParseForLHS IsLHS p -> fsep (
+      pwords "Could not parse the left-hand side" ++ [pretty p])
 
-    NoParseForLHS IsPatSyn p -> fsep $
-      pwords "Could not parse the pattern synonym" ++ [pretty p]
+    NoParseForLHS IsPatSyn p -> fsep (
+      pwords "Could not parse the pattern synonym" ++ [pretty p])
 
 {- UNUSED
     NoParseForPatternSynonym p -> fsep $
@@ -864,6 +895,100 @@ instance PrettyTCM TypeError where
         unambiguousP (C.OpAppP r op _ xs) = foldl C.AppP (C.IdentP op) xs
         unambiguousP e                    = e
 
+    OperatorInformation sects err ->
+      prettyTCM err
+        $+$
+      fsep (pwords "Operators used in the grammar:")
+        $$
+      nest 2
+        (if null sects then text "None" else
+         vcat (map text $
+               lines $
+               Boxes.render $
+               (\(col1, col2, col3) ->
+                   Boxes.hsep 1 Boxes.top $
+                   map (Boxes.vcat Boxes.left) [col1, col2, col3]) $
+               unzip3 $
+               map prettySect $
+               sortBy (compare `on` show . notaName . sectNotation) $
+               filter (not . closedWithoutHoles) sects))
+      where
+      trimLeft  = dropWhile isNormalHole
+      trimRight = reverse . dropWhile isNormalHole . reverse
+
+      closedWithoutHoles sect =
+        sectKind sect == NonfixNotation
+          &&
+        null [ () | NormalHole {} <- trimLeft $ trimRight $
+                                       notation (sectNotation sect) ]
+
+      prettyName n = Boxes.text $
+        P.render (P.pretty n) ++
+        " (" ++ P.render (P.pretty (nameBindingSite n)) ++ ")"
+
+      prettySect sect =
+        ( Boxes.text (P.render (P.pretty section))
+            Boxes.//
+          strut
+        , Boxes.text
+            ("(" ++
+             kind ++ " " ++
+             (if notaIsOperator nota
+              then "operator"
+              else "notation") ++
+             (if sectIsSection sect
+              then " section"
+              else "") ++
+             (case sectLevel sect of
+                Nothing          -> ""
+                Just Unrelated   -> ", unrelated"
+                Just (Related n) -> ", level " ++ show n) ++
+             ")")
+            Boxes.//
+          strut
+        , Boxes.text "["
+            Boxes.<>
+          Boxes.vcat Boxes.left
+            (map (\n -> prettyName n Boxes.<> Boxes.text ",") names ++
+             [prettyName name Boxes.<> Boxes.text "]"])
+        )
+        where
+        nota    = sectNotation sect
+        section = trim (notation nota)
+
+        trim = case sectKind sect of
+          InfixNotation   -> trimLeft . trimRight
+          PrefixNotation  -> trimRight
+          PostfixNotation -> trimLeft
+          NonfixNotation  -> id
+          NoNotation      -> __IMPOSSIBLE__
+
+        (names, name) = case Set.toList $ notaNames nota of
+          [] -> __IMPOSSIBLE__
+          ns -> (init ns, last ns)
+
+        strut = Boxes.emptyBox (length names) 0
+
+        kind = case sectKind sect of
+          PrefixNotation  -> "prefix"
+          PostfixNotation -> "postfix"
+          NonfixNotation  -> "closed"
+          NoNotation      -> __IMPOSSIBLE__
+          InfixNotation   ->
+            case fixityAssoc $ notaFixity nota of
+              NonAssoc   -> "infix"
+              LeftAssoc  -> "infixl"
+              RightAssoc -> "infixr"
+
+    OperatorChangeMessage err@(OperatorInformation [] _) ->
+      prettyTCM err
+    OperatorChangeMessage err ->
+      prettyTCM err
+        $+$
+      fsep (pwords $
+        "(the treatment of operators was changed in Agda 2.5.1, " ++
+        "so code that used to parse may have to be changed)")
+
 {- UNUSED
     AmbiguousParseForPatternSynonym p ps -> fsep (
       pwords "Don't know how to parse" ++ [pretty p <> text "."] ++
@@ -908,9 +1033,13 @@ instance PrettyTCM TypeError where
     CoverageCantSplitIrrelevantType a -> fsep $
       pwords "Cannot split on argument of irrelevant datatype" ++ [prettyTCM a]
 
+
     CoverageCantSplitType a -> fsep $
       pwords "Cannot split on argument of non-datatype" ++ [prettyTCM a]
 
+    CoverageNoExactSplit f cs -> fsep $
+      pwords "Exact splitting is enabled, but not all clauses can be preserved as definitional equalities in the translation to a case tree"
+
     SplitError e -> prettyTCM e
 
     WithoutKError a u v -> fsep $
@@ -918,9 +1047,36 @@ instance PrettyTCM TypeError where
       pwords "=" ++ [prettyTCM v] ++ pwords "of type" ++ [prettyTCM a] ++
       pwords "because K has been disabled."
 
+    UnifyConflict c c' -> fsep $
+      pwords "There was a conflict between the constructors " ++
+      [prettyTCM c] ++ pwords " and " ++ [prettyTCM c']
+
+    UnifyCycle i u -> fsep $
+      pwords "The variable " ++ [prettyTCM (var i)] ++
+      pwords "occurs strongly rigid in" ++ [prettyTCM u]
+
+    UnifyIndicesNotVars a u v ixs -> fsep $
+      pwords "Cannot apply injectivity to the equation" ++ [prettyTCM u] ++
+      pwords "=" ++ [prettyTCM v] ++ pwords "of type" ++ [prettyTCM a] ++
+      pwords "because I cannot generalize over the indices" ++
+      [prettyList $ map prettyTCM ixs]
+
+    UnificationRecursiveEq a i u -> fsep $
+      pwords "Cannot solve variable " ++ [prettyTCM (var i)] ++
+      pwords " of type " ++ [prettyTCM a] ++
+      pwords " with solution " ++ [prettyTCM u] ++
+      pwords " because the variable occurs in the solution," ++
+      pwords " or in the type one of the variables in the solution"
+
+    UnificationStuck tel us vs -> fsep $
+      pwords "I got stuck on unifying" ++ [prettyList (map prettyTCM us)] ++
+      pwords "with" ++ [prettyList (map prettyTCM vs)] ++
+      pwords "in the telescope" ++ [prettyTCM tel]
+
     NotStrictlyPositive d ocs -> fsep $
       pwords "The datatype" ++ [prettyTCM d] ++
-      pwords "is not strictly positive, because" ++ prettyOcc "it" ocs
+      pwords "is not strictly positive, because"
+      ++ prettyOcc "it" ocs
       where
         prettyOcc _ [] = []
         prettyOcc it (OccCon d c r : ocs) = concat
@@ -948,7 +1104,7 @@ instance PrettyTCM TypeError where
         com (_:_) = comma
 
     IFSNoCandidateInScope t -> fsep $
-      pwords "No variable of type" ++ [prettyTCM t] ++ pwords "was found in scope."
+      pwords "No instance of type" ++ [prettyTCM t] ++ pwords "was found in scope."
 
     UnquoteFailed e -> case e of
       BadVisibility msg arg -> fsep $
@@ -962,24 +1118,12 @@ instance PrettyTCM TypeError where
         pwords ("Use " ++ def ++ " instead of " ++ con ++ " for non-constructor")
         ++ [prettyTCM x]
 
-      NotAConstructor kind t ->
-        fwords "Unable to unquote the term"
+      NonCanonical kind t ->
+        fwords ("Cannot unquote non-canonical " ++ kind)
         $$ nest 2 (prettyTCM t)
-        $$ fwords ("of type " ++ kind ++ ". Reason: not a constructor.")
 
-      NotALiteral kind t ->
-        fwords "Unable to unquote the term"
-        $$ nest 2 (prettyTCM t)
-        $$ fwords ("of type " ++ kind ++ ". Reason: not a literal value.")
-
-      RhsUsesDottedVar ixs t ->
-        fwords "Unable to unquote the term"
-        $$ nest 2 (prettyTCM t)
-        $$ fwords "of type Clause. Reason: the right-hand side contains variables that are referring to a dot pattern."
-        $$ fwords ("Offending De Bruijn indices: "
-                   ++ intercalate ", " (map P.prettyShow ixs) ++ ".")
-
-      BlockedOnMeta m -> __IMPOSSIBLE__
+      BlockedOnMeta _ m -> fsep $
+        pwords $ "Unquote failed because of unsolved meta variables."
 
       UnquotePanic err -> __IMPOSSIBLE__
 
@@ -992,9 +1136,6 @@ instance PrettyTCM TypeError where
       in fsep $ [fwords ("Cannot set OPTION pragma" ++ plural)]
                 ++ map text xs ++ [fwords "with safe flag."]
 
-    SafeFlagNoTerminationCheck -> fsep $
-      pwords "Cannot use NO_TERMINATION_CHECK pragma with safe flag."
-
     SafeFlagNonTerminating -> fsep $
       pwords "Cannot use NON_TERMINATING pragma with safe flag."
 
@@ -1003,6 +1144,9 @@ instance PrettyTCM TypeError where
 
     SafeFlagPrimTrustMe -> fsep (pwords "Cannot use primTrustMe with safe flag")
 
+    SafeFlagNoPositivityCheck -> fsep $
+      pwords "Cannot use NO_POSITIVITY_CHECK pragma with safe flag."
+
     NeedOptionCopatterns -> fsep $
       pwords "Option --copatterns needed to enable destructor patterns"
     NeedOptionRewriting  -> fsep $
@@ -1013,13 +1157,13 @@ instance PrettyTCM TypeError where
       | n > 0 && not (null args) = parens
       | otherwise                = id
 
-    prettyArg :: I.Arg I.Pattern -> TCM Doc
-    prettyArg (Common.Arg info x) = case getHiding info of
+    prettyArg :: Arg (I.Pattern' a) -> TCM Doc
+    prettyArg (Arg info x) = case getHiding info of
       Hidden    -> braces $ prettyPat 0 x
       Instance  -> dbraces $ prettyPat 0 x
       NotHidden -> prettyPat 1 x
 
-    prettyPat :: Integer -> I.Pattern -> TCM Doc
+    prettyPat :: Integer -> (I.Pattern' a) -> TCM Doc
     prettyPat _ (I.VarP _) = text "_"
     prettyPat _ (I.DotP _) = text "._"
     prettyPat n (I.ConP c _ args) =
@@ -1044,7 +1188,7 @@ prettyInEqual t1 t2 = do
       (v1, v2) <- instantiate (t1, t2)
       case (ignoreSharing v1, ignoreSharing v2) of
         (I.Var i1 _, I.Var i2 _)
-          | i1 == i2  -> __IMPOSSIBLE__   -- if they're actually the same we would get the error on the arguments instead
+          | i1 == i2  -> generic -- possible, see issue 1826
           | otherwise -> varVar i1 i2
         (I.Def{}, I.Con{}) -> __IMPOSSIBLE__  -- ambiguous identifiers
         (I.Con{}, I.Def{}) -> __IMPOSSIBLE__
@@ -1054,10 +1198,11 @@ prettyInEqual t1 t2 = do
         (I.Con{}, I.Var{}) -> varCon
         _                  -> empty
   where
-    varDef, varCon :: TCM Doc
+    varDef, varCon, generic :: TCM Doc
     varDef = parens $ fwords "because one is a variable and one a defined identifier"
     varCon = parens $ fwords "because one is a variable and one a constructor"
-
+    generic = parens $ fwords $ "although these terms are looking the same, " ++
+      "they contain different but identically rendered identifiers somewhere"
     varVar :: Int -> Int -> TCM Doc
     varVar i j = parens $ fwords $
                    "because one has deBruijn index " ++ show i
@@ -1099,16 +1244,20 @@ instance PrettyTCM SplitError where
 
 instance PrettyTCM Call where
   prettyTCM c = case c of
-    CheckClause t cl -> fsep $
-      pwords "when checking that the clause"
-      ++ [AP.prettyA cl] ++ pwords "has type" ++ [prettyTCM t]
+    CheckClause t cl -> do
+      reportSLn "error.checkclause" 60 $ "prettyTCM CheckClause: cl = " ++ show (deepUnscope cl)
+      clc <- abstractToConcrete_ cl
+      reportSLn "error.checkclause" 40 $ "cl (Concrete) = " ++ show clc
+      fsep $
+        pwords "when checking that the clause"
+        ++ [prettyA cl] ++ pwords "has type" ++ [prettyTCM t]
 
     CheckPattern p tel t -> addCtxTel tel $ fsep $
       pwords "when checking that the pattern"
       ++ [prettyA p] ++ pwords "has type" ++ [prettyTCM t]
 
     CheckLetBinding b -> fsep $
-      pwords "when checking the let binding" ++ [AP.prettyA b]
+      pwords "when checking the let binding" ++ [prettyA b]
 
     InferExpr e -> fsep $ pwords "when inferring the type of" ++ [prettyA e]
 
@@ -1124,6 +1273,11 @@ instance PrettyTCM Call where
       pwords "when checking that the expression"
       ++ [prettyA e] ++ pwords "is a type"
 
+    CheckProjection _ x t -> fsep $
+      pwords "when checking the projection" ++
+      [ sep [ prettyTCM x <+> text ":"
+            , nest 2 $ prettyTCM t ] ]
+
     CheckArguments r es t0 t1 -> fsep $
       pwords "when checking that" ++
       map hPretty es ++
@@ -1163,7 +1317,7 @@ instance PrettyTCM Call where
       pwords "matches the inferred value" ++ [prettyTCM v]
 
     CheckPatternShadowing c -> fsep $
-      pwords "when checking the clause" ++ [AP.prettyA c]
+      pwords "when checking the clause" ++ [prettyA c]
 
     InferVar x ->
       fsep $ pwords "when inferring the type of" ++ [prettyTCM x]
@@ -1191,16 +1345,15 @@ instance PrettyTCM Call where
 
     CheckSectionApplication _ m1 modapp -> fsep $
       pwords "when checking the module application" ++
-      [prettyA $ A.Apply info m1 modapp empty empty]
+      [prettyA $ A.Apply info m1 modapp empty empty defaultImportDir]
       where
       info = A.ModuleInfo noRange noRange Nothing Nothing Nothing
 
+    ModuleContents -> fsep $ pwords "when retrieving the contents of a module"
+
     where
-    hPretty :: I.Arg (Named_ Expr) -> TCM Doc
-    hPretty a = do
-      info <- reify $ argInfo a
-      pretty =<< abstractToConcreteCtx (hiddenArgumentCtx (getHiding a))
-                   (Common.Arg info $ unArg a)
+    hPretty :: Arg (Named_ Expr) -> TCM Doc
+    hPretty a = pretty =<< abstractToConcreteCtx (hiddenArgumentCtx (getHiding a)) a
 
     simpleDecl = D.notSoNiceDeclaration
 
