@@ -1,8 +1,9 @@
-{-# LANGUAGE CPP #-}
+{-# LANGUAGE CPP               #-}
 {-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE PatternGuards #-}
-{-# LANGUAGE RelaxedPolyRec #-}
-{-# LANGUAGE TupleSections #-}
+{-# LANGUAGE FlexibleContexts  #-}
+{-# LANGUAGE PatternGuards     #-}
+{-# LANGUAGE RelaxedPolyRec    #-}
+{-# LANGUAGE TupleSections     #-}
 
 module Agda.TypeChecking.MetaVars where
 
@@ -13,13 +14,10 @@ import Data.List hiding (sort)
 import qualified Data.List as List
 import qualified Data.Map as Map
 import qualified Data.Foldable as Fold
-import qualified Data.Traversable as Trav
 
 import Agda.Syntax.Common
-import Agda.Syntax.Internal as I
+import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Generic
-import Agda.Syntax.Position
-import qualified Agda.Syntax.Abstract as A
 
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
@@ -27,7 +25,6 @@ import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Constraints
-import Agda.TypeChecking.Errors
 import Agda.TypeChecking.Free
 import Agda.TypeChecking.Level
 import Agda.TypeChecking.Records
@@ -35,6 +32,7 @@ import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.EtaContract
 import Agda.TypeChecking.SizedTypes (boundedSizeMetaHook, isSizeProblem)
+import {-# SOURCE #-} Agda.TypeChecking.CheckInternal
 
 -- import Agda.TypeChecking.CheckInternal
 -- import {-# SOURCE #-} Agda.TypeChecking.CheckInternal (checkInternal)
@@ -85,12 +83,12 @@ isBlockedTerm x = do
       if r then "  yes, because " ++ show i else "  no"
     return r
 
-isEtaExpandable :: MetaId -> TCM Bool
-isEtaExpandable x = do
+isEtaExpandable :: [MetaKind] -> MetaId -> TCM Bool
+isEtaExpandable kinds x = do
     i <- mvInstantiation <$> lookupMeta x
     return $ case i of
       Open{}                         -> True
-      OpenIFS{}                      -> False
+      OpenIFS{}                      -> notElem Records kinds
       InstV{}                        -> False
       InstS{}                        -> False
       BlockedConst{}                 -> False
@@ -103,14 +101,14 @@ isEtaExpandable x = do
 --   The instantiation should not be an 'InstV' or 'InstS' and the 'MetaId'
 --   should point to something 'Open' or a 'BlockedConst'.
 --   Further, the meta variable may not be 'Frozen'.
-assignTerm :: MetaId -> [I.Arg ArgName] -> Term -> TCM ()
+assignTerm :: MetaId -> [Arg ArgName] -> Term -> TCM ()
 assignTerm x tel v = do
      -- verify (new) invariants
     whenM (isFrozen x) __IMPOSSIBLE__
     assignTerm' x tel v
 
 -- | Skip frozen check.  Used for eta expanding frozen metas.
-assignTerm' :: MetaId -> [I.Arg ArgName] -> Term -> TCM ()
+assignTerm' :: MetaId -> [Arg ArgName] -> Term -> TCM ()
 assignTerm' x tel v = do
     reportSLn "tc.meta.assign" 70 $ prettyShow x ++ " := " ++ show v ++ "\n  in " ++ show tel
      -- verify (new) invariants
@@ -182,7 +180,7 @@ newTypeMeta_  = newTypeMeta =<< (workOnTypes $ newSortMeta)
 --   of type the output type of @t@ with name suggestion @s@ and initial
 --   solution candidates @cands@. If @t@ is a function type, then insert enough
 --   lambdas in front of it.
-newIFSMeta :: MetaNameSuggestion -> Type -> Maybe [(Term, Type)] -> TCM Term
+newIFSMeta :: MetaNameSuggestion -> Type -> Maybe [Candidate] -> TCM Term
 newIFSMeta s t cands = do
   TelV tel t' <- telView t
   addCtxTel tel $ do
@@ -191,7 +189,7 @@ newIFSMeta s t cands = do
     teleLam tel <$> newIFSMetaCtx s (telePi_ ctx t') vs (raise (size tel) cands)
 
 -- | Create a new value meta with specific dependencies.
-newIFSMetaCtx :: MetaNameSuggestion -> Type -> Args -> Maybe [(Term, Type)] -> TCM Term
+newIFSMetaCtx :: MetaNameSuggestion -> Type -> Args -> Maybe [Candidate] -> TCM Term
 newIFSMetaCtx s t vs cands = do
   reportSDoc "tc.meta.new" 50 $ fsep
     [ text "new ifs meta:"
@@ -205,7 +203,8 @@ newIFSMetaCtx s t vs cands = do
   reportSDoc "tc.meta.new" 50 $ fsep
     [ nest 2 $ pretty x <+> text ":" <+> prettyTCM t
     ]
-  addConstraint $ FindInScope x cands
+  addConstraint $ FindInScope x Nothing cands
+  etaExpandMetaSafe x
   return $ MetaV x $ map Apply vs
 
 
@@ -220,25 +219,24 @@ newValueMeta :: RunMetaOccursCheck -> Type -> TCM Term
 newValueMeta b t = do
   vs  <- getContextArgs
   tel <- getContextTelescope
-  newValueMetaCtx b (telePi_ tel t) vs
+  newValueMetaCtx b t tel (idP $ size tel) vs
 
-newValueMetaCtx :: RunMetaOccursCheck -> Type -> Args -> TCM Term
-newValueMetaCtx b t ctx =
-  instantiateFull =<< newValueMetaCtx' b t ctx
+newValueMetaCtx :: RunMetaOccursCheck -> Type -> Telescope -> Permutation -> Args -> TCM Term
+newValueMetaCtx b t tel perm ctx =
+  instantiateFull =<< newValueMetaCtx' b t tel perm ctx
 
 -- | Create a new value meta without η-expanding.
 newValueMeta' :: RunMetaOccursCheck -> Type -> TCM Term
 newValueMeta' b t = do
   vs  <- getContextArgs
   tel <- getContextTelescope
-  newValueMetaCtx' b (telePi_ tel t) vs
+  newValueMetaCtx' b t tel (idP $ size tel) vs
 
 -- | Create a new value meta with specific dependencies.
-newValueMetaCtx' :: RunMetaOccursCheck -> Type -> Args -> TCM Term
-newValueMetaCtx' b t vs = do
+newValueMetaCtx' :: RunMetaOccursCheck -> Type -> Telescope -> Permutation -> Args -> TCM Term
+newValueMetaCtx' b a tel perm vs = do
   i <- createMetaInfo' b
-  TelV tel a <- telView t
-  let perm = idP (size tel)
+  let t     = telePi_ tel a
   x <- newMeta i normalMetaPriority perm (HasType () t)
   reportSDoc "tc.meta.new" 50 $ fsep
     [ text "new meta:"
@@ -247,14 +245,14 @@ newValueMetaCtx' b t vs = do
     ]
   etaExpandMetaSafe x
   -- Andreas, 2012-09-24: for Metas X : Size< u add constraint X+1 <= u
-  let u = shared $ MetaV x $ map Apply vs
+  u <- shared $ MetaV x $ map Apply vs
   boundedSizeMetaHook u tel a
   return u
 
 newTelMeta :: Telescope -> TCM Args
 newTelMeta tel = newArgsMeta (abstract tel $ typeDontCare)
 
-type Condition = I.Dom Type -> Abs Type -> Bool
+type Condition = Dom Type -> Abs Type -> Bool
 
 trueCondition :: Condition
 trueCondition _ _ = True
@@ -266,26 +264,25 @@ newArgsMeta' :: Condition -> Type -> TCM Args
 newArgsMeta' condition t = do
   args <- getContextArgs
   tel  <- getContextTelescope
-  newArgsMetaCtx' condition t tel args
+  newArgsMetaCtx' condition t tel (idP $ size tel) args
 
-newArgsMetaCtx :: Type -> Telescope -> Args -> TCM Args
+newArgsMetaCtx :: Type -> Telescope -> Permutation -> Args -> TCM Args
 newArgsMetaCtx = newArgsMetaCtx' trueCondition
 
-newArgsMetaCtx' :: Condition -> Type -> Telescope -> Args -> TCM Args
-newArgsMetaCtx' condition (El s tm) tel ctx = do
+newArgsMetaCtx' :: Condition -> Type -> Telescope -> Permutation -> Args -> TCM Args
+newArgsMetaCtx' condition (El s tm) tel perm ctx = do
   tm <- reduce tm
   case ignoreSharing tm of
     Pi dom@(Dom info a) codom | condition dom codom -> do
-      arg  <- Arg info <$> do
-              applyRelevanceToContext (getRelevance info) $
+      u <- applyRelevanceToContext (getRelevance info) $
                {-
                  -- Andreas, 2010-09-24 skip irrelevant record fields when eta-expanding a meta var
                  -- Andreas, 2010-10-11 this is WRONG, see Issue 347
                 if r == Irrelevant then return DontCare else
                 -}
-                 newValueMetaCtx RunMetaOccursCheck (telePi_ tel a) ctx
-      args <- newArgsMetaCtx' condition (El s tm `piApply` [arg]) tel ctx
-      return $ arg : args
+                 newValueMetaCtx RunMetaOccursCheck a tel perm ctx
+      args <- newArgsMetaCtx' condition (codom `absApp` u) tel perm ctx
+      return $ Arg info u : args
     _  -> return []
 
 -- | Create a metavariable of record type. This is actually one metavariable
@@ -294,12 +291,12 @@ newRecordMeta :: QName -> Args -> TCM Term
 newRecordMeta r pars = do
   args <- getContextArgs
   tel  <- getContextTelescope
-  newRecordMetaCtx r pars tel args
+  newRecordMetaCtx r pars tel (idP $ size tel) args
 
-newRecordMetaCtx :: QName -> Args -> Telescope -> Args -> TCM Term
-newRecordMetaCtx r pars tel ctx = do
+newRecordMetaCtx :: QName -> Args -> Telescope -> Permutation -> Args -> TCM Term
+newRecordMetaCtx r pars tel perm ctx = do
   ftel   <- flip apply pars <$> getRecordFieldTypes r
-  fields <- newArgsMetaCtx (telePi_ ftel $ sort Prop) tel ctx
+  fields <- newArgsMetaCtx (telePi_ ftel $ sort Prop) tel perm ctx
   con    <- getRecordConstructor r
   return $ Con con fields
 
@@ -308,7 +305,7 @@ newQuestionMark ii t = do
   -- Do not run check for recursive occurrence of meta in definitions,
   -- because we want to give the recursive solution interactively (Issue 589)
   m  <- newValueMeta' DontRunMetaOccursCheck t
-  let MetaV x _ = ignoreSharing m
+  MetaV x _ <- return $ ignoreSharing m   -- needs to be strict!
   connectInteractionPoint ii x
   return m
 
@@ -365,8 +362,9 @@ postponeTypeCheckingProblem_ p = do
   postponeTypeCheckingProblem p (unblock p)
   where
     unblock (CheckExpr _ t)           = unblockedTester t
-    unblock (CheckArgs _ _ _ _ t _ _) = unblockedTester t  -- The type of the head of the application.
+    unblock (CheckArgs _ _ _ t _ _)   = unblockedTester t  -- The type of the head of the application.
     unblock (CheckLambda _ _ t)       = unblockedTester t
+    unblock (UnquoteTactic _ _ _)     = __IMPOSSIBLE__     -- unquote problems must be supply their own tester
 
 -- | Create a postponed type checking problem @e : t@ that waits for conditon
 --   @unblock@.  A new meta is created in the current context that has as
@@ -377,7 +375,7 @@ postponeTypeCheckingProblem p unblock = do
   i   <- createMetaInfo' DontRunMetaOccursCheck
   tel <- getContextTelescope
   cl  <- buildClosure p
-  let t = problemType p
+  t   <- problemType p
   m   <- newMeta' (PostponedTypeCheckingProblem cl unblock)
                   i normalMetaPriority (idP (size tel))
          $ HasType () $ telePi_ tel t
@@ -398,10 +396,11 @@ postponeTypeCheckingProblem p unblock = do
   return v
 
 -- | Type of the term that is produced by solving the 'TypeCheckingProblem'.
-problemType :: TypeCheckingProblem -> Type
-problemType (CheckExpr _ t           ) = t
-problemType (CheckArgs _ _ _ _ _ t _ ) = t  -- The target type of the application.
-problemType (CheckLambda _ _ t       ) = t
+problemType :: TypeCheckingProblem -> TCM Type
+problemType (CheckExpr _ t           ) = return t
+problemType (CheckArgs _ _ _ _ t _ )   = return t  -- The target type of the application.
+problemType (CheckLambda _ _ t       ) = return t
+problemType (UnquoteTactic tac hole t) = return t
 
 -- | Eta expand metavariables listening on the current meta.
 etaExpandListeners :: MetaId -> TCM ()
@@ -442,7 +441,7 @@ allMetaKinds = [minBound .. maxBound]
 -- | Eta expand a metavariable, if it is of the specified kind.
 --   Don't do anything if the metavariable is a blocked term.
 etaExpandMeta :: [MetaKind] -> MetaId -> TCM ()
-etaExpandMeta kinds m = whenM (isEtaExpandable m) $ do
+etaExpandMeta kinds m = whenM (isEtaExpandable kinds m) $ do
   verboseBracket "tc.meta.eta" 20 ("etaExpandMeta " ++ prettyShow m) $ do
     let waitFor x = do
           reportSDoc "tc.meta.eta" 20 $ do
@@ -468,7 +467,7 @@ etaExpandMeta kinds m = whenM (isEtaExpandable m) $ do
         ifM (isEtaRecord r) {- then -} (do
           let ps = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
           let expand = do
-                u <- withMetaInfo' meta $ newRecordMetaCtx r ps tel $ teleArgs tel
+                u <- withMetaInfo' meta $ newRecordMetaCtx r ps tel (idP $ size tel) $ teleArgs tel
                 inTopContext $ do
                   verboseS "tc.meta.eta" 15 $ do
                     du <- prettyTCM u
@@ -780,7 +779,6 @@ attemptInertRHSImprovement m args v = do
         Level{}    -> notInert
         MetaV{}    -> notInert
         DontCare{} -> notInert
-        ExtLam{}   -> __IMPOSSIBLE__
         Shared{}   -> __IMPOSSIBLE__
 
     ensureNeutral :: Term -> Term -> TCM ()
@@ -808,7 +806,6 @@ attemptInertRHSImprovement m args v = do
             MetaV{}    -> notNeutral v
             Con{}      -> notNeutral v
             Lam{}      -> notNeutral v
-            ExtLam{}   -> __IMPOSSIBLE__
             Shared{}   -> __IMPOSSIBLE__
 -- END UNUSED -}
 
@@ -921,7 +918,8 @@ subtypingForSizeLt dir   x mvar t args v cont = do
         _ -> fallback
 
 -- | Eta-expand bound variables like @z@ in @X (fst z)@.
-expandProjectedVars :: (Normalise a, TermLike a, Show a, PrettyTCM a, NoProjectedVar a, Subst a, PrettyTCM b, Subst b) =>
+expandProjectedVars :: (Normalise a, TermLike a, Show a, PrettyTCM a, NoProjectedVar a,
+                        Subst Term a, PrettyTCM b, Subst Term b) =>
   a -> b -> (a -> b -> TCM c) -> TCM c
 expandProjectedVars args v ret = loop (args, v) where
   loop (args, v) = do
@@ -938,7 +936,7 @@ expandProjectedVars args v ret = loop (args, v) where
       Left (ProjVarExc i _) -> etaExpandProjectedVar i (args, v) done loop
 
 -- | Eta-expand a de Bruijn index of record type in context and passed term(s).
-etaExpandProjectedVar :: (PrettyTCM a, Subst a) => Int -> a -> TCM c -> (a -> TCM c) -> TCM c
+etaExpandProjectedVar :: (PrettyTCM a, Subst Term a) => Int -> a -> TCM c -> (a -> TCM c) -> TCM c
 etaExpandProjectedVar i v fail succeed = do
   reportSDoc "tc.meta.assign.proj" 40 $
     text "trying to expand projected variable" <+> prettyTCM (var i)
@@ -969,7 +967,7 @@ instance NoProjectedVar Term where
       Con (ConHead _ Inductive (_:_)) vs -> noProjectedVar vs
       _ -> return ()
 
-instance NoProjectedVar a => NoProjectedVar (I.Arg a) where
+instance NoProjectedVar a => NoProjectedVar (Arg a) where
   noProjectedVar = Fold.mapM_ noProjectedVar
 
 instance NoProjectedVar a => NoProjectedVar [a] where
@@ -1050,7 +1048,7 @@ checkLinearity ids0 = do
         (throwError ())
 
 -- Intermediate result in the following function
-type Res = [(I.Arg Nat, Term)]
+type Res = [(Arg Nat, Term)]
 
 -- | Exceptions raised when substitution cannot be inverted.
 data InvertExcept
@@ -1086,7 +1084,7 @@ inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
       throwError CantInvert
     neutralArg = throwError NeutralArg
 
-    isVarOrIrrelevant :: Res -> (I.Arg Term, Term) -> ExceptT InvertExcept TCM Res
+    isVarOrIrrelevant :: Res -> (Arg Term, Term) -> ExceptT InvertExcept TCM Res
     isVarOrIrrelevant vars (arg, t) =
       case ignoreSharing <$> arg of
         -- i := x
@@ -1108,8 +1106,7 @@ inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
               | length fs == length vs -> do
                 let aux (Arg _ v) (Arg info' f) = (Arg ai v,) $ t `applyE` [Proj f] where
                      ai = ArgInfo
-                       { argInfoColors    = argInfoColors info -- TODO guilhem
-                       , argInfoHiding    = min (getHiding info) (getHiding info')
+                       { argInfoHiding    = min (getHiding info) (getHiding info')
                        , argInfoRelevance = max (getRelevance info) (getRelevance info')
                        }
                 res <- loop $ zipWith aux vs fs
@@ -1135,7 +1132,6 @@ inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
         Arg _ Pi{}       -> neutralArg
         Arg _ Sort{}     -> neutralArg
         Arg _ Level{}    -> neutralArg
-        Arg _ ExtLam{}   -> __IMPOSSIBLE__
 
         Arg info (Shared p) -> isVarOrIrrelevant vars (Arg info $ derefPtr p, t)
 
@@ -1144,8 +1140,8 @@ inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
     append res vars = foldr cons vars res
 
     -- adding an irrelevant entry only if not present
-    cons :: (I.Arg Nat, Term) -> Res -> Res
-    cons a@(Arg (ArgInfo _ Irrelevant _) i, t) vars    -- TODO? UnusedArg?!
+    cons :: (Arg Nat, Term) -> Res -> Res
+    cons a@(Arg (ArgInfo _ Irrelevant) i, t) vars    -- TODO? UnusedArg?!
       | any ((i==) . unArg . fst) vars  = vars
       | otherwise                       = a : vars
     -- adding a relevant entry:
