@@ -12,9 +12,8 @@ import Control.Monad.Writer
 import qualified Data.Map as Map
 import qualified Data.Set as Set
 import qualified Data.Foldable as Fold
-import qualified Data.Traversable as Trav
 
-import Agda.Syntax.Common as Common
+import Agda.Syntax.Common
 import Agda.Syntax.Internal
 import Agda.Syntax.Position
 import Agda.Syntax.Scope.Base
@@ -25,6 +24,7 @@ import Agda.TypeChecking.Monad.Closure
 import Agda.TypeChecking.Monad.Options (reportSLn)
 import Agda.TypeChecking.Monad.Context
 import Agda.TypeChecking.Substitute
+import {-# SOURCE #-} Agda.TypeChecking.Telescope
 
 import Agda.Utils.Functor ((<.>))
 import Agda.Utils.Lens
@@ -87,7 +87,7 @@ getMetaTypeInContext m = do
   case j of
     HasType{ jMetaType = t } -> do
       vs <- getContextArgs
-      return $ piApply t $ permute (takeP (size vs) p) vs
+      piApplyM t $ permute (takeP (size vs) p) vs
     IsSort{}                 -> __IMPOSSIBLE__
 
 -- | Check whether all metas are instantiated.
@@ -126,7 +126,7 @@ instance IsInstantiatedMeta a => IsInstantiatedMeta [a] where
 instance IsInstantiatedMeta a => IsInstantiatedMeta (Maybe a) where
   isInstantiatedMeta = isInstantiatedMeta . maybeToList
 
-instance IsInstantiatedMeta a => IsInstantiatedMeta (Common.Arg c a) where
+instance IsInstantiatedMeta a => IsInstantiatedMeta (Arg a) where
   isInstantiatedMeta = isInstantiatedMeta . unArg
 
 -- | Does not worry about raising.
