@@ -6,6 +6,9 @@
 {-# LANGUAGE PatternGuards      #-}
 {-# LANGUAGE StandaloneDeriving #-}
 {-# LANGUAGE TupleSections      #-}
+{-# LANGUAGE FunctionalDependencies #-}
+{-# LANGUAGE MultiParamTypeClasses  #-}
+{-# LANGUAGE UndecidableInstances   #-}
 
 #if __GLASGOW_HASKELL__ <= 708
 {-# LANGUAGE OverlappingInstances #-}
@@ -15,24 +18,25 @@
 
 module Agda.TypeChecking.Substitute
   ( module Agda.TypeChecking.Substitute
-  , Substitution(..)
+  , Substitution'(..), Substitution
   ) where
 
-import Control.Arrow ((***), first, second)
+import Control.Arrow ((***), second)
 
 import Data.Function
 import Data.Functor
 import Data.List hiding (sort, drop)
 import qualified Data.List as List
 import Data.Map (Map)
+import Data.Maybe
 import Data.Monoid
 import Data.Typeable (Typeable)
 
 import Debug.Trace (trace)
 
-import Agda.Syntax.Common hiding (Arg, Dom, NamedArg)
-import qualified Agda.Syntax.Common as Common
+import Agda.Syntax.Common
 import Agda.Syntax.Internal
+import Agda.Syntax.Internal.Pattern
 import Agda.Syntax.Position
 
 import Agda.TypeChecking.Monad.Base
@@ -43,7 +47,6 @@ import Agda.TypeChecking.Positivity.Occurrence as Occ
 import Agda.Utils.Empty
 import Agda.Utils.Functor
 import Agda.Utils.List
-import Agda.Utils.Monad
 import Agda.Utils.Permutation
 import Agda.Utils.Size
 import Agda.Utils.Tuple
@@ -65,9 +68,13 @@ class Apply t where
   applyE t es  = apply  t $ map argFromElim es
     -- precondition: all @es@ are @Apply@s
 
--- | Apply to a single argument.
+-- | Apply to some default arguments.
+applys :: Apply t => t -> [Term] -> t
+applys t vs = apply t $ map defaultArg vs
+
+-- | Apply to a single default argument.
 apply1 :: Apply t => t -> Term -> t
-apply1 t u = apply t [ defaultArg u ]
+apply1 t u = applys t [ u ]
 
 instance Apply Term where
   applyE m [] = m
@@ -88,7 +95,6 @@ instance Apply Term where
       Sort _      -> __IMPOSSIBLE__
       DontCare mv -> dontCare $ mv `applyE` es  -- Andreas, 2011-10-02
         -- need to go under DontCare, since "with" might resurrect irrelevant term
-      ExtLam _ _  -> __IMPOSSIBLE__
 
 -- | If $v$ is a record value, @canProject f v@
 --   returns its field @f@.
@@ -133,18 +139,35 @@ defApp f [] (Apply a : es) | Just v <- canProject f (unArg a)
 defApp f es0 es = Def f $ es0 ++ es
 
 -- protect irrelevant fields (see issue 610)
-argToDontCare :: Common.Arg c Term -> Term
-argToDontCare (Common.Arg ai v)
+argToDontCare :: Arg Term -> Term
+argToDontCare (Arg ai v)
   | Irrelevant <- getRelevance ai     = dontCare v
   | otherwise                         = v
 
-instance Apply Type where
-  apply = piApply
-  -- Maybe an @applyE@ instance would be useful here as well.
-  -- A record type could be applied to a projection name
-  -- to yield the field type.
-  -- However, this works only in the monad where we can
-  -- look up the fields of a record type.
+-- Andreas, 2016-01-19: In connection with debugging issue #1783,
+-- I consider the Apply instance for Type harmful, as piApply is not
+-- safe if the type is not sufficiently reduced.
+-- (piApply is not in the monad and hence cannot unfold type synonyms).
+--
+-- Without apply for types, one has to at least use piApply and be
+-- aware of doing something which has a precondition
+-- (type sufficiently reduced).
+--
+-- By grepping for piApply, one can quickly get an overview over
+-- potentially harmful uses.
+--
+-- In general, piApplyM is preferable over piApply since it is more robust
+-- and fails earlier than piApply, which may only fail at serialization time,
+-- when all thunks are forced.
+
+-- REMOVED:
+-- instance Apply Type where
+--   apply = piApply
+--   -- Maybe an @applyE@ instance would be useful here as well.
+--   -- A record type could be applied to a projection name
+--   -- to yield the field type.
+--   -- However, this works only in the monad where we can
+--   -- look up the fields of a record type.
 
 instance Apply Sort where
   applyE s [] = s
@@ -155,14 +178,14 @@ instance Apply a => Apply (Ptr a) where
 
 -- @applyE@ does not make sense for telecopes, definitions, clauses etc.
 
-instance Subst a => Apply (Tele a) where
+instance Subst Term a => Apply (Tele a) where
   apply tel               []       = tel
   apply EmptyTel          _        = __IMPOSSIBLE__
   apply (ExtendTel _ tel) (t : ts) = lazyAbsApp tel (unArg t) `apply` ts
 
 instance Apply Definition where
-  apply (Defn info x t pol occ df m c inst d) args =
-    Defn info x (piApply t args) (apply pol args) (apply occ args) df m c inst (apply d args)
+  apply (Defn info x t pol occ df m c inst copy d) args =
+    Defn info x (piApply t args) (apply pol args) (apply occ args) df m c inst copy (apply d args)
 
 instance Apply RewriteRule where
   apply r args = RewriteRule
@@ -247,7 +270,7 @@ instance Apply Defn where
         }
     Record{ recPars = np, recConType = t, recClause = cl, recTel = tel
           {-, recArgOccurrences = occ-} } ->
-      d { recPars = np - size args, recConType = apply t args
+      d { recPars = np - size args, recConType = piApply t args
         , recClause = apply cl args, recTel = apply tel args
 --        , recArgOccurrences = List.drop (length args) occ
         }
@@ -260,27 +283,24 @@ instance Apply PrimFun where
     apply (PrimFun x ar def) args   = PrimFun x (ar - size args) $ \vs -> def (args ++ vs)
 
 instance Apply Clause where
-    apply (Clause r tel perm ps b t) args =
+    apply (Clause r tel ps b t catchall) args =
       Clause r
              (apply tel args)
-             (apply perm args)
              (List.drop (size args) ps)
              (apply b args)
              (applySubst (parallelS (map unArg args)) t)
+             catchall
 
 instance Apply CompiledClauses where
   apply cc args = case cc of
     Fail     -> Fail
     Done hs t
-      | length hs >= len -> Done (List.drop len hs)
-                                 (applySubst
-                                    (parallelS $
-                                       [ var i | i <- [0..length hs - len - 1]] ++
-                                       map unArg args)
-                                    t)
+      | length hs >= len ->
+         let sub = parallelS $ map var [0..length hs - len - 1] ++ map unArg args
+         in  Done (List.drop len hs) $ applySubst sub t
       | otherwise -> __IMPOSSIBLE__
     Case n bs
-      | n >= len  -> Case (n - len) (apply bs args)
+      | unArg n >= len -> Case (n <&> \ m -> m - len) (apply bs args)
       | otherwise -> __IMPOSSIBLE__
     where
       len = length args
@@ -364,8 +384,12 @@ instance Abstract Permutation where
     where
       m = size tel
 
--- | The type must contain the right number of pis without have to perform any
--- reduction.
+-- | @(x:A)->B(x) `piApply` [u] = B(u)@
+--
+--   Precondition: The type must contain the right number of pis without
+--   having to perform any reduction.
+--
+--   @piApply@ is potentially unsafe, the monadic 'piApplyM' is preferable.
 piApply :: Type -> Args -> Type
 piApply t []                      = t
 piApply (El _ (Pi  _ b)) (a:args) = lazyAbsApp b (unArg a) `piApply` args
@@ -396,8 +420,8 @@ instance Abstract Telescope where
   ExtendTel arg xtel `abstract` tel = ExtendTel arg $ xtel <&> (`abstract` tel)
 
 instance Abstract Definition where
-  abstract tel (Defn info x t pol occ df m c inst d) =
-    Defn info x (abstract tel t) (abstract tel pol) (abstract tel occ) df m c inst (abstract tel d)
+  abstract tel (Defn info x t pol occ df m c inst copy d) =
+    Defn info x (abstract tel t) (abstract tel pol) (abstract tel occ) df m c inst copy (abstract tel d)
 
 -- | @tel ⊢ (Γ ⊢ lhs ↦ rhs : t)@ becomes @tel, Γ ⊢ lhs ↦ rhs : t)@
 --   we do not need to change lhs, rhs, and t since they live in Γ.
@@ -474,16 +498,18 @@ instance Abstract PrimFun where
         where n = size tel
 
 instance Abstract Clause where
-  abstract tel (Clause r tel' perm ps b t) =
-    Clause r (abstract tel tel') (abstract tel perm)
-           (namedTelVars tel ++ ps) (abstract tel b)
+  abstract tel (Clause r tel' ps b t catchall) =
+    Clause r (abstract tel tel')
+           (namedTelVars m tel ++ ps) (abstract tel b)
            t -- nothing to do for t, since it lives under the telescope
+           catchall
+      where m = size tel + size tel'
 
 instance Abstract CompiledClauses where
   abstract tel Fail = Fail
   abstract tel (Done xs t) = Done (map (argFromDom . fmap fst) (telToList tel) ++ xs) t
   abstract tel (Case n bs) =
-    Case (n + fromIntegral (size tel)) (abstract tel bs)
+    Case (n <&> \ i -> i + fromIntegral (size tel)) (abstract tel bs)
 
 instance Abstract a => Abstract (WithArity a) where
   abstract tel (WithArity n a) = WithArity n $ abstract tel a
@@ -492,14 +518,14 @@ instance Abstract a => Abstract (Case a) where
   abstract tel (Branches cop cs ls m) =
     Branches cop (abstract tel cs) (abstract tel ls) (abstract tel m)
 
-telVars :: Telescope -> [Arg Pattern]
-telVars = map (fmap namedThing) . namedTelVars
+telVars :: Int -> Telescope -> [Arg DeBruijnPattern]
+telVars m = map (fmap namedThing) . (namedTelVars m)
 
-namedTelVars :: Telescope -> [NamedArg Pattern]
-namedTelVars EmptyTel                            = []
-namedTelVars (ExtendTel (Common.Dom info a) tel) =
-  Common.Arg info (namedVarP $ absName tel) :
-  namedTelVars (unAbs tel)
+namedTelVars :: Int -> Telescope -> [NamedArg DeBruijnPattern]
+namedTelVars m EmptyTel                     = []
+namedTelVars m (ExtendTel (Dom info a) tel) =
+  Arg info (namedDBVarP (m-1) $ absName tel) :
+  namedTelVars (m-1) (unAbs tel)
 
 instance Abstract FunctionInverse where
   abstract tel NotInjective  = NotInjective
@@ -525,7 +551,7 @@ instance Abstract v => Abstract (Map k v) where
 abstractArgs :: Abstract a => Args -> a -> a
 abstractArgs args x = abstract tel x
     where
-        tel   = foldr (\(Common.Arg info x) -> ExtendTel (Common.Dom info $ sort Prop) . Abs x)
+        tel   = foldr (\(Arg info x) -> ExtendTel (Dom info $ sort Prop) . Abs x)
                       EmptyTel
               $ zipWith (<$) names args
         names = cycle $ map (stringToArgName . (:[])) ['a'..'z']
@@ -534,38 +560,51 @@ abstractArgs args x = abstract tel x
 -- * Explicit substitutions
 ---------------------------------------------------------------------------
 
+class DeBruijn a where
+  debruijnVar  :: Int -> a
+  debruijnVar = debruijnNamedVar underscore
+  debruijnNamedVar :: String -> Int -> a
+  debruijnNamedVar _ = debruijnVar
+  debruijnView :: a -> Maybe Int
+
+instance DeBruijn Term where
+  debruijnVar = var
+  debruijnView (Var n []) = Just n
+  debruijnView _ = Nothing
+
 -- See Syntax.Internal for the definition.
 
-idS :: Substitution
+idS :: Substitution' a
 idS = IdS
 
-wkS :: Int -> Substitution -> Substitution
+wkS :: Int -> Substitution' a -> Substitution' a
 wkS 0 rho        = rho
 wkS n (Wk m rho) = Wk (n + m) rho
 wkS n EmptyS     = EmptyS
 wkS n rho        = Wk n rho
 
-raiseS :: Int -> Substitution
+raiseS :: Int -> Substitution' a
 raiseS n = wkS n idS
 
-consS :: Term -> Substitution -> Substitution
-consS (Var n []) (Wk m rho)
-  | n + 1 == m = wkS (m - 1) (liftS 1 rho)
+consS :: DeBruijn a => a -> Substitution' a -> Substitution' a
+consS t (Wk m rho)
+  | Just n <- debruijnView t,
+    n + 1 == m = wkS (m - 1) (liftS 1 rho)
 consS u rho = seq u (u :# rho)
 
 -- | To replace index @n@ by term @u@, do @applySubst (singletonS n u)@.
-singletonS :: Int -> Term -> Substitution
-singletonS n u = map var [0..n-1] ++# consS u (raiseS n)
-  -- ALT: foldl (\ s i -> var i `consS` s) (consS u $ raiseS n) $ downFrom n
+singletonS :: DeBruijn a => Int -> a -> Substitution' a
+singletonS n u = map debruijnVar [0..n-1] ++# consS u (raiseS n)
+  -- ALT: foldl (\ s i -> debruijnVar i `consS` s) (consS u $ raiseS n) $ downFrom n
 
 -- | Lift a substitution under k binders.
-liftS :: Int -> Substitution -> Substitution
+liftS :: Int -> Substitution' a -> Substitution' a
 liftS 0 rho          = rho
 liftS k IdS          = IdS
 liftS k (Lift n rho) = Lift (n + k) rho
 liftS k rho          = Lift k rho
 
-dropS :: Int -> Substitution -> Substitution
+dropS :: Int -> Substitution' a -> Substitution' a
 dropS 0 rho                = rho
 dropS n IdS                = raiseS n
 dropS n (Wk m rho)         = wkS m (dropS n rho)
@@ -576,7 +615,7 @@ dropS n (Lift m rho)       = wkS 1 $ dropS (n - 1) $ liftS (m - 1) rho
 dropS n EmptyS             = __IMPOSSIBLE__
 
 -- | @applySubst (ρ `composeS` σ) v == applySubst ρ (applySubst σ v)@
-composeS :: Substitution -> Substitution -> Substitution
+composeS :: Subst a a => Substitution' a -> Substitution' a -> Substitution' a
 composeS rho IdS = rho
 composeS IdS sgm = sgm
 composeS rho EmptyS = EmptyS
@@ -590,7 +629,7 @@ composeS rho (Lift n sgm) = lookupS rho 0 :# composeS rho (wkS 1 (liftS (n - 1)
 -- If Γ ⊢ ρ : Δ, Θ then splitS |Θ| ρ = (σ, δ), with
 --   Γ ⊢ σ : Δ
 --   Γ ⊢ δ : Θσ
-splitS :: Int -> Substitution -> (Substitution, Substitution)
+splitS :: Int -> Substitution' a -> (Substitution' a, Substitution' a)
 splitS 0 rho                  = (rho, EmptyS)
 splitS n (u :# rho)           = second (u :#) $ splitS (n - 1) rho
 splitS n (Strengthen err rho) = second (Strengthen err) $ splitS (n - 1) rho
@@ -602,32 +641,32 @@ splitS n EmptyS               = __IMPOSSIBLE__
 
 infixr 4 ++#
 
-(++#) :: [Term] -> Substitution -> Substitution
+(++#) :: DeBruijn a => [a] -> Substitution' a -> Substitution' a
 us ++# rho = foldr consS rho us
 
-prependS :: Empty -> [Maybe Term] -> Substitution -> Substitution
+prependS :: DeBruijn a => Empty -> [Maybe a] -> Substitution' a -> Substitution' a
 prependS err us rho = foldr f rho us
   where
     f Nothing  rho = Strengthen err rho
     f (Just u) rho = consS u rho
 
-parallelS :: [Term] -> Substitution
+parallelS :: DeBruijn a => [a] -> Substitution' a
 parallelS us = us ++# idS
 
-compactS :: Empty -> [Maybe Term] -> Substitution
+compactS :: DeBruijn a => Empty -> [Maybe a] -> Substitution' a
 compactS err us = prependS err us idS
 
 -- | Γ ⊢ (strengthenS ⊥ |Δ|) : Γ,Δ
-strengthenS :: Empty -> Int -> Substitution
+strengthenS :: Empty -> Int -> Substitution' a
 strengthenS err n
   | n < 0     = __IMPOSSIBLE__
   | otherwise = iterate (Strengthen err) idS !! n
 
-lookupS :: Substitution -> Nat -> Term
+lookupS :: Subst a a => Substitution' a -> Nat -> a
 lookupS rho i = case rho of
-  IdS                    -> var i
+  IdS                    -> debruijnVar i
   Wk n IdS               -> let j = i + n in
-                            if  j < 0 then __IMPOSSIBLE__ else var j
+                            if  j < 0 then __IMPOSSIBLE__ else debruijnVar j
   Wk n rho               -> applySubst (raiseS n) (lookupS rho i)
   u :# rho   | i == 0    -> u
              | i < 0     -> __IMPOSSIBLE__
@@ -636,7 +675,7 @@ lookupS rho i = case rho of
              | i == 0    -> absurd err
              | i < 0     -> __IMPOSSIBLE__
              | otherwise -> lookupS rho (i - 1)
-  Lift n rho | i < n     -> var i
+  Lift n rho | i < n     -> debruijnVar i
              | otherwise -> raise n $ lookupS rho (i - n)
   EmptyS                 -> __IMPOSSIBLE__
 
@@ -653,31 +692,31 @@ lookupS rho i = case rho of
 -- -----------
 -- Γ ⊢ tρ : σρ
 
-class Subst t where
-  applySubst :: Substitution -> t -> t
+class DeBruijn t => Subst t a | a -> t where
+  applySubst :: Substitution' t -> a -> a
 
-raise :: Subst t => Nat -> t -> t
+raise :: Subst t a => Nat -> a -> a
 raise = raiseFrom 0
 
-raiseFrom :: Subst t => Nat -> Nat -> t -> t
+raiseFrom :: Subst t a => Nat -> Nat -> a -> a
 raiseFrom n k = applySubst (liftS n $ raiseS k)
 
 -- | Replace de Bruijn index i by a 'Term' in something.
-subst :: Subst t => Int -> Term -> t -> t
+subst :: Subst t a => Int -> t -> a -> a
 subst i u = applySubst $ singletonS i u
 
-strengthen :: Subst t => Empty -> t -> t
+strengthen :: Subst t a => Empty -> a -> a
 strengthen err = applySubst (compactS err [Nothing])
 
 -- | Replace what is now de Bruijn index 0, but go under n binders.
 --   @substUnder n u == subst n (raise n u)@.
-substUnder :: Subst t => Nat -> Term -> t -> t
+substUnder :: Subst t a => Nat -> t -> a -> a
 substUnder n u = applySubst (liftS n (singletonS 0 u))
 
-instance Subst Substitution where
+instance Subst a a => Subst a (Substitution' a) where
   applySubst rho sgm = composeS rho sgm
 
-instance Subst Term where
+instance Subst Term Term where
   applySubst IdS t = t
   applySubst rho t    = case t of
     Var i es    -> lookupS rho i `applyE` applySubst rho es
@@ -691,15 +730,14 @@ instance Subst Term where
     Sort s      -> sortTm $ applySubst rho s
     Shared p    -> Shared $ applySubst rho p
     DontCare mv -> dontCare $ applySubst rho mv
-    ExtLam cs es-> ExtLam (applySubst rho cs) (applySubst rho es)
 
-instance Subst a => Subst (Ptr a) where
+instance Subst t a => Subst t (Ptr a) where
   applySubst rho = fmap (applySubst rho)
 
-instance (Subst a) => Subst (Type' a) where
+instance Subst Term a => Subst Term (Type' a) where
   applySubst rho (El s t) = applySubst rho s `El` applySubst rho t
 
-instance Subst Sort where
+instance Subst Term Sort where
   applySubst rho s = case s of
     Type n     -> levelSort $ sub n
     Prop       -> Prop
@@ -708,36 +746,33 @@ instance Subst Sort where
     DLub s1 s2 -> DLub (sub s1) (sub s2)
     where sub x = applySubst rho x
 
-instance Subst Level where
+instance Subst Term Level where
   applySubst rho (Max as) = Max $ applySubst rho as
 
-instance Subst PlusLevel where
+instance Subst Term PlusLevel where
   applySubst rho l@ClosedLevel{} = l
   applySubst rho (Plus n l) = Plus n $ applySubst rho l
 
-instance Subst LevelAtom where
+instance Subst Term LevelAtom where
   applySubst rho (MetaLevel m vs)   = MetaLevel m    $ applySubst rho vs
   applySubst rho (BlockedLevel m v) = BlockedLevel m $ applySubst rho v
   applySubst rho (NeutralLevel _ v) = UnreducedLevel $ applySubst rho v
   applySubst rho (UnreducedLevel v) = UnreducedLevel $ applySubst rho v
 
-instance Subst Bool where
+instance Subst Term Name where
   applySubst rho = id
 
 #if __GLASGOW_HASKELL__ >= 710
-instance {-# OVERLAPPING #-} Subst String where
+instance {-# OVERLAPPING #-} Subst Term String where
 #else
-instance Subst String where
+instance Subst Term String where
 #endif
   applySubst rho = id
 
-instance Subst Name where
-  applySubst rho = id
-
-instance Subst ConPatternInfo where
+instance Subst Term ConPatternInfo where
   applySubst rho (ConPatternInfo mr mt) = ConPatternInfo mr $ applySubst rho mt
 
-instance Subst Pattern where
+instance Subst Term Pattern where
   applySubst rho p = case p of
     ConP c mt ps -> ConP c (applySubst rho mt) $ applySubst rho ps
     DotP t       -> DotP $ applySubst rho t
@@ -745,9 +780,9 @@ instance Subst Pattern where
     LitP l       -> p
     ProjP _      -> p
 
-instance Subst NLPat where
+instance Subst Term NLPat where
   applySubst rho p = case p of
-    PVar i -> p
+    PVar id i -> p
     PWild  -> p
     PDef f es -> PDef f $ applySubst rho es
     PLam i u -> PLam i $ applySubst rho u
@@ -755,26 +790,34 @@ instance Subst NLPat where
     PBoundVar i es -> PBoundVar i $ applySubst rho es
     PTerm u -> PTerm $ applySubst rho u
 
-instance Subst t => Subst (Blocked t) where
+instance Subst Term RewriteRule where
+  applySubst rho (RewriteRule q gamma lhs rhs t) =
+    RewriteRule q (applySubst rho gamma)
+                  (applySubst (liftS n rho) lhs)
+                  (applySubst (liftS n rho) rhs)
+                  (applySubst (liftS n rho) t)
+    where n = size gamma
+
+instance Subst t a => Subst t (Blocked a) where
   applySubst rho b = fmap (applySubst rho) b
 
-instance Subst DisplayForm where
+instance Subst Term DisplayForm where
   applySubst rho (Display n ps v) =
     Display n (applySubst (liftS 1 rho) ps)
               (applySubst (liftS n rho) v)
 
-instance Subst DisplayTerm where
+instance Subst Term DisplayTerm where
   applySubst rho (DTerm v)        = DTerm $ applySubst rho v
   applySubst rho (DDot v)         = DDot  $ applySubst rho v
   applySubst rho (DCon c vs)      = DCon c $ applySubst rho vs
   applySubst rho (DDef c es)      = DDef c $ applySubst rho es
   applySubst rho (DWithApp v vs ws) = uncurry3 DWithApp $ applySubst rho (v, vs, ws)
 
-instance Subst a => Subst (Tele a) where
+instance Subst t a => Subst t (Tele a) where
   applySubst rho  EmptyTel         = EmptyTel
   applySubst rho (ExtendTel t tel) = uncurry ExtendTel $ applySubst rho (t, tel)
 
-instance Subst Constraint where
+instance Subst Term Constraint where
   applySubst rho c = case c of
     ValueCmp cmp a u v       -> ValueCmp cmp (rf a) (rf u) (rf v)
     ElimCmp ps a v e1 e2     -> ElimCmp ps (rf a) (rf v) (rf e1) (rf e2)
@@ -785,56 +828,65 @@ instance Subst Constraint where
     Guarded c cs             -> Guarded (rf c) cs
     IsEmpty r a              -> IsEmpty r (rf a)
     CheckSizeLtSat t         -> CheckSizeLtSat (rf t)
-    FindInScope m cands      -> FindInScope m (rf cands)
+    FindInScope m b cands    -> FindInScope m b (rf cands)
     UnBlock{}                -> c
     where
       rf x = applySubst rho x
 
-instance Subst a => Subst (Elim' a) where
+instance Subst t a => Subst t (Elim' a) where
   applySubst rho e = case e of
     Apply v -> Apply $ applySubst rho v
     Proj{}  -> e
 
-instance Subst a => Subst (Abs a) where
+instance Subst t a => Subst t (Abs a) where
   applySubst rho (Abs x a)   = Abs x $ applySubst (liftS 1 rho) a
   applySubst rho (NoAbs x a) = NoAbs x $ applySubst rho a
 
-instance Subst a => Subst (Arg a) where
+instance Subst t a => Subst t (Arg a) where
   applySubst rho = fmap (applySubst rho)
 
-instance Subst a => Subst (Named name a) where
+instance Subst t a => Subst t (Named name a) where
   applySubst rho = fmap (applySubst rho)
 
-instance Subst a => Subst (Dom a) where
+instance Subst t a => Subst t (Dom a) where
   applySubst rho = fmap (applySubst rho)
 
-instance Subst a => Subst (Maybe a) where
+instance Subst t a => Subst t (Maybe a) where
   applySubst rho = fmap (applySubst rho)
 
-instance Subst a => Subst [a] where
+instance Subst t a => Subst t [a] where
   applySubst rho = map (applySubst rho)
 
-instance Subst () where
+instance Subst Term () where
   applySubst _ _ = ()
 
-instance (Subst a, Subst b) => Subst (a,b) where
+instance (Subst t a, Subst t b) => Subst t (a, b) where
   applySubst rho (x,y) = (applySubst rho x, applySubst rho y)
 
-instance (Subst a, Subst b, Subst c) => Subst (a,b,c) where
+instance (Subst t a, Subst t b, Subst t c) => Subst t (a, b, c) where
   applySubst rho (x,y,z) = (applySubst rho x, applySubst rho y, applySubst rho z)
 
-instance (Subst a, Subst b, Subst c, Subst d) => Subst (a,b,c,d) where
+instance (Subst t a, Subst t b, Subst t c, Subst t d) => Subst t (a, b, c, d) where
   applySubst rho (x,y,z,u) = (applySubst rho x, applySubst rho y, applySubst rho z, applySubst rho u)
 
-instance Subst ClauseBody where
+instance Subst Term ClauseBody where
   applySubst rho (Body t) = Body $ applySubst rho t
   applySubst rho (Bind b) = Bind $ applySubst rho b
   applySubst _   NoBody   = NoBody
 
-instance Subst Clause where
-  -- NOTE: This only happens when reifying extended lambdas, in which case there are
-  -- no interesting dot patterns and we don't care about the type.
-  applySubst rho c = c { clauseBody = applySubst rho $ clauseBody c }
+instance Subst Term Candidate where
+  applySubst rho (Candidate u t eti) = Candidate (applySubst rho u) (applySubst rho t) eti
+
+instance Subst Term EqualityView where
+  applySubst rho (OtherType t) = OtherType
+    (applySubst rho t)
+  applySubst rho (EqualityType s eq l t a b) = EqualityType
+    (applySubst rho s)
+    eq
+    (applySubst rho l)
+    (applySubst rho t)
+    (applySubst rho a)
+    (applySubst rho b)
 
 ---------------------------------------------------------------------------
 -- * Telescopes
@@ -842,7 +894,10 @@ instance Subst Clause where
 
 type TelView = TelV Type
 data TelV a  = TelV { theTel :: Tele (Dom a), theCore :: a }
-  deriving (Typeable, Show, Eq, Ord, Functor)
+  deriving (Typeable, Show, Functor)
+
+deriving instance (Subst t a, Eq  a) => Eq  (TelV a)
+deriving instance (Subst t a, Ord a) => Ord (TelV a)
 
 type ListTel' a = [Dom (a, Type)]
 type ListTel = ListTel' ArgName
@@ -850,7 +905,7 @@ type ListTel = ListTel' ArgName
 telFromList' :: (a -> ArgName) -> ListTel' a -> Telescope
 telFromList' f = foldr extTel EmptyTel
   where
-    extTel (Common.Dom info (x, a)) = ExtendTel (Common.Dom info a) . Abs (f x)
+    extTel (Dom info (x, a)) = ExtendTel (Dom info a) . Abs (f x)
 
 telFromList :: ListTel -> Telescope
 telFromList = telFromList' id
@@ -862,7 +917,7 @@ telToList (ExtendTel arg tel) = fmap (absName tel,) arg : telToList (absBody tel
   -- since 'absBody' raises.
 
 telToArgs :: Telescope -> [Arg ArgName]
-telToArgs tel = [ Common.Arg (domInfo d) (fst $ unDom d) | d <- telToList tel ]
+telToArgs tel = [ Arg (domInfo d) (fst $ unDom d) | d <- telToList tel ]
 
 -- | Turn a typed binding @(x1 .. xn : A)@ into a telescope.
 bindsToTel' :: (Name -> a) -> [Name] -> Dom Type -> ListTel' a
@@ -898,7 +953,7 @@ telView'UpTo n t = case ignoreSharing $ unEl t of
 
 -- | @mkPi dom t = telePi (telFromList [dom]) t@
 mkPi :: Dom (ArgName, Type) -> Type -> Type
-mkPi (Common.Dom info (x, a)) b = el $ Pi (Common.Dom info a) (mkAbs x b)
+mkPi (Dom info (x, a)) b = el $ Pi (Dom info a) (mkAbs x b)
   where
     el = El $ dLub (getSort a) (Abs x (getSort b)) -- dLub checks x freeIn
 
@@ -943,7 +998,7 @@ class TeleNoAbs a where
   teleNoAbs :: a -> Term -> Term
 
 instance TeleNoAbs ListTel where
-  teleNoAbs tel t = foldr (\ (Common.Dom ai (x, _)) -> Lam ai . NoAbs x) t tel
+  teleNoAbs tel t = foldr (\ (Dom ai (x, _)) -> Lam ai . NoAbs x) t tel
 
 instance TeleNoAbs Telescope where
   teleNoAbs tel = teleNoAbs $ telToList tel
@@ -955,7 +1010,7 @@ instance TeleNoAbs Telescope where
 dLub :: Sort -> Abs Sort -> Sort
 dLub s1 (NoAbs _ s2) = sLub s1 s2
 dLub s1 b@(Abs _ s2) = case occurrence 0 s2 of
-  Flexible      -> DLub s1 b
+  Flexible _    -> DLub s1 b
   Irrelevantly  -> DLub s1 b
   NoOccurrence  -> sLub s1 (noabsApp __IMPOSSIBLE__ b)
 --  Free.Unused   -> sLub s1 (absApp b __IMPOSSIBLE__) -- triggers Issue784
@@ -970,31 +1025,31 @@ dLub s1 b@(Abs _ s2) = case occurrence 0 s2 of
 ---------------------------------------------------------------------------
 
 -- | Instantiate an abstraction. Strict in the term.
-absApp :: Subst t => Abs t -> Term -> t
+absApp :: Subst t a => Abs a -> t -> a
 absApp (Abs   _ v) u = subst 0 u v
 absApp (NoAbs _ v) _ = v
 
 -- | Instantiate an abstraction. Lazy in the term, which allow it to be
 --   __IMPOSSIBLE__ in the case where the variable shouldn't be used but we
 --   cannot use 'noabsApp'. Used in Apply.
-lazyAbsApp :: Subst t => Abs t -> Term -> t
+lazyAbsApp :: Subst t a => Abs a -> t -> a
 lazyAbsApp (Abs   _ v) u = applySubst (u :# IdS) v  -- Note: do not use consS here!
 lazyAbsApp (NoAbs _ v) _ = v
 
 -- | Instantiate an abstraction that doesn't use its argument.
-noabsApp :: Subst t => Empty -> Abs t -> t
+noabsApp :: Subst t a => Empty -> Abs a -> a
 noabsApp err (Abs   _ v) = strengthen err v
 noabsApp _   (NoAbs _ v) = v
 
-absBody :: Subst t => Abs t -> t
+absBody :: Subst t a => Abs a -> a
 absBody (Abs   _ v) = v
 absBody (NoAbs _ v) = raise 1 v
 
-mkAbs :: (Subst a, Free a) => ArgName -> a -> Abs a
+mkAbs :: (Subst t a, Free a) => ArgName -> a -> Abs a
 mkAbs x v | 0 `freeIn` v = Abs x v
           | otherwise    = NoAbs x (raise (-1) v)
 
-reAbs :: (Subst a, Free a) => Abs a -> Abs a
+reAbs :: (Subst t a, Free a) => Abs a -> Abs a
 reAbs (NoAbs x v) = NoAbs x v
 reAbs (Abs x v)   = mkAbs x v
 
@@ -1004,7 +1059,7 @@ reAbs (Abs x v)   = mkAbs x v
 --   at point of application of @k@ and the content of @b@
 --   are at the same context.
 --   Precondition: @a@ and @b@ are at the same context at call time.
-underAbs :: Subst a => (a -> b -> b) -> a -> Abs b -> Abs b
+underAbs :: Subst t a => (a -> b -> b) -> a -> Abs b -> Abs b
 underAbs cont a b = case b of
   Abs   x t -> Abs   x $ cont (raise 1 a) t
   NoAbs x t -> NoAbs x $ cont a t
@@ -1013,7 +1068,7 @@ underAbs cont a b = case b of
 --   performs operation @k@ on @a@ and the body of @b@,
 --   and puts the 'Lam's back.  @a@ is raised correctly
 --   according to the number of abstractions.
-underLambdas :: Subst a => Int -> (a -> Term -> Term) -> a -> Term -> Term
+underLambdas :: Subst Term a => Int -> (a -> Term -> Term) -> a -> Term -> Term
 underLambdas n cont a v = loop n a v where
   loop 0 a v = cont a v
   loop n a v = case ignoreSharing v of
@@ -1072,11 +1127,12 @@ deriving instance Eq NotBlocked
 deriving instance Ord NotBlocked
 deriving instance Eq t => Eq (Blocked t)
 deriving instance Ord t => Ord (Blocked t)
+deriving instance Eq Candidate
 
-deriving instance (Subst a, Eq a) => Eq (Elim' a)
-deriving instance (Subst a, Ord a) => Ord (Elim' a)
-deriving instance (Subst a, Eq a) => Eq (Tele a)
-deriving instance (Subst a, Ord a) => Ord (Tele a)
+deriving instance (Subst t a, Eq a)  => Eq  (Elim' a)
+deriving instance (Subst t a, Ord a) => Ord (Elim' a)
+deriving instance (Subst t a, Eq a)  => Eq  (Tele a)
+deriving instance (Subst t a, Ord a) => Ord (Tele a)
 
 deriving instance Eq Constraint
 deriving instance Eq Section
@@ -1116,8 +1172,6 @@ instance Eq Term where
   _          == _            = False
 
 instance Ord Term where
-  ExtLam{}   `compare` _          = __IMPOSSIBLE__
-  _          `compare` ExtLam{}   = __IMPOSSIBLE__
   Shared a   `compare` Shared x | a == x = EQ
   Shared a   `compare` x          = compare (derefPtr a) x
   a          `compare` Shared x   = compare a (derefPtr x)
@@ -1150,12 +1204,12 @@ instance Ord Term where
   _          `compare` MetaV{}    = GT
   DontCare{} `compare` DontCare{} = EQ
 
-instance (Subst a, Eq a) => Eq (Abs a) where
+instance (Subst t a, Eq a) => Eq (Abs a) where
   NoAbs _ a == NoAbs _ b = a == b
   Abs   _ a == Abs   _ b = a == b
   a         == b         = absBody a == absBody b
 
-instance (Subst a, Ord a) => Ord (Abs a) where
+instance (Subst t a, Ord a) => Ord (Abs a) where
   NoAbs _ a `compare` NoAbs _ b = a `compare` b
   Abs   _ a `compare` Abs   _ b = a `compare` b
   a         `compare` b         = absBody a `compare` absBody b
