@@ -12,7 +12,7 @@ import Control.Monad.State
 import qualified Data.List as List (sort, nub)
 
 import Agda.Syntax.Position
-import Agda.Syntax.Common as Common
+import Agda.Syntax.Common
 import Agda.Syntax.Literal
 import Agda.Syntax.Fixity
 import Agda.Syntax.Internal as I
@@ -154,7 +154,7 @@ makeConfiguration ds cs ps vs = TermConf
                         { nameId          = NameId n 1
                         , nameConcrete    = C.Name noRange [C.Id s]
                         , nameBindingSite = noRange
-                        , nameFixity      = defaultFixity'
+                        , nameFixity      = noFixity'
                         }
                       }
 
@@ -196,10 +196,10 @@ instance GenC Hiding where
       HiddenFreqs {hiddenFreq = hideF, notHiddenFreq = nohideF } =
         hiddenFreqs $ tcFrequencies conf
 
-instance (GenC c, GenC a) => GenC (Common.Arg c a) where
+instance GenC a => GenC (Arg a) where
   genC conf = (\ (h, a) -> Arg (setHiding h defaultArgInfo) a) <$> genC conf
 
-instance (GenC c, GenC a) => GenC (Common.Dom c a) where
+instance GenC a => GenC (Dom a) where
   genC conf = (\ (h, a) -> Dom (setHiding h defaultArgInfo) a) <$> genC conf
 
 instance GenC a => GenC (Abs a) where
@@ -245,7 +245,7 @@ instance GenC Integer where
 
 instance GenC Literal where
   genC conf = oneof (concat $ zipWith gen useLits
-              [ uncurry LitInt    <$> genC conf
+              [ uncurry LitNat    <$> genC conf
               , uncurry LitFloat  <$> genC conf
               , uncurry LitString <$> genC conf
               , uncurry LitChar   <$> genC conf
@@ -399,13 +399,14 @@ instance ShrinkC ConName ConHead where
   noShrink = unConName
 
 instance ShrinkC Literal Literal where
-  shrinkC _ (LitInt _ 0) = []
-  shrinkC conf l         = LitInt noRange 0 : case l of
-      LitInt    r n -> LitInt    r <$> shrink n
+  shrinkC _ (LitNat _ 0) = []
+  shrinkC conf l         = LitNat noRange 0 : case l of
+      LitNat    r n -> LitNat    r <$> shrink n
       LitString r s -> LitString r <$> shrinkC conf s
       LitChar   r c -> LitChar   r <$> shrinkC conf c
       LitFloat  r x -> LitFloat  r <$> shrink x
       LitQName  r x -> []
+      LitMeta{}     -> []
   noShrink = id
 
 instance ShrinkC Char Char where
@@ -424,11 +425,11 @@ instance ShrinkC a b => ShrinkC (Abs a) (Abs b) where
   shrinkC conf (Abs   s x) = Abs s <$> shrinkC (extendConf conf) x
   noShrink = fmap noShrink
 
-instance ShrinkC a b => ShrinkC (I.Arg a) (I.Arg b) where
+instance ShrinkC a b => ShrinkC (Arg a) (Arg b) where
   shrinkC conf (Arg info x) = (\ (h,x) -> Arg (setHiding h info) x) <$> shrinkC conf (argInfoHiding info, x)
   noShrink = fmap noShrink
 
-instance ShrinkC a b => ShrinkC (I.Dom a) (I.Dom b) where
+instance ShrinkC a b => ShrinkC (Dom a) (Dom b) where
   shrinkC conf (Dom info x) = (\ (h,x) -> Dom (setHiding h info) x) <$> shrinkC conf (argInfoHiding info, x)
   noShrink = fmap noShrink
 
@@ -484,7 +485,6 @@ instance ShrinkC Term Term where
                     (MetaV m <$> shrinkC conf (NoType es))
     DontCare _   -> []
     Shared{}     -> __IMPOSSIBLE__
-    ExtLam _ _   -> __IMPOSSIBLE__
     where
       validType t
         | not (tcIsType conf) = True
@@ -517,7 +517,6 @@ instance KillVar Term where
     MetaV m args           -> MetaV m     $ killVar i args
     DontCare mv            -> DontCare    $ killVar i mv
     Shared{}               -> __IMPOSSIBLE__
-    ExtLam _ _             -> __IMPOSSIBLE__
 
 instance KillVar Type where
   killVar i (El s t) = El s $ killVar i t
@@ -529,10 +528,10 @@ instance KillVar Telescope where
 instance KillVar a => KillVar (Elim' a) where
   killVar i = fmap (killVar i)
 
-instance KillVar a => KillVar (I.Arg a) where
+instance KillVar a => KillVar (Arg a) where
   killVar i = fmap (killVar i)
 
-instance KillVar a => KillVar (I.Dom a) where
+instance KillVar a => KillVar (Dom a) where
   killVar i = fmap (killVar i)
 
 instance KillVar a => KillVar (Abs a) where
